import os
import sys
import time
import uuid

import rclpy
from rclpy.action import ActionClient
from rclpy.node import Node

from threading import RLock
from typing import Any
from rclpy.task import Future

from lrs_msgs_common.srv import InAirCheck, GetHomePosition, GetCurrentPosition, SignalGoal, GimbalCheck
from lrs_msgs_common.action import FlyPath, Land, TakeOff

from geographic_msgs.msg import GeoPoint
from unique_identifier_msgs.msg import UUID

from lrs_util.coordtrans import CoordTrans

class GetCurrentPositionClient():
    def __init__(self, node, shutdown=False):
        self.node = node
        self.result = None
        self.shutdown = shutdown        
        self.cli = node.create_client(GetCurrentPosition, 'get_current_position')
        while not self.cli.wait_for_service(timeout_sec=1.0):
            self.node.get_logger().info('GetCurrentPosition service not available, waiting again...')
        self.req = GetCurrentPosition.Request()

    def send_request(self):
        self.future = self.cli.call_async(self.req)
        self.future.add_done_callback(self.get_result_callback)        

    def get_result_callback(self, future):
        self.result = future.result()
        self.node.get_logger().info(f'Result: {self.result}')
        if self.shutdown:
            rclpy.shutdown()        
        

class InAirCheckClient():
    def __init__(self, node, shutdown=False, callback_group=None):
        self.node = node
        self.result = None
        self.shutdown = shutdown
        self.cli = node.create_client(InAirCheck, 'in_air_check')
        while not self.cli.wait_for_service(timeout_sec=1.0):
            self.node.get_logger().info('InAirCheck service not available, waiting again...')
        self.req = InAirCheck.Request()

    def send_request(self):
        self.result = None        
        self.future = self.cli.call_async(self.req)
        self.future.add_done_callback(self.get_result_callback)

    def get_result_callback(self, future):
        self.result = future.result()
        self.node.get_logger().info(f'Result: {self.result}')
        if self.shutdown:
            rclpy.shutdown()        


class PatchRclpyIssue1123(ActionClient):

    _lock: RLock = None  # type: ignore

    @property
    def _cpp_client_handle_lock(self) -> RLock:
        if self._lock is None:
            self._lock = RLock()
        return self._lock

    async def execute(self, *args: Any, **kwargs: Any) -> None:
        with self._cpp_client_handle_lock:
            return await super().execute(*args, **kwargs)  # type: ignore

    def send_goal_async(self, *args: Any, **kwargs: Any) -> Future:
        with self._cpp_client_handle_lock:
            return super().send_goal_async(*args, **kwargs)

    def _cancel_goal_async(self, *args: Any, **kwargs: Any) -> Future:
        with self._cpp_client_handle_lock:
            return super()._cancel_goal_async(*args, **kwargs)

    def _get_result_async(self, *args: Any, **kwargs: Any) -> Future:
        with self._cpp_client_handle_lock:
            return super()._get_result_async(*args, **kwargs)

class TrajectoryActionClient(Node):

    def __init__(self, node, shutdown=False, callback_group=None):
        self.node = node
        self.shutdown = shutdown
        self.result = None
        self._action_client = PatchRclpyIssue1123(node, Trajectory, 'trajectory', callback_group=callback_group)
        self.have_goal_response = False
        self.goal_accepted = False

    def send_goal(self, target_pose, duration=60.0, precise_positioning=False):

        goal_msg = Trajectory.Goal()
        goal_msg.target_pose = target_pose
        goal_msg.duration = Duration(seconds=duration, nanoseconds=0).to_msg()
        goal_msg.precise_positioning = precise_positioning
        guuid = uuid.uuid4()
        goal_uuid = UUID(uuid=list(guuid.bytes))
        print(str(guuid))
        print(str(goal_uuid))

        self.node.get_logger().info(f'GOAL UUI: {str(goal_uuid)}')

        self._action_client.wait_for_server()
        self._send_goal_future = self._action_client.send_goal_async(goal_msg, self.feedback_callback, goal_uuid=goal_uuid)
        self._send_goal_future.add_done_callback(self.goal_response_callback)

        return goal_uuid

    def goal_response_callback(self, future):
        goal_handle = future.result()
        if not goal_handle.accepted:
            self.node.get_logger().info('Goal rejected :(')
            self.have_goal_response = True
            return

        self.node.get_logger().info('Goal accepted :)')
        
        self.goal_accepted = True
        self._get_result_future = goal_handle.get_result_async()
        self._get_result_future.add_done_callback(self.get_result_callback)
        self.have_goal_response = True

    def get_result_callback(self, future):
        self.result = future.result().result
        self.node.get_logger().info(f'Result: {self.result.success} - {self.result.message}')
        if self.shutdown:
            rclpy.shutdown()        

    def feedback_callback(self, feedback_msg):
        feedback = feedback_msg.feedback
        self.node.get_logger().info(f'Received feedback {feedback.feedback}')        

class MoveToActionClient(Node):

    def __init__(self, node, shutdown=False, callback_group=None):
        self.node = node
        self.shutdown = shutdown
        self.result = None
        self._action_client = PatchRclpyIssue1123(node, MoveTo, 'move_to', callback_group=callback_group)
        self.have_goal_response = False
        self.goal_accepted = False

    def send_goal(self, target_pose):

        goal_msg = MoveTo.Goal()
        goal_msg.pose = target_pose
        guuid = uuid.uuid4()
        goal_uuid = UUID(uuid=list(guuid.bytes))
        print(str(guuid))
        print(str(goal_uuid))

        self.node.get_logger().info(f'MoveToAction GOAL UUI: {str(goal_uuid)}')

        self._action_client.wait_for_server()
        self.node.get_logger().info(f'MoveToAction send goal: {str(goal_uuid)}')
        self._send_goal_future = self._action_client.send_goal_async(goal_msg, self.feedback_callback, goal_uuid=goal_uuid)
        self._send_goal_future.add_done_callback(self.goal_response_callback)

        return goal_uuid

    def goal_response_callback(self, future):
        goal_handle = future.result()
        if not goal_handle.accepted:
            self.node.get_logger().info('Goal rejected :(')
            self.have_goal_response = True
            return

        self.node.get_logger().info('Goal accepted :)')
        
        self.goal_accepted = True
        self._get_result_future = goal_handle.get_result_async()
        self._get_result_future.add_done_callback(self.get_result_callback)
        self.have_goal_response = True

    def get_result_callback(self, future):
        self.result = future.result().result
        self.node.get_logger().info(f'Result: {self.result.success} - {self.result.message}')
        if self.shutdown:
            rclpy.shutdown()        

    def feedback_callback(self, feedback_msg):
        feedback = feedback_msg.feedback
        self.node.get_logger().info(f'Received feedback {feedback.current_pose}')        

class FlyPathActionClient(Node):

    def __init__(self, node, shutdown=False, callback_group=None):
        self.node = node
        self.shutdown = shutdown
        self.result = None
        self._action_client = PatchRclpyIssue1123(node, FlyPath, 'fly_path', callback_group=callback_group)
        self.have_goal_response = False
        self.goal_accepted = False
        

    def send_goal(self, geopoints, speed):
        self.result = None        
        goal_msg = FlyPath.Goal()
        goal_msg.geopoints = geopoints
        goal_msg.speed = speed

        guuid = uuid.uuid4()
        goal_uuid = UUID(uuid=list(guuid.bytes))
        print(str(guuid))
        print(str(goal_uuid))

        self.node.get_logger().info(f'GOAL UUI: {str(goal_uuid)}')

        self._action_client.wait_for_server()
        self._send_goal_future = self._action_client.send_goal_async(goal_msg, self.feedback_callback, goal_uuid=goal_uuid)
        self._send_goal_future.add_done_callback(self.goal_response_callback)

        return goal_uuid

    def goal_response_callback(self, future):
        goal_handle = future.result()
        if not goal_handle.accepted:
            self.node.get_logger().info('Goal rejected :(')
            self.have_goal_response = True
            return

        self.node.get_logger().info('Goal accepted :)')
        
        self.goal_accepted = True
        self._get_result_future = goal_handle.get_result_async()
        self._get_result_future.add_done_callback(self.get_result_callback)
        self.have_goal_response = True

    def get_result_callback(self, future):
        self.result = future.result().result
        self.abort_flag = self.result.abort_flag
        self.enough_flag = self.result.enough_flag
        self.node.get_logger().info(f'Result: {self.result.success} - {self.result.message} - {self.result}')
        if self.shutdown:
            rclpy.shutdown()        

    def feedback_callback(self, feedback_msg):
        feedback = feedback_msg.feedback
        self.node.get_logger().info(f'Received feedback {feedback.current_point}')        


class LandActionClient():

    def __init__(self, node, shutdown=False, callback_group=None):
        self.node = node
        self._action_client = PatchRclpyIssue1123(node, Land, 'land', callback_group=callback_group)
        self.shutdown = shutdown
        self.result = None
        self.have_goal_response = False
        self.goal_accepted = False
        
    def send_goal(self):
        goal_msg = Land.Goal()
        self._action_client.wait_for_server()
        self._send_goal_future = self._action_client.send_goal_async(goal_msg, self.feedback_callback)
        self._send_goal_future.add_done_callback(self.goal_response_callback)

    def goal_response_callback(self, future):
        goal_handle = future.result()
        if not goal_handle.accepted:
            self.node.get_logger().info('Goal rejected :(')
            self.have_goal_response = True
            return

        self.goal_accepted = True
        self.node.get_logger().info('Goal accepted :)')
        
        self._get_result_future = goal_handle.get_result_async()
        self._get_result_future.add_done_callback(self.get_result_callback)
        self.have_goal_response = True

    def get_result_callback(self, future):
        self.result = future.result().result
        self.node.get_logger().info(f'Result: {self.result.success} - {self.result.message}')
        if self.shutdown:
            rclpy.shutdown()        

    def feedback_callback(self, feedback_msg):
        feedback = feedback_msg.feedback
        self.node.get_logger().info(f'Received feedback {feedback.current_point}')        
        
class TakeOffActionClient():

    def __init__(self, node, shutdown=False, callback_group=None):
        self.node = node
        if callback_group:
            self._action_client = PatchRclpyIssue1123(node, TakeOff, 'take_off', callback_group=callback_group)
        else:
            self._action_client = PatchRclpyIssue1123(node, TakeOff, 'take_off')
        self.shutdown = shutdown
        self.result = None
        self.have_goal_response = False
        self.goal_accepted = False

    def send_goal(self):
        goal_msg = TakeOff.Goal()
        self._action_client.wait_for_server()
        self._send_goal_future = self._action_client.send_goal_async(goal_msg, self.feedback_callback)
        self._send_goal_future.add_done_callback(self.goal_response_callback)

    def goal_response_callback(self, future):
        goal_handle = future.result()
        if not goal_handle.accepted:
            self.node.get_logger().info('Goal rejected :(')
            self.have_goal_response = True
            return

        self.goal_accepted = True
        self.node.get_logger().info('Goal accepted :)')
        
        self._get_result_future = goal_handle.get_result_async()
        self._get_result_future.add_done_callback(self.get_result_callback)
        self.have_goal_response = True

    def get_result_callback(self, future):
        self.result = future.result().result
        self.node.get_logger().info(f'Result: {self.result.success} - {self.result.message}')
        if self.shutdown:
            rclpy.shutdown()        

    def feedback_callback(self, feedback_msg):
        feedback = feedback_msg.feedback
        self.node.get_logger().info(f'Received feedback {feedback.current_point}')
            
