import os
import json
import uuid
import threading

have_pygeodesi=True
try:
    import pygeodesy
    from pygeodesy.ellipsoidalKarney import LatLon
except:
    have_pygeodesi=False

import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup
from rclpy.qos import qos_profile_system_default, qos_profile_sensor_data, QoSProfile, QoSReliabilityPolicy, QoSHistoryPolicy

from lrs_team_msgs.srv import AddTask, SetTaskPriority, SignalTask
from lrs_msgs_tst.msg import TSTNodeStatus
from lrs_srvs_tst.srv import TSTGetExecInfo, TSTSendSignal, TSTSignalTree, TSTSignalUnit
from std_msgs.msg import String
from lrs_msgs_common.srv import SendTopicMsgToUnit, OctreeGetHeight

from lrs_util.jsonutil import json_basic_moveto, json_moveto, json_basic_search_position, json_search_position
from lrs_util.jsonutil import json_basic_move_path, json_move_path, json_basic_search_area, json_search_area
from lrs_util.jsonutil import json_tree_from_task

from lrs_exec.exec_tree import ExecTree
from lrs_srvs_exec.srv import TSTExecute
from ros_kdb_interfaces.srv import GetAltitude

class ExecServer(ExecTree):
    def __init__(self):
        super().__init__("json_api_exec_server")

        self.transport = "mqtt" # "ros", "zenoh", "mqtt", ...
        if os.getenv("TST_REMOTE_TRANSPORT"):
            self.transport = os.getenv("TST_REMOTE_TRANSPORT")

        self.threadlist = []
        
        self.exec_command_sub = self.create_subscription(String, "exec_command", self.exec_command_callback,
                                                         QoSProfile(
                                                             reliability=QoSReliabilityPolicy.RELIABLE,
                                                             history=QoSHistoryPolicy.KEEP_LAST,
                                                             depth=100
                                                         ),
                                                         callback_group=self.group)

        self.tst_command_sub = self.create_subscription(String, "tst_command", self.tst_command_callback,
                                                         QoSProfile(
                                                             reliability=QoSReliabilityPolicy.RELIABLE,
                                                             history=QoSHistoryPolicy.KEEP_LAST,
                                                             depth=100
                                                         ),
                                                        callback_group=self.group)
        
        self.tst_node_status_sub = self.create_subscription(TSTNodeStatus, "tst_node_status", self.tst_node_status_callback,
                                                            QoSProfile(
                                                                reliability=QoSReliabilityPolicy.RELIABLE,
                                                                history=QoSHistoryPolicy.KEEP_LAST,
                                                                depth=100
                                                            ),
                                                            callback_group=self.group)

        self.kdb_get_altitude_client = self.create_client(GetAltitude, "kdb_server/get_altitude", callback_group=self.group)        
        self.octree_get_height_client = self.create_client(OctreeGetHeight, "octree_get_height", callback_group=self.group)        
        self.add_task_client = self.create_client(AddTask, "team/add_task", callback_group=self.group)        
        self.stmtus_client = self.create_client(SendTopicMsgToUnit, "send_topic_msg_to_unit", callback_group=self.group)
        self.send_signal_client = self.create_client(TSTSendSignal, "exec/tstfactory/send_signal", callback_group=self.group)
        
        self.signal_tree_client = self.create_client(TSTSignalTree, "exec/tstfactory/signal_tree", callback_group=self.group)
        self.signal_unit_client = self.create_client(TSTSignalUnit, "exec/tstfactory/signal_unit", callback_group=self.group)

        self.tst_execute_client = self.create_client(TSTExecute, f'tst_execute', callback_group=self.group)        
        
        self.task_status_map = {}
        self.task_to_node_map = {}
        self.agent_uuid = str(uuid.uuid4())

        self.tst_status_map = {}

        geoid_file = "/usr/share/GeographicLib/geoids/egm96-5.pgm"
        self.can_compute_geoid_height = False
        if have_pygeodesi and os.path.isfile(geoid_file):
            self.ginterpolator = pygeodesy.GeoidKarney(geoid_file)
            self.can_compute_geoid_height = True

        self.exec_response_pub = self.create_publisher(String, "exec_response", 5, callback_group=self.group)            
        self.exec_feedback_pub = self.create_publisher(String, "exec_feedback", 5, callback_group=self.group)            

        self.get_exec_info_client = self.create_client(TSTGetExecInfo, "exec/tstfactory/get_exec_info", callback_group=self.group)        
        self.inside_timer = False
        self.timer = self.create_timer(0.5, self.timer_callback, callback_group=self.group)

    def get_exec_info(self, nuuid):
        # print("get_exec_info:", nuuid)
        req = TSTGetExecInfo.Request()
        req.node_uuid = nuuid
        future = self.get_exec_info_client.call_async(req)
        resp = self.wait_future(future)        
        # print("get_exec_info resp:", resp)
        return resp

    def send_pong(self, resp_to_uuid):
        resp = {}
        resp["com-uuid"] = str(uuid.uuid4())
        resp["response"] = "pong"
        resp["response-to"] = resp_to_uuid
        msgstr = json.dumps(resp, sort_keys=True, indent=4, separators=(',', ': '))
        print("SENDING ping RESPONSE:", msgstr)
        msg = String()
        msg.data = msgstr
        self.exec_response_pub.publish(msg)
        

    def add_task(self, team_name, task_obj, priority=50, name="dummyname", task_type="json-tst"):
        try:
            print("ADD_TASK:", team_name, priority, name, task_obj)
            self.add_task_client.wait_for_service()
            print("GOT add_task_client service")
            req = AddTask.Request()
            req.team_name = team_name
            req.task.uuid = str(uuid.uuid4())
            req.task.priority = priority
            req.task.type = task_type
            req.task.task = json.dumps(task_obj, sort_keys=True, indent=4, separators=(',', ': '))
            req.task.name = name
            self.add_task_client.call_async(req)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION: {type(exc)} - {exc}')
    

    def timer_callback(self):
        try:
            # self.get_logger().error(f'ENTER timer_callback')
            if self.inside_timer:
                self.get_logger().error(f'EXIT timer_callback: alredy inside')
                return
            self.inside_timer = True
            task_to_node_map = self.task_to_node_map.copy()
            for key in task_to_node_map:
                node_uuid = task_to_node_map[key]
                if not key in self.task_status_map:
                    self.task_status_map[key] = "initial"
                if self.task_status_map[key] in ["finished", "failed"]:
                    continue
                new_status = "running"
                ##self.get_logger().error(f'BEFORE get_exec_info: {node_uuid}')            
                resp = self.get_exec_info(node_uuid)
                ##self.get_logger().error(f'AFTER get_exec_info: {node_uuid}')                            
                if resp.success:
                    if resp.exec_status.finished:
                        if resp.exec_status.succeeded:
                            new_status = "finished"
                        else:
                            new_status = "failed"
                    else:
                        if resp.exec_status.executing:
                            new_status = "running"
                        else:
                            new_status = "starting"
                if new_status != self.task_status_map[key]:
                    self.send_feedback(self.get_namespace().lstrip("/"), key, new_status)
                    self.task_status_map[key] = new_status
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION timer_callback: {type(exc)} - {exc}')
        self.inside_timer = False
        ##self.get_logger().error(f'EXIT timer_callback')
        
    def send_signal(self, node_uuid, signal):
        try:
            print("send_signal:", node_uuid)
            req = TSTSendSignal.Request()
            req.node_uuid = node_uuid
            req.signal = signal
            future = self.send_signal_client.call_async(req)
            resp = self.wait_future(future)        
            print("send_signal_client resp:", resp)
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_signal: {type(exc)} - {exc}')

    def signal_tree(self, tree_uuid, signal):
        try:
            self.get_logger().info(f'signal_tree: {tree_uuid} - {signal}')
            req = TSTSignalTree.Request()
            req.node_uuid = tree_uuid
            req.signal = signal
            future = self.signal_tree_client.call_async(req)
            resp = self.wait_future(future)        
            self.get_logger().info(f'signal_tree_client resp: {resp}')
            return resp
        except Exception as exc:
            print("EXCEPTION signal_tree", type(exc))
            print(exc)

    def signal_unit(self, unit, signal):
        try:
            self.get_logger().info(f'signal_unit: {unit} - {signal}')
            req = TSTSignalUnit.Request()
            req.unit = unit
            req.signal = signal
            future = self.signal_unit_client.call_async(req)
            resp = self.wait_future(future)        
            print("signal_unit_client resp:", resp)
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION signal_unit: {type(exc)} - {exc}')            
        
    def send_topic_msg_to_unit(self, unit, topic, msgstr):
        try:
            ### print("send_topic_msg_to_unit:", unit, topic, msgstr)
            req = SendTopicMsgToUnit.Request()
            req.unit = unit
            req.topic = topic
            req.msg = msgstr
            future = self.stmtus_client.call_async(req)
            resp = self.wait_future(future)        
            ### print("send_topic_msg_to_unit resp:", resp)
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_topic_msg_to_unit: {type(exc)} - {exc}')                        

    def send_signal_response(self, unit, response_uuid, response, topic="exec/response", fail_reason=""):
        resp = None
        try:
            msg = {}
            msg["com-uuid"] = str(uuid.uuid4())
            msg["response-to"] = response_uuid
            msg["response"] = response
            msgstr = json.dumps(msg, sort_keys=True, indent=4, separators=(',', ': '))
            # print(msgstr)
            resp = self.send_topic_msg_to_unit(unit, topic, msgstr)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_signal_response: {type(exc)} - {exc}')                        
        return resp

    def send_response(self, unit, response_uuid, task_uuid, response, fail_reason=""):
        resp = None
        self.get_logger().info(f"send_response: {unit} {task_uuid} {response}")
        
        try:
            msg = {}
            msg["com-uuid"] = str(uuid.uuid4())
            msg["response-to"] = response_uuid
            msg["agent-uuid"] = self.agent_uuid
            msg["task-uuid"] = task_uuid
            msg["response"] = response
            msg["fail-reason"] = fail_reason
            msgstr = json.dumps(msg, sort_keys=True, indent=4, separators=(',', ': '))
            print("SENDING TASK RESPONSE:", msgstr)

            if self.transport == "zenoh":
                msg = String()
                msg.data = msgstr
                self.exec_response_pub.publish(msg)

            if self.transport == "mqtt":
                topic = "exec/response"
                print("send_response:", unit, topic, msgstr)
                resp = self.send_topic_msg_to_unit(unit, topic, msgstr)
                
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_response: {type(exc)} - {exc}')                        
        return resp
    
    def send_tst_response(self, response_uuid, root_uuid, response, fail_reason=""):
        resp = None        
        try:
            msg = {}
            msg["com-uuid"] = str(uuid.uuid4())
            msg["response-to"] = response_uuid
            msg["root-uuid"] = root_uuid
            msg["response"] = response
            msg["fail-reason"] = fail_reason
            msgstr = json.dumps(msg, sort_keys=True, indent=4, separators=(',', ': '))
            topic = "tst/response"
            print("send_tst_response:", topic, msgstr)
            resp = self.send_topic_msg_to_unit(self.get_namespace(), topic, msgstr)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_tst_response: {type(exc)} - {exc}')                                    
        return resp
    
    def send_feedback(self, unit, task_uuid, status):
        resp = None        
        try:
            msg = {}
            msg["com-uuid"] = str(uuid.uuid4())
            msg["agent-uuid"] = self.agent_uuid
            msg["task-uuid"] = task_uuid
            msg["status"] = status
            msgstr = json.dumps(msg, sort_keys=True, indent=4, separators=(',', ': '))
            print("Send TASK FEEDBACK:", msgstr)

            if self.transport == "zenoh":
                msg = String()
                msg.data = msgstr
                self.exec_feedback_pub.publish(msg)

            if self.transport == "mqtt":            
                topic = "exec/feedback"
                ### print("Send feedback on topic:", topic)
                self.get_logger().info(f'TASK FEEDBACK: {topic} - {msgstr}')
                resp = self.send_topic_msg_to_unit(unit, topic, msgstr)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_feedback: {type(exc)} - {exc}')
        return resp

    def send_tst_feedback(self, unit, node_uuid, status, fail_reason="", root_flag=False):
        resp = None        
        try:
            ## self.get_logger().info(f'send_tst_feedback {unit} - {node_uuid} - {status} - {fail_reason} - {root_flag}')            
            msg = {}
            msg["com-uuid"] = str(uuid.uuid4())
            msg["node-uuid"] = node_uuid
            msg["status"] = status
            msg["fail-reason"] = fail_reason
            msg["root-flag"] = root_flag
            msgstr = json.dumps(msg, sort_keys=True, indent=4, separators=(',', ': '))
            # print(msgstr)
            topic = "tst/feedback"
            ### print("Send tst feedback on topic:", topic)
            self.get_logger().info(f'TST FEEDBACK: {topic} - {msgstr}')                                    
            resp = self.send_topic_msg_to_unit(self.get_namespace().lstrip("/"), topic, msgstr)            
            ##oldish resp = self.send_topic_msg_to_unit(unit, topic, msgstr)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_tst_feedback: {type(exc)} - {exc}')            
        return resp

    def send_sensor(self, unit, sensor, jobj):
        resp = None        
        try:
            msgstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
            # print(msgstr)
            topic = f"sensor/{sensor}"
            print("Send sensor on topic:", topic)
            resp = self.send_topic_msg_to_unit(unit, topic, msgstr)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_sensor: {type(exc)} - {exc}')
        return resp
    
    def get_lat_lon_alt(self, jobj):
        try:
            lat = jobj["latitude"]
            lon = jobj["longitude"]
            alt = jobj["altitude"]
            return (lat, lon, alt)
        except Exception as exc: 
            self.get_logger().error(f'EXCEPTION get_lat_lon_alt: {type(exc)} - {exc}')                       

        
    def do_command(self, execunit, task, root_uuid, task_uuid=""):
        try:
            if self.get_namespace() == execunit:
                if task["name"] == "get-geoid-height":
                    lat = task["params"]["latitude"]
                    lon = task["params"]["longitude"]
                    geoid_height = -10000.0
                    valid = False
                    if self.can_compute_geoid_height:
                        valid = True
                        single_position=LatLon(lat, lon)
                        geoid_height = self.ginterpolator(single_position)
                    print("geoid_height:", geoid_height)
                    jobj = {
                        "type": "geoid-height",
                        "task_uuid": task_uuid,
                        "longitude": lon,
                        "latitude": lat,
                        "geoid_height": geoid_height,
                        "valid": valid
                    }
                    self.send_sensor(execunit, "geoid_height", jobj)
                    return (True, "", True)
                    
                if task["name"] == "get-octomap-height":
                    lat = task["params"]["latitude"]
                    lon = task["params"]["longitude"]
                    self.get_logger().info(f"get-octomap-height lon lat: {lon} {lat}")
                    (x, y, z) = self.coordtrans.wgs84_to_world(lon, lat)
                    self.get_logger().info(f"get-octomap-height: {x} {y} {z}")
                    octomap_height = -10000.0
                    req = OctreeGetHeight.Request()
                    req.x = x
                    req.y = y
                    future = self.octree_get_height_client.call_async(req)
                    resp = self.wait_future(future)        
                    valid = resp.success
                    if valid:
                        octomap_height = resp.height
                    self.get_logger().info(f"octomap-height: {octomap_height} - {valid}")
                    # compute it
                    jobj = {
                        "type": "octomap-height",
                        "task_uuid": task_uuid,
                        "longitude": lon,
                        "latitude": lat,
                        "octomap_height": octomap_height,
                        "valid": valid
                    }
                    self.send_sensor(execunit, "octomap_height", jobj)
                    return (True, "", True)
                    
                if task["name"] == "get-ellipsoid-height":
                    lat = task["params"]["latitude"]
                    lon = task["params"]["longitude"]
                    srid = 4326
                    self.get_logger().info(f"get-ellipsoid-height lon lat: {lon} {lat} {srid}")
                    (x, y, z) = self.coordtrans.wgs84_to_world(lon, lat)
                    self.get_logger().info(f"get-ellipsoid-height: {x} {y} {z}")
                    octomap_height = -10000.0
                    req = GetAltitude.Request()
                    req.latitude = lat
                    req.longitude = lon
                    req.srid = srid
                    future = self.kdb_get_altitude_client.call_async(req)
                    resp = self.wait_future(future)        
                    valid = True
                    if valid:
                        ellipsoid_height = resp.altitude
                    self.get_logger().info(f"ellipsoid-height: {ellipsoid_height} - {valid}")
                    jobj = {
                        "type": "ellipsoid-height",
                        "task_uuid": task_uuid,
                        "longitude": lon,
                        "latitude": lat,
                        "ellipsoid_height": ellipsoid_height,
                        "valid": valid
                    }
                    self.send_sensor(execunit, "ellipsoid_height", jobj)
                    return (True, "", True)
                    
                if task["name"] == "get-origin":
                    (lon, lat, alt) = self.coordtrans.world_to_wgs84(0, 0, 0)
                    print("origin:", lon, lat, alt)
                    jobj = {
                        "type": "origin",
                        "task_uuid": task_uuid,
                        "longitude": lon,
                        "latitude": lat,
                        "altitude": alt
                    }
                    self.send_sensor(execunit, "origin", jobj)
                    return (True, "", True)
                    
                if task["name"] == "move-toxxx":
                    params = task["params"]
                    (lat, lon, alt) = self.get_lat_lon_alt(params["waypoint"])
                    speed = params["speed"]
                    camera_data_recording = False
                    if "camera-data-recording" in params:
                        camera_data_recording = params["camera-data-recording"]
                    jobj = json_moveto(self.get_namespace(), lat, lon, alt, speed, seq_node_uuid=root_uuid, camera_data_recording=camera_data_recording)
                    ## print("JSON do_command:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
                    resp = self.new_handle_tree(jobj)
                    # (res, reason) = self.new_handle_tree(jobj)
                    ## print("HANDLE TREE RES:", resp)
                    return (resp.success, resp.reason, False)
                else:
                    print("JSON do_command task:", json.dumps(task, sort_keys=True, indent=4, separators=(',', ': ')))                    
                    jobj = json_tree_from_task(self.get_namespace(), task["name"], task["params"], seq_node_uuid=root_uuid)
                    print("JSON do_command:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
                    task_name = task["name"]
                    print("TASK_NAME:", task_name)
                    queue_flag = False
                    if task_name in ["move-to", "move-path", "search-area", "look-at-position"] and self.get_namespace() == "/op0":
                        if task_name == "move-to":
                            team_name = "examine"
                            jobj["common_params"]["execunit"] = ""
                            jobj["children"][0]["common_params"]["execunit"] = ""
                            print("JSON-TASK queue_task:", json.dumps(task, sort_keys=True, indent=4, separators=(',', ': ')))

                            self.add_task(team_name, task, name="move-to" ,task_type="json-task")
                            queue_flag = True
                        if task_name == "search-area":
                            team_name = "searchair"
                            jobj["common_params"]["execunit"] = ""
                            jobj["children"][0]["common_params"]["execunit"] = ""
                            print("JSON queue_task:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
                            self.add_task(team_name, jobj, name="search-area")
                            queue_flag = True
                        if task_name == "move-path":
                            team_name = "searchsurface"
                            jobj["common_params"]["execunit"] = ""
                            jobj["children"][0]["common_params"]["execunit"] = ""
                            search_area = True
                            if search_area:
                                #jobj["children"][0]["params"]["name"] = "search-area"
                                #jobj["children"][0]["params"]["area"] = jobj["children"][0]["params"]["waypoints"]
                                #del jobj["children"][0]["params"]["waypoints"]
                                #jobj["children"][0]["params"]["target-type"] = "person"
                                #jobj["children"][0]["params"]["area-type"] = "water"
                                #jobj["children"][0]["params"]["target-size"] = 2
                                task["name"] = "search-area"
                                task["params"]["area"] = task["params"]["waypoints"]
                                del task["params"]["waypoints"]
                                task["params"]["target-type"] = "person"
                                task["params"]["area-type"] = "water"
                                task["params"]["target-size"] = 2
                            print("JSON queue_task:", json.dumps(task, sort_keys=True, indent=4, separators=(',', ': ')))
                            self.add_task(team_name, task, name="search-area" ,task_type="json-task")
                            queue_flag = True
                        if task_name == "look-at-position":
                            team_name = "sup"
                            jobj["common_params"]["execunit"] = ""
                            jobj["children"][0]["common_params"]["execunit"] = ""
                            print("JSON queue_task:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
                            self.add_task(team_name, jobj, name="look-at-position")
                            queue_flag = True
                            
                        resp = TSTExecute.Response()
                        resp.success = True
                    else:
                        resp = self.new_handle_tree(jobj)
                    # (res, reason) = self.new_handle_tree(jobj)
                    print("HANDLE TREE RES:", resp)
                    return (resp.success, resp.reason, queue_flag)                    
                
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION do_command: {type(exc)} - {exc}')                       
            return (False, f'do_command exception: {exc}')
        
    def do_tree(self, treeobj):
        try:
            print("JSON do_tree:", json.dumps(treeobj, sort_keys=True, indent=4, separators=(',', ': ')))
            resp = self.new_handle_tree(treeobj)
            if resp:
                res = resp.success
                reason = resp.reason
            else:
                res = False
                reason = "Call in new_handle_tree failed"
            print("HANDLE TREE RES:", res, reason)
            return (res, reason)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION do_tree: {type(exc)} - {exc}')                       
            return (False, f'do_tree exception: {exc}')
        
    def handle_task_command(self, jobj):
        try:
            # print(jobj)
            com_uuid = jobj["com-uuid"]
            task_uuid = str(uuid.uuid4())
            if "task-uuid" in jobj:
                task_uuid = jobj["task-uuid"]
            command = jobj["command"]
            unit = self.get_namespace().lstrip("/")
            print("UNIT:", unit)

            if command == "ping":            
                print("COMMAND IS ping")
                self.send_pong(com_uuid)
                
            if command == "signal-task":
                signal = jobj["signal"]
                signal_task_uuid = jobj["task-uuid"]
                if signal_task_uuid not in self.task_to_node_map:
                    self.get_logger().error(f"handle_task_command Could not find task uuid: {signal_task_uuid}")
                    return
                tree_uuid = self.task_to_node_map[signal_task_uuid]
                self.get_logger().info(f"Signal Task: {signal_task_uuid} - Tree: {tree_uuid}")

                response = "ok"
                resp = None
                if signal in ["$abort", "$enough", "$pause", "$continue"]:
                    resp = self.signal_tree(tree_uuid, signal)
                if not resp or not resp.success:
                    response = "fail"
                self.send_signal_response(unit, com_uuid, response)

            if command == "start-task":
                print("COMMAND IS start-task")
                task = jobj["task"]
                execunit = unit
                root_uuid = str(uuid.uuid4())                
                if "execution-unit" in jobj:
                    execunit = jobj["execution-unit"]
                if execunit != unit:
                    print("UNIT and EXECUNIT must be the same", unit, execunit)
                    return
                execunit = f'/{unit}'
                (res, failcause, queue_flag) = self.do_command(execunit, task, root_uuid, task_uuid=task_uuid)
                print("RES do_command:", res, "-", failcause, "-", queue_flag, "ENDLINE")
                if res:
                    if not queue_flag:
                        print("Updating task_to_node_map:", task_uuid, root_uuid)
                        self.task_to_node_map[task_uuid] = root_uuid
                    self.send_response(unit, com_uuid, task_uuid, "running")
                    if queue_flag:
                        print("SEND FINISHED")
                        self.send_feedback(self.get_namespace().lstrip("/"), task_uuid, "finished")
                else:
                    self.send_response(unit, com_uuid, task_uuid, "failed", failcause)
                    
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION handle_task_command: {type(exc)} - {exc}')                                   
        
    def exec_command_callback(self, msg):
        try:
            self.get_logger().info(f"exec_command_callback: {msg}")            
            msgobj = json.loads(msg.data)
            print("JSON:", json.dumps(msgobj, sort_keys=True, indent=4, separators=(',', ': ')))
            self.handle_task_command(msgobj)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION exec_command_callback: {type(exc)} - {exc}')                                               
    
    def handle_tst_command(self, jobj):
        try:
            self.get_logger().info(f'handle_tst_command: {jobj}')

            com_uuid = jobj["com-uuid"]
            command = jobj["command"]
            unit = self.get_namespace().lstrip("/")
            
            if command == "signal-tst-node":
                signal = jobj["signal"]
                node_uuid = jobj["node-uuid"]
                response = "ok"
                resp = self.send_signal(node_uuid, signal)
                if not resp or not resp.success:
                    response = "fail"
                self.send_signal_response(unit, com_uuid, response, topic="tst/response")                
                
            if command == "signal-tst-tree":
                signal = jobj["signal"]
                node_uuid = jobj["node-uuid"]
                response = "ok"
                resp = None
                if signal in ["$abort", "$enough", "$pause", "$continue"]:
                    resp = self.signal_tree(node_uuid, signal)
                if not resp:
                    response = "fail"
                self.send_signal_response(unit, com_uuid, response, topic="tst/response")                
                
            if command == "signal-unit":
                signal = jobj["signal"]
                unit = jobj["unit"]
                response = "ok"
                resp = None
                if signal in ["$abort", "$enough", "$pause", "$continue"]:                
                    resp = self.signal_unit(unit, signal)
                if not resp:
                    response = "fail"
                self.send_signal_response(unit, com_uuid, response, topic="tst/response")                
                
            if command == "start-tst":
                print("COMMAND IS start-tst")
                receiver = jobj["receiver"]
                treeobj = jobj["tst"]
                comparams = treeobj["common_params"]
                if not "node-uuid" in comparams or comparams["node-uuid"] == "":
                    comparams["node-uuid"] = str(uuid.uuid4())
                root_uuid = comparams["node-uuid"]
                # treestr = json.dumps(treeobj, sort_keys=True, indent=4, separators=(',', ': '))
                # send_tst_response(unit, com_uuid, root_uuid, "started")
                #start_tst(treestr)
                (res, failcause) = self.do_tree(treeobj)
                print("RES do_command:", res, failcause)
                if res:
                    # self.task_to_node_map[task_uuid] = root_uuid
                    self.send_tst_response(com_uuid, root_uuid, "started")
                else:
                    self.send_tst_response(com_uuid, root_uuid, "failed", fail_reason=failcause)
                    
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION handle_tst_command: {type(exc)} - {exc}')                                                           
    
        
    def tst_command_callback(self, msg):
        try:
            self.get_logger().info(f'tst_command_callback: {msg}')
            msgobj = json.loads(msg.data)
            print("JSON:", json.dumps(msgobj, sort_keys=True, indent=4, separators=(',', ': ')))
            self.handle_tst_command(msgobj)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION tst_command_callback: {type(exc)} - {exc}')

    def tst_node_status_callback(self, msg):
        try:
            ###self.get_logger().info(f'tst_node_status_callback 1: {msg}')
            ns = self.get_namespace()
            #        resp = get_node(ns + "/exec", msg.id)
            #        tni = resp.info
            #        if msg.ns == ns and tni.execution_ns == ns and not tni.root_uuid in internal_root_uuid_map:
            ## self.get_logger().error(f'tst_node_status_callback NS check: {ns} - {msg.ns}')
            if msg.ns == ns:
                ### self.get_logger().info(f'tst_node_status_callback 2: {msg}')
                status = "not-active"
                fail_reason = ""
                # The node is on this agent (it must be since the sub is on this agent, so test probably unneccessary)
                if msg.exec_status.finished:
                    if msg.exec_status.succeeded:
                        status = "succeeded"
                    else:
                        status = "failed"
                        fail_reason = msg.exec_status.fail_reason
                else:
                    if msg.exec_status.executing:
                        status = "running"
                    else:
                        if msg.exec_status.waiting:
                            status = "waiting"
                        else:
                            if msg.exec_status.active:
                                status = "active"
                            if msg.exec_status.revoked:
                                status = "revoked"
                            if msg.exec_status.aborted:
                                status = "aborted"
                            if msg.exec_status.paused:
                                status = "paused"
                root_flag = False
                if msg.id == msg.root_id:
                    root_flag = True
                ###self.get_logger().info(f'tst_node_status_callback STATUS - root flag: {status} - {root_flag}')
                #if msg.node_uuid in self.tst_status_map:
                #    self.get_logger().error(f'tst_node_status_callback map: {msg.ns} - {msg.node_uuid} - {self.tst_status_map[msg.node_uuid]} - {status}')                    

                if (not msg.node_uuid in self.tst_status_map) or (self.tst_status_map[msg.node_uuid] != status):
                    ###self.get_logger().info(f'tst_node_status_callback FEEDBACK {ns}: {status}')
                    self.send_tst_feedback(ns.lstrip("/"), msg.node_uuid, status, fail_reason=fail_reason, root_flag=root_flag)
                    self.tst_status_map[msg.node_uuid] = status
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION tst_node_status_callback: {type(exc)} - {exc}')
        ###self.get_logger().info(f'EXIT tst_node_status_callback 3: {msg}')
        
def main(args=None):
    print('Hi from JSON Exec Server.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=32)
    node = ExecServer()
    executor.add_node(node)

    print("Spinning exec server")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass
    
    # rclpy.spin(node)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    #node.destroy_node()
    #rclpy.shutdown()
    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
        
