import time
import os
import json

import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup

from lrs_srvs_wdb.srv import WDBWorldToGeo
from lrs_srvs_tst.srv import TSTCreateTree, TSTDistributeTreeFromUuid, TSTSetStartExecutorFlagFromUuid
from lrs_msgs_common.srv import GetResponse
from lrs_msgs_common.msg import DisplayRequest
from lrs_srvs_tst.srv import TSTExpandTree, TSTAddConstraintsToTree, TSTSignalTree
from lrs_srvs_exec.srv import TSTExecute

from lrs_util.constraints import TreeConstraints
from lrs_util.coordtrans import CoordTrans

class ExecTree(Node):

    def __init__(self, node_name):
        super().__init__(node_name)
        self.group = ReentrantCallbackGroup()

        self.coordtrans = CoordTrans(location=os.environ.get("LRS_LOCATION"))
        self.declare_parameter("external", False)
        self.external = self.get_parameter('external').get_parameter_value().bool_value
        external_prefix = ""
        if self.external:
            external_prefix = "/ext"

        self.add_constraints_client = self.create_client(TSTAddConstraintsToTree, "tst_add_constraints_to_tree", callback_group=self.group)
        self.expand_tree_client = self.create_client(TSTExpandTree, "tst_expand_tree", callback_group=self.group)
        self.world_to_geo_client = self.create_client(WDBWorldToGeo, "wdb/world_to_geo", callback_group=self.group)
        self.create_tree_client = self.create_client(TSTCreateTree, "exec/tstfactory/create_tree", callback_group=self.group)
        self.distribute_tree_client = self.create_client(TSTDistributeTreeFromUuid, "exec/tstfactory/distribute_tree_from_uuid", callback_group=self.group)
        self.get_response_client = self.create_client(GetResponse, "exec/tstfactory/get_response", callback_group=self.group)
        self.start_executor_client = self.create_client(TSTSetStartExecutorFlagFromUuid, "exec/tstfactory/set_start_executor_from_uuid", callback_group=self.group)

        self.signal_tree_client = self.create_client(TSTSignalTree, '/op0/exec/tstfactory/signal_tree', callback_group=self.group)        

        self.disp_pub = self.create_publisher(DisplayRequest, f"{external_prefix}/tst_display_request", 10)

        self.display = True

        self.delayed_obj = None

    def signal_tree(self, tree_uuid, signal):
        try:
            print("signal_tree:", tree_uuid, signal)
            req = TSTSignalTree.Request()
            req.node_uuid = tree_uuid
            req.signal = signal
            future = self.signal_tree_client.call_async(req)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION signal_tree: {type(exc)} - {exc}')

    def add_delayed_tree(self, tree, delay=0.5):
        self.delayed_obj = tree
        self.delay_timer = self.create_timer(delay, self.do_delayed_command, callback_group=self.group)

    def do_delayed_command(self):
        self.delay_timer.cancel()
        resp = self.new_handle_tree(self.delayed_obj)
        print(f'do_delayed_command new_handle_tree resp: {resp}')
        
    def wait_future(self, future, print_waiting=False):
        while True:
            if future.done():
                break
        resp = future.result()
        return resp

    def expand_tree(self, jobj):
        jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
        self.get_logger().error(f'expand_tree: {jobjstr}')
        try:
            req = TSTExpandTree.Request()
            req.input_json = jobjstr
            future = self.expand_tree_client.call_async(req)
            resp = self.wait_future(future, print_waiting=True)
            self.get_logger().error(f'expand_tree resp: {resp}')
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION expand_tree: {type(exc)} - {exc}')
    
    def add_constraints(self, jobj):
        jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
        print("add_constraints:", jobjstr)
        try:
            req = TSTAddConstraintsToTree.Request()
            req.input_json = jobjstr
            future = self.add_constraints_client.call_async(req)
            resp = self.wait_future(future, print_waiting=True)
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION add_constraints: {type(exc)} - {exc}')
    
    def world_to_geo(self, x, y, z):
        print("world_to_geo:", x, y, z)
        try:
            req = WDBWorldToGeo.Request()
            req.x = float(x)
            req.y = float(y)
            req.z = float(z)
            future = self.world_to_geo_client.call_async(req)
            resp = self.wait_future(future, print_waiting=True)
            return resp
        except Exception as exc:
            print("world_to_geo",  type(exc))
            print(exc)
        return None

    def world_to_geo_path(self, path):
        try:
            geopath = []
            for el in path:
                # resp = self.world_to_geo(el[0], el[1], el[2])
                (lon, lat, alt) = self.coordtrans.world_to_wgs84(el[0], el[1], el[2])
                geopath.append({"altitude": alt,
                                "latitude": lat,
                                "longitude": lon,
                                "rostype": "GeoPoint"})
            return geopath
        except Exception as exc:
            print("world_to_geo_path",  type(exc))
            print(exc)
        return None
    
    def create_tree(self, jobjstr):
        try:
            req = TSTCreateTree.Request()
            req.json = jobjstr
            req.free_id = 1
            req.approve_flag = True
            future = self.create_tree_client.call_async(req)
            resp = self.wait_future(future, print_waiting=True)
            return resp
        except Exception as exc:
            print("create_tree",  type(exc))
            print(exc)
        return None

    def distribute_tree(self, uuidstr, units=[]):
        try:
            req = TSTDistributeTreeFromUuid.Request()
            req.uuid = uuidstr
            req.units = units
            future = self.distribute_tree_client.call_async(req)
            resp = self.wait_future(future, print_waiting=True)
            return resp
        except Exception as exc:
            print("distribute_tree",  type(exc))
            print(exc)
        return None

    def get_response(self, uuidstr):
        try:
            req = GetResponse.Request()
            req.cuuid = uuidstr
            future = self.get_response_client.call_async(req)
            resp = self.wait_future(future, print_waiting=True)
            return resp
        except Exception as exc:
            print("get_response",  type(exc))
            print(exc)
        return None

    def start_executor(self, uuidstr):
        try:
            req = TSTSetStartExecutorFlagFromUuid.Request()
            req.uuid = uuidstr
            req.val = True
            future = self.start_executor_client.call_async(req)
            resp = self.wait_future(future, print_waiting=True)
            return resp
        except Exception as exc:
            print("start_executor",  type(exc))
            print(exc)
        return None

    

    def xyz_to_latlonalt(self, x, y, z):
        (lat, lon, alt) = (0, 0, 0)
        try:
            ###resp = self.world_to_geo(x, y, z)
            # print("RESP:", resp)
            (lon, lat, alt) = self.coordtrans.world_to_wgs84(x, y, z)
        except Exception as exc:
            print("xyz_to_latlonalt", type(exc))
            print(exc)
        return (lat, lon, alt)

    def wait_for_response(self, cuuid, waittime=10):
        count = 0
        delta = 0.1
        while count < waittime/delta:
            try:
                resp = self.get_response(cuuid)
                if resp and resp.success:
                    return (True, "")                
                time.sleep(delta)
                count += 1
            except rospy.ServiceException as exc:
                print("wait_for_response", type(exc))
                print(exc)
                print("Cannot get response")
                return (False, "Cannot get response")
        # print("wait_for_response timeout:", waittime, resp)
        return (False, "Timeout")

    def new_handle_tree(self, jobj, channel=0):
        try:
            jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
            self.get_logger().info(f'******************new_handle_tree: {jobjstr}')
            ###self.get_logger().info(f'new_handle_tree: {jobjstr}')            
            self.get_logger().info(f'new_handle_tree: wait_for_service tst_execute')
            self.tst_execute_client.wait_for_service()
            self.get_logger().info(f'new_handle_tree: GOT service tst_execute')            
            req = TSTExecute.Request()
            req.json_tst = jobjstr
            req.channel = channel
            future = self.tst_execute_client.call_async(req)
            ## self.get_logger().info(f'wait on future: {req}')                        
            resp = self.wait_future(future, print_waiting=True)
            ## self.get_logger().info(f'new_handle_tree resp: {resp}')            
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION new_handle_tree: {type(exc)} - {exc}')
            
    def handle_tree(self, jobj, exectree=True, add_constraints=False, expand=True, distribute=True, displaytype="exec"):
        try:
            self.get_logger().info(f'handle_tree add_constraints: {add_constraints} expand: {expand}')

            if expand:
                self.get_logger().info("HANDLE TREE EXPAND")
                # expand tree
                resp = self.expand_tree(jobj)
                self.get_logger().info(f"HANDLE TREE EXPAND response: {resp}")
                if resp and resp.success:
                    jobj = json.loads(resp.expanded_json)
                else:
                    self.get_logger().error("HANDLE TREE EXPAND FAILED")
                self.get_logger().info("HANDLE TREE EXPAND FINISHED")
                
            if add_constraints:
                # get constraints
                resp = self.add_constraints(jobj)
                if resp and resp.success:
                    jobj = json.loads(resp.output_json)

                # check constraints

                tc = TreeConstraints(jobj=jobj)
                res = tc.check_constraints_isolated()
                
                if not res:
                    print("CONSTRAINT CHECK FAILED")
                    return (False, "CONSTRAINT CHECK FAILED")
            
            jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
            ## print("JOBJSTR:", jobjstr)

            ## print("create_tree")
            resp = self.create_tree(jobjstr)
            ## print("create_tree RESP:", resp)
            
            if resp.success:
                rootid = resp.id
                free_id = resp.free_id
                root_uuid = resp.global_tree_uuid
            else:
                return (False, resp.reason)

            if distribute:
                resp = self.distribute_tree(root_uuid)
                ## print("distribute_tree RESP:", resp)
                (res, failreason) = self.wait_for_response(resp.com_uuid)
                if not res:
                    return (res, failreason)

            if self.display:
                msg = DisplayRequest()
                msg.title="tst_command"
                msg.type=displaytype
                msg.root_uuid=root_uuid
                ## print("DISPLAY PUB:", msg)
                # time.sleep(0.5) # seems not to be needed... was needed in ros1
                self.disp_pub.publish(msg)

            if exectree:
                # possibly add set start time here
                resp = self.start_executor(root_uuid)
                if not resp.success:
                    return (False, "Setting the start executor flag in the root node failed")
                if resp.com_uuid:
                    (res, failreason) = self.wait_for_response(resp.com_uuid)
                    if not res:
                        return (res, failreason + " - set_start_executor failed in recv response")
                return (True, "")

            if self.display:
                return (True, "DISPLAY")

            if exectree:
                print("wait_for_node_finished_from_uuid")
                #if not wait_for_node_finished_from_uuid (nsexec, root_uuid):
                #res = False
                #failcause = "Execution failed"

            return (True, "")
        
        except Exception as exc:
            self.get_logger().error(f'handle_tree: {type(exc)} - {exc}')
            return (False, f'Handle tree exception: {exc}')
    
