import threading
import time
import json
import rclpy
from rclpy.node import Node
from timeit import default_timer as timer

from rclpy.callback_groups import ReentrantCallbackGroup

from lrs_exec.parameters import Parameters

# Needed for Task Execution Level and TST Execution Level
from lrs_srvs_exec.srv import TSTCreateExecutor, TSTDestroyExecutor, TSTSignalExecutor, TSTStartExecutor
from lrs_srvs_exec.srv import TSTExecutorJsonExpand, TSTExecutorJsonConstraints
from lrs_srvs_tst.srv import TSTGetNode

# Needed for Delegation Level
from lrs_srvs_exec.srv import TSTExecutorCheckFromTni, TSTExecutorExpand
from lrs_srvs_exec.srv import TSTExecutorIsDelegationExpandable, TSTExecutorGetConstraintsFromTni
from lrs_srvs_wdb.srv import WDBAddCapability

# Needed for Team Tasks Queue Level
from lrs_srvs_exec.srv import TSTExecutorCost

from lrs_srvs_tst.srv import TSTSetStartExecutorFlag

class ExecutorFactory(Node):

    def add_serv(self, type, srvname, callback):
        srv = self.create_service(type, f'tst_executor/{self.ntype}/{srvname}', callback, callback_group=self.group)
        self.srvs.append(srv)

    def __init__(self, type, expandable=False, handle=None):
        ntype = type.replace("-", "_")
        #print("ExecutorFactory NTYPE:", ntype)
        super().__init__(ntype)
        #print("ExecutorFactory NAMESPACE:", self.get_namespace())
        self.myhandle = handle
        self.group = ReentrantCallbackGroup()
        self.verbose = False
        self.expandable = expandable
        self.type = type
        self.ntype = ntype
        self.executor_map = {}
        self.uuid_executor_map = {}
        self.threads = []
        self.futures = []
        self.futures_desc = []
        timer_period = 0.1  # seconds
        self.timer = self.create_timer(timer_period, self.timer_callback)
        self.srvs = []
        self.executors_lock = threading.Lock()

        ### Services for Task and TST Execution (or more precice execution of a TST node)

        print(f'Create services with prefix: tst_executor/{self.ntype}')

        # Task Execution Level services
        self.add_serv(TSTCreateExecutor, "create_executor", self.create_executor)  
        self.add_serv(TSTDestroyExecutor, "destroy_executor", self.destroy_executor)  
        self.add_serv(TSTStartExecutor, "start_executor", self.start_executor)  
        self.add_serv(TSTSignalExecutor, "signal_executor", self.signal_executor)
        self.add_serv(TSTExecutorJsonExpand, "executor_json_expand", self.executor_json_expand)
        self.add_serv(TSTExecutorJsonConstraints, "executor_json_constraints", self.executor_json_constraints)

        self.set_start_executor_flag_client = self.create_client(TSTSetStartExecutorFlag, "exec/tstfactory/set_start_executor", callback_group=self.group)

        
        # Task Execution Level clients
        self.get_node_client = self.create_client(TSTGetNode, "exec/tstfactory/get_node", callback_group=self.group)
        

        # print("TASK LEVEL CREATED")

        # Delegation Level services
        self.add_serv(TSTExecutorCheckFromTni, "executor_check", self.executor_check)  
        self.add_serv(TSTExecutorExpand, "executor_expand", self.executor_expand)  
        self.add_serv(TSTExecutorIsDelegationExpandable, "executor_is_delegation_expandable", self.is_delegation_expandable)  
        self.add_serv(TSTExecutorGetConstraintsFromTni, "executor_get_constraints", self.get_constraints)

        # print("DELEGATION LEVEL CREATED")        

        # Delegation Level clients
        self.add_capability_client = self.create_client(WDBAddCapability, "wdb/add_capability", callback_group=self.group)

        # Team Task Queue Level Services

        self.add_serv(TSTExecutorCost, "executor_cost", self.executor_cost)  
        
        # Delegation Level stuff
        self.delegation_expandable_flag = False
        # self.get_logger().error(f'executor_factory: {ntype} - before register_capability') # register is fast
        self.register_capability()
        # self.get_logger().error(f'executor_factory: {ntype} - after register_capability')        

    def lprint(self, *val):
        if self.verbose:
            print(val)

    def set_delegation_expandable(self, val):
        self.delegation_expandable_flag = val

    def remove_finished_executors(self):
        for id in self.executor_map.keys():
            if self.executor_map[id]:
                if self.executor_map[id].basic_start_exited:
                    self.get_logger().error(f'remove finished executors: {id}')
                    self.executor_map[id] = None
                    ### self.executor_map.pop(id) ### mmodifes for loop I think
                    
    def timer_callback(self):
        ## self.remove_finished_executors()
        for i in range(len(self.futures)):
            future = self.futures[i]
            if future:
                if future.done():
                    resp = future.result()
                    # print(f'{self.futures_desc[i]}: {resp}')
                    self.futures[i] = None

    def set_start_executor_flag(self, id, flag):
        try:
            self.get_logger().error(f'set_start_executor_flag: {id} {flag} {type(self)}')
            req = TSTSetStartExecutorFlag.Request()
            req.id = id
            req.val = flag
            future = self.set_start_executor_flag_client.call_async(req)
            resp = self.wait_future(future)
            self.get_logger().error(f'set_start_executor_flag resp: {resp}')
            return resp
        except Exception as exc:        
            self.get_logger().error(f'EXCEPTION: {exc} - {type(exc)}')
    
                    
    def register_capability(self):
        self.get_logger().info(f'register_capability: wait for add capability server')
        self.add_capability_client.wait_for_service()
        self.get_logger().info(f'register_capability: capability server available: {self.ntype}')
        req = WDBAddCapability.Request()        
        req.capability = self.ntype
        req.expandable = self.expandable
        self.get_logger().info(f'call add capability in wdb: {req}')
        future = self.add_capability_client.call_async(req)
        self.futures.append(future)
        self.futures_desc.append(f'register_capability {self.ntype}')
        #resp = self.wait_future(future)        
        #print("register_capability:", resp)
        #return resp

    def wait_future(self, future, print_waiting=False):
        while not future.done():
            if print_waiting:
                print(f"wait_future executor")
        resp = future.result()
        return resp

    def get_node(self, id):
        self.lprint("exec get_node:", id)
        req = TSTGetNode.Request()
        req.id = id
        future = self.get_node_client.call_async(req)
        resp = self.wait_future(future)        
        self.lprint("exec get_node resp:", resp.info.type)
        return resp

    def create_executor(self, request, response):
        self.get_logger().info(f'executor_factory create_executor: {request}')
        start = timer ()
        try:
            id = request.id
            if id in self.executor_map:
                response.success = True
                response.error = 0
                response.reason = ""
                return response
            if self.myhandle:
                exec = self.get_executor(self, id, handle=self.myhandle)
            else:
                exec = self.get_executor(self, id)
            self.get_logger().info(f'create_executor - have exec: {request}')
            run_prepare = request.run_prepare
            if run_prepare and False: #debugging, test to run prepare in executor tick
                exec.prepare()
            self.get_logger().info(f'create_executor- after prepare: {request}')                
            resp = self.get_node(id)
            if resp.success:
                self.uuid_executor_map[resp.info.node_uuid] = exec
            self.executor_map[id] = exec
            response.success = True
            response.error = 0
            response.reason = ""
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION create_executor: {request} - {type(exc)} - {exc}')            
            response.success = False
            response.error = 1
            response.reason = "Exception"
        end = timer ()
        self.get_logger().info(f'create_executor response: {response} - {end-start}')        
        return response

    def destroy_executor(self, request, response):
        self.get_logger().error(f'destroy_executor: {request}')        
        id = request.id        
        if id in self.executor_map:
            self.executor_map.pop(id)
            response.success = True
            response.error = 0
            response.reason = ""
        else:
            response.success = False
            response.error = 1
            response.reason = f'Executor with id={id} does not exist'
        return response
    
    def start_executor(self, request, response):
        try:
            self.get_logger().info(f'start_executor: {request}')        
            id = request.id
            while not id in self.executor_map:
                self.get_logger().error("Wait for executor1")
                time.sleep(0.1)
            while not self.executor_map[id]:
                self.get_logger().error("Wait for executor2")
                time.sleep(0.1)
            if id in self.executor_map and self.executor_map[id]:
                self.get_logger().info(f"Call basic start on {type(self.executor_map[id])}")

                if self.executor_map[id].use_ticking:
                    self.get_logger().info(f'executor_factory: Use ticking, no threads')
                    if self.executor_map[id].basic_do_before_work():
                        self.executor_map[id].start()
                else:
                    self.get_logger().error(f'Create New Thread ({len(self.threads)})**************************************')
                    thread = threading.Thread(target=self.executor_map[id].basic_start)
                    thread.start()
                    self.threads.append(thread)
                response.success = True
                response.error = 0
                response.reason = ""
            else:
                self.get_logger().error(f'start_executor: Executor with id={id} does not exist')
                response.success = False
                response.error = 1
                response.reason = f'Executor with id={id} does not exist'
        except Exception as exc:
            self.get_logger().error(f'start_executor: Executor with id={id} throw exception - {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Executor with id={id} throw exception'
        self.get_logger().info(f'start_executor response: {response}')                    
        return response
    
    def signal_executor(self, request, response):
        try:
            self.get_logger().info(f'executor_factory signal_executor: {request}')            
            node_uuid = request.node_uuid
            signal = request.signal
            if node_uuid in self.uuid_executor_map:
                # bool can_do_signal = uuid_to_exec_map[req.node_uuid]->can_do_signal(req.signal);
                can_do_signal = True

                if can_do_signal:
                    self.get_logger().info(f'executor_factory push_signal: {node_uuid}')            
                    self.uuid_executor_map[node_uuid].push_signal(signal)
                    self.get_logger().info(f'executor_factory AFTER push_signal: {node_uuid}')            
                    response.success = True
                    response.error = 0
                    response.reason = ""
                else:
                    response.success = False
                    response.error = 4
                    response.reason = "Executor cannot do signal"
            else:
                response.success = False
                response.error = 1
                response.reason = "Could not find executor in uuid_executor_map"
        except Exception as exc:
            self.get_logger().error(f'signal_executor Exception: {type(exc)} - {exc}')            
            response.success = False
            response.error = 2
            response.reason = f'signal_executor throw exception'
        return response

    #
    # Delegation Level services
    #

    def executor_check(self, request, response):
        try:
            self.lprint("executor check:", request)
            params = Parameters(request.tni)
            (success, fail_reason) = self.check(params)
            if success:
                # std::vector<std::string> path_uuids = get_path_uuids ();
                # res.path_uuids = path_uuids;
                response.path_uuids = []
                response.success = True
                response.error = 0
                response.reason = ""
            else:
                response.success = False
                response.error = 1
                response.reason = fail_reason
        except Exception as exc:
            print(type(exc))
            print(exc)
            response.success = False
            response.error = 2
            response.reason = f'executor_check throw exception'
        self.lprint("executor check:", response)            
        return response

    def json_constraints(self, node_json):
        res = ""
        try:
            jobj = json.loads(node_json)
            if "name" in jobj:
                name = jobj["name"].replace("-", "_")
        except Exception as exc:
            print("EXCEPTION:", type(exc))
            print(exc)
        return res

    def executor_json_constraints(self, request, response):
        try:
            self.lprint("executor constraints:", request)
            response.constraints_json = self.json_constraints(request.node_json)
            response.success = True
            response.error = 0
            response.reason = ""
        except Exception as exc:
            print(type(exc))
            print(exc)
            response.free_id = 0
            response.success = False
            response.error = 2
            response.reason = f'executor_json_constraints throw exception'
        self.lprint("executor_json_constraints response:", response)                    
        return response
    
    def json_expand(self, node_json):
        res = ""
        try:
            jobj = json.loads(node_json)
            if "name" in jobj:
                name = jobj["name"].replace("-", "_")
        except Exception as exc:
            print("EXCEPTION:", type(exc))
            print(exc)
        return res

    def executor_json_expand(self, request, response):
        try:
            self.lprint("executor expand:", request)
            response.expansion_json = self.json_expand(request.node_json)
            response.success = True
            response.error = 0
            response.reason = ""
        except Exception as exc:
            print(type(exc))
            print(exc)
            response.free_id = 0
            response.success = False
            response.error = 2
            response.reason = f'executor_expand throw exception'
        self.lprint("executor_expand response:", response)                    
        return response
    
    def executor_expand(self, request, response):
        try:
            self.lprint("executor expand:", request)
            (fid, expansion_can_be_tried) = self.expand(request.id, request.free_id, request.possible_units, request.expansion_try)
            response.expansion_can_be_tried = expansion_can_be_tried
            if fid:
                response.free_id = fid
                response.success = True
                response.error = 0
                response.reason = ""
            else:
                response.free_id = 0
                response.success = False
                response.error = 2
                response.reason = "Expand failed"
        except Exception as exc:
            print(type(exc))
            print(exc)
            response.free_id = 0
            response.success = False
            response.error = 2
            response.reason = f'executor_expand throw exception'
        self.lprint("executor_expand response:", response)                    
        return response
    
    def is_delegation_expandable(self, request, response):
        try:
            self.lprint("executor is_delegation_expandable:", request)
            response.delegation_expandable_flag = self.delegation_expandable_flag
            response.success = True
            response.error = 0
            response.reason = ""
        except Exception as exc:            
            print(type(exc))
            print(exc)
            response.free_id = 0
            response.success = False
            response.error = 1
            response.reason = f'is_delegation_expandable throw exception'
        self.lprint("executor is_delegation_expandable:", response)                    
        return response

    # normally an executor type does not add constraints (has not been used in the system yet)
    def get_constraints(self, request, response):
        self.lprint("executor get_constraints:", request)
        response.csp_json = ""
        response.success = True
        response.error = 0
        response.reason = ""
        print("executor get_constraints:", response)                    
        return response

    #
    # Team Tasks Queue Level services
    #

    def executor_cost(self, request, response):
        self.lprint("executor cost (NYI):", request)
        response.cost = 0
        response.success = True
        response.error = 0
        response.reason = ""
        self.lprint("executor cost:", response)                    
        return response


    #
    # Delegation Level - methods to be possibly overridden
    #

    def expand(self, node_id, free_id, possible_units, expansion_try):
        try:
            self.lprint("executor_factory expand: BASIC NULL EXPAND:", node_id, free_id, possible_units, expansion_try)
            expansion_can_be_tried = expansion_try
            resp = self.get_node(node_id)
            if resp.success:
                tni = resp.info
                self.lprint("expand: got TNI")
            else:
                self.lprint("expand: Could not get node info")
                return (0, expansion_can_be_tried)
        except Exception as exc:            
            print(type(exc))
            print(exc)
        return (free_id, expansion_can_be_tried)


    def check(self, params):
        self.lprint("executor_factory check: Default, always success")
        fail_reason = ""        
        return (True, fail_reason)

    #
    # Team Task Queue Level - methods to be possibly overridden
    #
    
    def cost(self, params):
        self.lprint("executor_factory cost: Default, return 0 cost")
        fail_reason = ""
        return (0, fail_reason)
    
