import time
import rclpy
import importlib.metadata
from rclpy.executors import MultiThreadedExecutor

from lrs_exec.executor import Executor
from lrs_exec.executor_factory import ExecutorFactory

class ComputationExecutor(Executor):
    def __init__(self, node, id, computers):
        super().__init__(node, id)
        self.computers = computers

    def start(self):
        try:
            self.node.get_logger().error(f'{__name__} start computation')
            self.params.print()

            language = self.params.get_value("language")
            description = self.params.get_value("description")
            parameters = self.params.get_value("parameters", "")

            self.node.get_logger().info(f'LANGUAGE: {language}')
            self.node.get_logger().info(f'DESCRIPTION: {description}')
            self.node.get_logger().info(f'PARAMETERS2: {parameters}')

            if language in self.computers:
                computer = self.computers[language]
            else:
                eps = importlib.metadata.entry_points().select(group='lrs_exec.computation.create_computation_interface', name=language)
                if len(eps) > 0:
                    if len(eps) > 1:
                        self.fail(f"Got {len(eps)} entry points for {language}, should be only one.")
                        return
                    else:
                        (ep,) = eps
                        computer = ep.load()()
                        self.computers[language] = computer
                else:
                    self.fail(f"No computer for language '{language}'")
                    return

            computer.execute_computation(description, parameters)
            self.node.get_logger().info(f'Computation was executed succesfully.')
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION start: {type(exc)} - {exc}')
            self.fail(f"Computation Node executor failed.")
        self.node.get_logger().info(f'Computation executor is finished.')

class ComputationExecutorFactory(ExecutorFactory):
    def __init__(self):
        super().__init__("computation")
        self.computers = dict()

    def get_executor(self, node, id):
        return ComputationExecutor(node, id, self.computers)

def main(args=None):
    rclpy.init(args=args)
    
    executor = MultiThreadedExecutor(num_threads=8)    
    
    node = ComputationExecutorFactory()
    
    executor.add_node(node)    
    print(f"Spinning {__name__}")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
    
