import time
import uuid
import json
import os
import rclpy
from rclpy.executors import MultiThreadedExecutor

from lrs_exec.ticked_executor import TickedExecutor
from lrs_exec.ticked_executor_factory import TickedExecutorFactory
from lrs_util.json_exec_node import JsonExecNode
from lrs_util.coordtrans import CoordTrans

from lrs_exec.exec_tree import ExecTree
from lrs_srvs_exec.srv import TSTExecute

from std_msgs.msg import String

class InputHandlerExecutor(TickedExecutor):
    def __init__(self, node, id):
        super().__init__(node, id, blocking=False)
        self.node = node
        self.input_data = ""
        self.prime = True

        location = os.environ.get("LRS_LOCATION", default="")
        if location == "":
            self.node.get_logger().error("LRS_LOCATION is not set")
        else:
            self.coordtrans = CoordTrans(location=location)
        
        self.tst_execute_client = self.node.create_client(TSTExecute, f'tst_execute', callback_group=self.node.group)
        print("Wait for services...")
        self.tst_execute_client.wait_for_service()
        print("Services are up and running...")

    def input_callback(self, msg):
        print("input_callback:", msg)
        self.input_data = msg.data
        
    def do_work_begin(self):
        try:
            self.node.get_logger().info(f'do_work_begin: input_handler')
            self.params.print()
            # topic = self.params.get_value("topic")

            self.gp = self.params.get_value("waypoint")
            self.node.get_logger().info(f'GP: {self.gp}')            

            topic = "stream_deck_button"
            self.input_sub = self.node.create_subscription(String, topic, self.input_callback, 10, callback_group=self.node.group)        
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION do_work_begin: {type(exc)} - {exc}')

    def add_person(self, lat, lon, alt, object="medication"):
        try:
            self.node.get_logger().info(f'add_person: {lat} {lon} {alt}')            

            query = f"""salient region:
                            action: insert
                            dataset: granso
                            geometry: POINT({lon} {lat})
                            classes: ["http://askco.re/thing#human"]
                            properties:
                                "http://askco.re/thing/property#needs": "http://askco.re/thing#{object}"
                            """
            print("QUERY:", query)
            query_node = JsonExecNode("", name="query", language="kdb/krql", query=query)
            return query_node
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION add_persson: {type(exc)} - {exc}')            
        
            
    def add_object(self, x, y, z, object="medication"):
        try:
            
            #if self.lat > 1.0:
            #    (lat, lon, alt) = (self.lat, self.lon, self.alt)
            #else:

            (lon, lat, alt) = self.coordtrans.world_to_wgs84(x, y, z)            
            self.node.get_logger().info(f'add_object: {lat} {lon} {alt}')            

            query = f"""salient region:
                            action: insert
                            dataset: granso
                            geometry: POINT({lon} {lat})
                            classes: ["http://askco.re/thing#depot"]
                            properties:
                                "http://askco.re/thing#{object}": 1
                            """
            print("QUERY:", query)
            query_node = JsonExecNode("", name="query", language="kdb/krql", query=query)
            return query_node
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION add_object: {type(exc)} - {exc}')

    def clear_graph(self):
        try:
            self.node.get_logger().info(f'clear_graph')            

            graph_name="granso"
            
            query = f"""DROP SILENT GRAPH <{graph_name}>; CREATE GRAPH <{graph_name}>"""
            print("QUERY:", query)
            query_node = JsonExecNode("", name="query", language="kdb/sparql", query=query)
            return query_node
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION clear_graph: {type(exc)} - {exc}')
            
    def mission(self, object="medication"):
        try:
            mission_document = "retrieve: { what: [document, format], from: missions, matches: { name: deliver_supplies }}"
            mission_mode = "kdql"
            parameters_document = f"{{ source_graph: granso, supply_type: http://askco.re/thing#{object} }}"
            parameters_format = "yaml"
            parameters_mode = "raw"
            mission = JsonExecNode("", name="mission", mission_document=mission_document, mission_mode=mission_mode, parameters_document=parameters_document, parameters_format=parameters_format, parameters_mode=parameters_mode)
            return mission
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION mission: {type(exc)} - {exc}')
            
    def do_work(self):
        try:
            # self.node.get_logger().info(f'do_work: input_handler')
            # self.params.print()

            lat = self.gp.latitude
            lon = self.gp.longitude
            alt = self.gp.altitude

            if self.input_data:
                print("GOT INPUT:", self.input_data)

                ns = self.node.get_namespace()

                say_string = "Sorry we could not help"
                deliver_type = ""
                if self.input_data == "medicine" or self.input_data == "food":
                    say_string = f"A delivery of {self.input_data} has been requested"
                    deliver_type = self.input_data

                #
                # Add say
                #
                    
                say = JsonExecNode(ns, name="say", string=say_string, gain="15")
                tree = JsonExecNode(ns, name="seq", node_uuid=str(uuid.uuid4()))
                tree.add_child(say)
                    
                say2 = JsonExecNode("", name="say", string=say_string)

                tree2 = JsonExecNode("", name="seq", node_uuid=str(uuid.uuid4()))
                tree2.add_child(say2)
                if not self.prime:
                    task_string = json.dumps(tree2.get_json(), sort_keys=True, indent=4, separators=(',', ': '))
                    task_type = "json-tst"
                    task_priority = 50
                    team_name = "speak"
                    task_name = "say-from-husky"

                    jobj = JsonExecNode("/op0", name="add-task-to-team-queue", task_string=task_string, team_name=team_name, task_priority=task_priority, task_type=task_type, task_name=task_name)
                    tree.add_child(jobj)
                tree.add_child(say)
                print("TREE:", tree)
                exec_tree = ExecTree("input_handler")
                exec_tree.tst_execute_client = self.tst_execute_client
                resp = exec_tree.new_handle_tree(tree.get_json(), channel=1)
                print("RESP:", resp)


                #
                # Add delivery trees
                #

                if deliver_type:
                    delivertree = JsonExecNode("", name="seq", node_uuid=str(uuid.uuid4()))
                    object = "medication"
                    if deliver_type == "food":
                        object = "food"                        
                    clear_graph_node = self.clear_graph()
                    person_node = self.add_person(lat, lon, alt, object=object)
                    ##object_node1 = self.add_object(41.0, 15.0, 2.0, object="medication")
                    ##object_node2 = self.add_object(38.0, 18.0, 2.0, object="food")
                    object_node1 = self.add_object(1.2, 3.0, 2.0, object="medication")
                    object_node2 = self.add_object(3.7, 0.2, 2.0, object="food")
                    mission_node = self.mission(object=object)
                    (hlon, hlat, halt) = self.coordtrans.world_to_wgs84(-3.8, 2.7, 2.0)
                    move_to = JsonExecNode("", name="move-to", lat=hlat, lon=hlon, alt=halt, speed="standard")
                    
                    delivertree.add_child(clear_graph_node)
                    delivertree.add_child(person_node)
                    delivertree.add_child(object_node1)
                    delivertree.add_child(object_node2)
                    delivertree.add_child(mission_node)
                    delivertree.add_child(move_to)
                    
                    task_string = json.dumps(delivertree.get_json(), sort_keys=True, indent=4, separators=(',', ': '))
                    print("TASK STRING:", task_string)
                    task_type = "json-tst"
                    task_priority = 50
                    team_name = "deliver"
                    task_name = f"deliver-{deliver_type}"

                    jobj = JsonExecNode("/op0", name="add-task-to-team-queue", task_string=task_string, team_name=team_name, task_priority=task_priority, task_type=task_type, task_name=task_name)
                    tree = JsonExecNode(ns, name="seq", node_uuid=str(uuid.uuid4()))
                    tree.add_child(jobj)
                    print("TREE:", tree)
                    exec_tree = ExecTree("input_handler")
                    exec_tree.tst_execute_client = self.tst_execute_client
                    resp = exec_tree.new_handle_tree(tree.get_json(), channel=2)
                    print("RESP:", resp)



                
                self.do_work_finished()
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION do_work: {type(exc)} - {exc}')

class InputHandlerExecutorFactory(TickedExecutorFactory):
    def __init__(self):
        super().__init__("input-handler")

    def get_executor(self, node, id):
        return self.add_executor(InputHandlerExecutor(node, id))

def main(args=None):
    rclpy.init(args=args)
    
    executor = MultiThreadedExecutor(num_threads=8)    
    
    node = InputHandlerExecutorFactory()
    
    executor.add_node(node)    
    print(f"Spinning {__name__}")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
    
