import time
import rclpy
from rclpy.executors import MultiThreadedExecutor
from rclpy.action import ActionClient

from lrs_exec.executor import Executor
from lrs_exec.executor_factory import ExecutorFactory

from ros_mission_server_interfaces.action import ExecuteMission

class MissionExecutor(Executor):
    def __init__(self, node, id):
        super().__init__(node, id)
        self.action_client = ActionClient(self.node, ExecuteMission, 'execute_mission')

    def start(self):
        try:
            self.node.get_logger().error(f'{__name__} start mission')
            self.params.print()

            mission_request = ExecuteMission.Goal()
            mission_request.mission.mode = self.params.get_value("mission_mode")
            mission_request.mission.format = self.params.get_value_always("mission_format", "")
            mission_request.mission.document = self.params.get_value("mission_document")
            mission_request.parameters.mode = self.params.get_value("parameters_mode")
            mission_request.parameters.format = self.params.get_value_always("parameters_format", "")
            mission_request.parameters.document = self.params.get_value("parameters_document")

            self.node.get_logger().info(f'MISSION: {mission_request.mission.mode} (format {mission_request.mission.format}) with {mission_request.mission.document}')
            self.node.get_logger().info(f'PARAMETERS: {mission_request.parameters.mode} (format {mission_request.parameters.format}) with {mission_request.parameters.document}')

            self.action_client.wait_for_server()

            res = self.action_client.send_goal(mission_request).result
            if res.success:
                self.node.get_logger().info(f'Mission was executed succesfully.')
            else:
                self.node.get_logger().error(f'Mission failed to execute with error "{res.message}"')
                self.fail(f'Mission failed to execute with error "{res.message}"')
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION start: {type(exc)} - {exc}')
            self.fail(f"Mission Node executor failed.")
        self.node.get_logger().info(f'Mission executor is finished.')

class MissionExecutorFactory(ExecutorFactory):
    def __init__(self):
        super().__init__("mission")

    def get_executor(self, node, id):
        return MissionExecutor(node, id)

def main(args=None):
    rclpy.init(args=args)
    
    executor = MultiThreadedExecutor(num_threads=8)    
    
    node = MissionExecutorFactory()
    
    executor.add_node(node)    
    print(f"Spinning {__name__}")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
    
