import time
import rclpy
from rclpy.executors import MultiThreadedExecutor

from lrs_exec.ticked_executor import TickedExecutor
from lrs_exec.ticked_executor_factory import TickedExecutorFactory

from std_msgs.msg import String

class MonitorExecutor(TickedExecutor):
    def __init__(self, node, id):
        super().__init__(node, id, blocking=False)
        self.set_signals(["$abort", "$pause", "$continue", "$enough", "$finished"])
        self.node = node
        self.pub = None
        self.monitor_finished = False

    def monitor_response_callback(self, data):
        print("monitor_response_callback:", data)
        if data.data == "finished":
            print("FINISHED")
            self.monitor_finished = True
        
    def do_work_begin(self):
        try:
            self.node.get_logger().info(f'{__name__} do_work_begin: ')
            self.params.print()            

            topic = self.params.get_value("topic", default="monitor")
            
            string = self.params.get_value("string", default="start")

            print("STRING:", string)
            print("TOPIC:", topic)
            self.pub = self.node.create_publisher(String, topic, 10)
            self.sub = self.node.create_subscription(String, "monitor_response", self.monitor_response_callback, 10)
            
            msg = String()
            msg.data = string
            self.pub.publish(msg)
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION do_work_begin: {type(exc)} - {exc}')

    def do_work(self):
        try:
            if self.have_signal():
                signal = self.pop_signal()
                self.node.get_logger().info(f'Monitor got signal: {signal}')
                if signal == "$finished":
                    self.signal_finish()
                    self.do_work_finished()
                    return
                if signal == "$enough":
                    self.do_work_finished()
                if signal == "$abort":
                    self.abort_fail("Got abort signal")
                    self.do_work_finished()
                if signal == "$pause":
                    self.paused = True
                if signal == "$continue":
                    self.paused = False
            if self.monitor_finished:
                self.do_work_finished()
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION start: {type(exc)} - {exc}')
            self.fail(f'Exception: {exc}')
            

class MonitorExecutorFactory(TickedExecutorFactory):
    def __init__(self):
        super().__init__("monitor")
        ### topic = "stream_deck_config"

    def get_executor(self, node, id):
        return self.add_executor(MonitorExecutor(node, id))

def main(args=None):
    rclpy.init(args=args)
    
    executor = MultiThreadedExecutor(num_threads=8)    
    
    node = MonitorExecutorFactory()
    
    executor.add_node(node)    
    print(f"Spinning {__name__}")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
    
