import time
import rclpy
import timeit
from timeit import default_timer as timer
from rclpy.executors import MultiThreadedExecutor

from lrs_exec.executor import Executor
from lrs_exec.executor_factory import ExecutorFactory
from lrs_msgs_tst.msg import TSTNodeStatus

class SeqExecutor(Executor):
    def __init__(self, node, id, collection=None):
        super().__init__(node, id)
        self.timings_start = {}
        self.timings_end = {}
        self.use_ticking = True
        self.uuids = []
        self.uuid_to_id = {}
        self.collection = collection

        # tick part
        self.state = ""
        self.next_child = 0
        self.old_state = ""
        self.inside_tick = False
        self.started_flag = False
        self.current_uuid = ""
        self.finished = False

        self.stuck_in_tick = 0

    def tick(self):
        if self.inside_tick:
            if self.stuck_in_tick < 5:
                self.node.get_logger().error(f'seq {self.id}: {self.state} - INSIDE TICK: RETURN')
            self.stuck_in_tick += 1
            return
        self.inside_tick = True
        self.stuck_in_tick = 0;
        ## self.node.get_logger().error(f'seq {self.id}: TICK: {self.started_flag}')
        try:
            if self.started_flag:
                if self.old_state != self.state:
                    self.node.get_logger().info(f'seq {self.id}: {self.old_state} -> {self.state}')
                    self.old_state = self.state
                if self.state == "next":
                    if len(self.uuids) > self.next_child:
                        uid = self.uuids[self.next_child]
                        self.node.get_logger().info(f'seq: call set start executor flag: {self.next_child}')
                        self.next_child += 1
                        ##self.node.get_logger().error(f'seq: call set start executor flag')
                        if self.collection:
                            self.collection.start_executor(self.uuid_to_id[uid], uid)
                        else:
                            resp = self.set_start_executor_flag(self.uuid_to_id[uid], True)
                            self.node.get_logger().info(f'seq: resp set start executor flag: {resp}')
                            if not resp.success:
                                self.fail("Could not start executor for child")
                                self.state = "finished"
                        self.finished = False
                        self.current_uuid = uid
                        self.state = "waiting"
                    else:
                        self.state = "finished"
                    self.inside_tick = False
                    return
                if self.state == "waiting":
                    if self.finished:
                        self.state = "next"
                    self.inside_tick = False
                    return
                if self.state == "finished":
                    if self.normal_exit():
                        self.wait_for_postwork_conditions()
                    else:
                        self.node.get_logger().error(f'seq {self.id}: {self.state} - abnormal exit')
                        unlock_res = self.unlock_resources()
                        self.node.get_logger().error(f'seq {self.id}: {self.state} - after unlock resources')
                        exec_unlock_res = True
                        if self.tni.use_execution_lock:
                            exec_unlock_res = self.execution_lock_release(self, tni.root_uuid)
                            self.node.get_logger().error(f'seq {self.id}: {self.state} - after exec lock release')
                    self.state = "exit"
                    self.inside_tick = False
                    return
        except Exception as exc:
            self.node.get_logger().error(f'Exception notify_node_status: {type(exc)} - {exc}')
        self.inside_tick = False
        
    def notify_node_status(self, msg):
        try:
            self.node.get_logger().error(f'seq node_status: {msg.node_uuid} - {self.current_uuid}')
            if self.current_uuid == msg.node_uuid:
                self.node.get_logger().error(f'***seq node_status {self.id}: {msg.id}')
                if msg.exec_status.finished:
                    if not msg.exec_status.succeeded:
                        self.fail("Child execution failed")
                    self.node.get_logger().error(f'***seq node_status FINISHED {self.id}: {msg.id}')
                    self.finished = True
        except Exception as exc:
            self.node.get_logger().error(f'Exception notify_node_status: {type(exc)} - {exc}')

        
    def start(self):
        try:
            self.node.get_logger().info(f'seq start: {len(self.tni.children)}')
            for cid in self.tni.children:
                self.node.get_logger().info(f'seq start cid: {cid}')
                val = self.get_uuid_from_id(cid)
                if val:
                    self.uuid_to_id[val] = cid
                    self.uuids.append(val)
                else:
                    self.node.get_logger().info(f'could not get node uuid for child: {cid}')
                    self.fail("Could not get node uuid for a child")
                    self.started_flag = True
                    self.next_child = 0
                    self.state = "finished"
                    return

            self.node.get_logger().info(f'seq: uuids: {self.uuids}')                    

            self.started_flag = True
            self.next_child = 0
            self.state = "next"
            return
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION start: {type(exc)} - {exc}')
            self.fail(f"seq Exception: {exc}")
            

class SeqExecutorFactory(ExecutorFactory):
    def __init__(self, collection=None):
        super().__init__("seq")
        self.node_status_sub = self.create_subscription(TSTNodeStatus, "tst_node_status", self.node_status, 10, callback_group=self.group)
        self.executors = []
        # was 0.05
        self.tick_timer = self.create_timer(0.2, self.tick, callback_group=self.group)
        self.collection = collection

    def tick(self):
        ## self.get_logger().error(f'seq: FACTICK: {len(self.executors)}')
        executors = self.executors.copy()
        try:
            for obj in executors:
                ## self.get_logger().error(f'seq: OBJTICK: {len(self.executors)}')                        
                obj.tick()
            remove_list = []
            for obj in executors:                
                if obj.state == "exit":
                    remove_list.append(obj)
            self.executors_lock.acquire()
            for obj in remove_list:
                if obj in self.executors:
                    self.get_logger().error(f'Remove from seq executors: {obj.id}')
                    self.executors.remove(obj)
            self.executors_lock.release()
            
        except Exception as exc:
            self.get_logger().error(f'Seq fac tick Exception: {type(exc)} - {exc}')
                    
        
    def node_status(self, msg):
        try:
            # self.get_logger().error(f'node_status: {msg.node_uuid} {msg.id}')

            self.executors_lock.acquire()
            for obj in self.executors:
                obj.notify_node_status(msg)
            self.executors_lock.release()                            
        except Exception as exc:
            self.get_logger().error(f'Exception node_status: {type(exc)} - {exc}')

    def get_executor(self, node, id):
        ##self.get_logger().error(f'seq get_executor: {node} {id}')
        obj = SeqExecutor(node, id, collection=self.collection)
        ##self.get_logger().error(f'seq get_executor HAVE OBJ: {node} {id}')
        self.executors_lock.acquire()                            
        self.executors.append(obj)
        self.executors_lock.release()
        return obj

def main(args=None):
    rclpy.init(args=args)
    
    executor = MultiThreadedExecutor(num_threads=8)    
    
    node = SeqExecutorFactory()
    
    executor.add_node(node)    
    print(f"Spinning {__name__}")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
