import time
import rclpy
from rclpy.executors import MultiThreadedExecutor

from lrs_exec.executor import Executor
from lrs_exec.executor_factory import ExecutorFactory
from lrs_exec.parameters import Parameters
from lrs_msgs_tst.msg import TSTNodeStatus

class TestIfExecutor(Executor):
    def __init__(self, node, id, collection=None):
        super().__init__(node, id)
        self.use_ticking = True
        self.uuids = []
        self.uuid_to_id = {}
        self.collection = collection
        
        # tick part
        self.state = ""
        self.next_child = 0
        self.old_state = ""
        self.inside_tick = False
        self.started_flag = False
        self.current_uuid = ""
        self.finished = False

    def cleanup(self):
         self.node.destroy_subscription(self.node_status_sub)
         self.node_status_sub = None
        
    def tick(self):
        if self.inside_tick:
            return
        self.inside_tick = True
        ## self.node.get_logger().error(f'seq {self.id}: TICK: {self.started_flag}')
        try:
            if self.started_flag:
                if self.old_state != self.state:
                    self.node.get_logger().info(f'test-if {self.id}: {self.old_state} -> {self.state}')
                    self.old_state = self.state
                    
                if self.state == "condition":
                    uid = self.uuids[0]
                    if self.collection:
                        self.collection.start_executor(self.uuid_to_id[uid], uid)
                        self.finished = False
                        self.current_uuid = uid
                        self.state = "running-condition"
                    else:
                        resp = self.set_start_executor_flag(self.uuid_to_id[uid], True)
                        if not resp.success:
                            self.fail("Could not start executor for child 0")
                            self.state = "finished"
                        else:
                            self.finished = False
                            self.current_uuid = uid
                            self.state = "running-condition"
                    self.inside_tick = False
                    return

                if self.state == "running-condition":
                    if self.finished:
                        self.state = "condition-finished"
                        self.finished = False
                    self.inside_tick = False
                    return
                
                if self.state == "condition-finished":
                    resp = self.get_node_from_uuid(self.current_uuid)
                    # print(f'{__name__}: {resp}')
                    params = Parameters(resp.info)

                    answer = params.get_value("answer")
                    self.node.get_logger().info(f'test_if answer: {answer}')
            
                    if not answer:
                        self.node.get_logger().error("Query node did not return any answer")
                        self.fail("Query node did not return any answer")
                        self.state = "finished"
                        self.inside_tick = False
                        return

                    if not answer in ["unknown", "true", "false"]:
                        self.node.get_logger().error(f"Query node returned unexpected value: {answer}")
                        self.fail(f"Query node returned unexpected value: {answer}")
                        self.state = "finished"
                        self.inside_tick = False
                        return

                    self.state = answer
                    self.inside_tick = False
                    return

                if self.state == "true":
                    index = 1
                    uid = self.uuids[index]
                    if self.collection:
                        self.collection.start_executor(self.uuid_to_id[uid], uid)
                        self.current_uuid = uid
                        self.state = "waiting"
                    else:
                        resp = self.set_start_executor_flag(self.uuid_to_id[uid], True)
                        if not resp.success:
                            self.fail(f"Could not start executor for child {index}")
                            self.state = "finished"
                        else:
                            self.current_uuid = uid
                            self.state = "waiting"
                    self.inside_tick = False
                    return
                    
                if self.state == "false":
                    index = 2
                    uid = self.uuids[index]
                    if self.collection:
                        self.collection.start_executor(self.uuid_to_id[uid], uid)
                        self.current_uuid = uid
                        self.state = "waiting"
                    else:
                        resp = self.set_start_executor_flag(self.uuid_to_id[uid], True)
                        if not resp.success:
                            self.fail(f"Could not start executor for child {index}")
                            self.state = "finished"
                        else:
                            self.current_uuid = uid
                            self.state = "waiting"
                    self.inside_tick = False
                    return
                    
                if self.state == "unknown":
                    index = 3
                    uid = self.uuids[index]
                    if self.collection:
                        self.collection.start_executor(self.uuid_to_id[uid], uid)
                        self.current_uuid = uid
                        self.state = "waiting"
                    else:
                        resp = self.set_start_executor_flag(self.uuid_to_id[uid], True)
                        if not resp.success:
                            self.fail(f"Could not start executor for child {index}")
                            self.state = "finished"
                        else:
                            self.current_uuid = uid
                            self.state = "waiting"
                    self.inside_tick = False
                    return
                    
                if self.state == "waiting":
                    if self.finished:
                        self.state = "finished"
                    self.inside_tick = False
                    return
                
                if self.state == "finished":
                    if self.normal_exit():
                        self.wait_for_postwork_conditions()
                    else:
                        unlock_res = self.unlock_resources()
                        exec_unlock_res = True
                        if self.tni.use_execution_lock:
                            exec_unlock_res = self.execution_lock_release(self, tni.root_uuid)
                    self.state = "exit"
                    self.inside_tick = False
                    return
        except Exception as exc:
            self.node.get_logger().error(f'Exception notify_node_status: {type(exc)} - {exc}')
        self.inside_tick = False

    def notify_node_status(self, msg):
        try:
            ## self.node.get_logger().error(f'seq node_status: {msg.node_uuid} - {self.current_uuid}')
            if self.current_uuid == msg.node_uuid:
                ### self.node.get_logger().error(f'***seq node_status {self.id}: {msg.id}')
                if msg.exec_status.finished:
                    if not msg.exec_status.succeeded:
                        self.fail("Child execution failed")
                    ### self.node.get_logger().error(f'***seq node_status FINISHED {self.id}: {msg.id}')
                    self.finished = True
        except Exception as exc:
            self.node.get_logger().error(f'Exception notify_node_status: {type(exc)} - {exc}')

        
    def start(self):
        try:
            self.node.get_logger().info(f'TestIf start: {len(self.tni.children)}')
            self.uuid_to_id = {}
            self.uuids = []
            for cid in self.tni.children:
                self.node.get_logger().info(f'seq start cid: {cid}')
                val = self.get_uuid_from_id(cid)
                if val:
                    self.uuid_to_id[val] = cid
                    self.uuids.append(val)
                else:
                    self.node.get_logger().info(f'could not get node uuid for child: {cid}')                    
                    self.fail("Could not get node uuid for a child")
                    self.started_flag = True
                    self.next_child = 0
                    self.state = "finished"
                    return

            self.started_flag = True
            self.state = "condition"
            return
                
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION start: {type(exc)} - {exc}')
            self.fail(f"{__name__} Exception: {exc}")

        self.node.get_logger().info(f'TestIf end')            

class TestIfExecutorFactory(ExecutorFactory):
    def __init__(self, collection=None):
        super().__init__("test-if")
        
    def node_status(self, msg):
        try:
            # self.get_logger().error(f'node_status: {msg.node_uuid} {msg.id}')
            self.executors_lock.acquire()                    
            for obj in self.executors:
                obj.notify_node_status(msg)
            self.executors_lock.release()                                        
        except Exception as exc:
            self.get_logger().error(f'Exception node_status: {type(exc)} - {exc}')
        
    #def get_executor(self, node, id):
    #    obj = TestIfExecutor(node, id, collection=self.collection)
    #    self.executors_lock.acquire()                            
    #    self.executors.append(obj)
    #    self.executors_lock.release()                            
    #    return obj

def main(args=None):
    rclpy.init(args=args)
    
    executor = MultiThreadedExecutor(num_threads=8)    
    
    node = TestIfExecutorFactory()
    
    executor.add_node(node)    
    print(f"Spinning {__name__}")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
    
