import time
import rclpy
import importlib.metadata
from rclpy.executors import MultiThreadedExecutor

from lrs_exec.executor import Executor
from lrs_exec.executor_factory import ExecutorFactory

class QueryExecutor(Executor):
    def __init__(self, node, id, queriers):
        super().__init__(node, id)
        self.queriers = queriers

    def start(self):
        try:
            self.node.get_logger().error(f'{__name__} start query')
            self.params.print()

            language = self.params.get_value("language")
            query = self.params.get_value("query")
            bindings = self.params.get_value("bindings")

            self.node.get_logger().info(f'LANGUAGE: {language}')
            self.node.get_logger().info(f'QUERY: {query}')
            self.node.get_logger().info(f'BINDINGS: {bindings}')

            if language in self.queriers:
                querier = self.queriers[language]
            else:
                eps = importlib.metadata.entry_points().select(group='lrs_exec.query.create_query_interface', name=language)
                if len(eps) > 0:
                    if len(eps) > 1:
                        self.fail(f"Got {len(eps)} entry points for {language}, should be only one.")
                        return
                    else:
                        (ep,) = eps
                        querier = ep.load()()
                        self.queriers[language] = querier
                else:
                    self.fail(f"No querier for language '{language}'")
                    return

            querier.execute_query(query, bindings)
            self.node.get_logger().info(f'Query was executed succesfully.')
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION start: {type(exc)} - {exc}')
            self.fail(f"Query Node executor failed.")
        self.node.get_logger().info(f'Query executor is finished.')

class QueryExecutorFactory(ExecutorFactory):
    def __init__(self):
        super().__init__("query")
        self.queriers = dict()

    def get_executor(self, node, id):
        return QueryExecutor(node, id, self.queriers)

def main(args=None):
    rclpy.init(args=args)
    
    executor = MultiThreadedExecutor(num_threads=8)    
    
    node = QueryExecutorFactory()
    
    executor.add_node(node)    
    print(f"Spinning {__name__}")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
    
