import os
import time
import rclpy
import json
import uuid
from rclpy.executors import MultiThreadedExecutor

from lrs_exec.ticked_executor import TickedExecutor
from lrs_exec.executor_factory import ExecutorFactory
from lrs_exec.ticked_executor_factory import TickedExecutorFactory
from lrs_util.jsonutil import json_basic_seq

from lrs_msgs_common.srv import SendTopicMsgToUnit
from lrs_msgs_common.msg import UnitTopicMsg
from std_msgs.msg import String

class RemoteExecutor(TickedExecutor):
    def __init__(self, node, id, stmtu_client=None):
        super().__init__(node, id, blocking=False)
        self.set_signals(["$abort", "$pause", "$continue", "$enough", "$finished"])

        self.transport = "mqtt" # "zenoh", "mqtt", ...
        if os.getenv("TST_REMOTE_TRANSPORT"):
            self.transport = os.getenv("TST_REMOTE_TRANSPORT")

        self.stmtu_client = stmtu_client

        self.com_uuid = None
        self.pong_uuid = None

        # tick part
        self.state = ""
        self.next_child = 0
        self.old_state = ""
        self.current_uuid = ""
        self.finished = False

        self.pub_tst_command = None
        self.response = {}
        self.feedback = {}

        self.use_tst_execution = False

        self.root_uuid = "dummy_root_uuid"
        self.task_uuid = "dummy_task_uuid"

        self.pong_response = {}
        self.have_pong = False
        self.have_started = False
        self.counter = 0
        self.old_status = ""
        self.status = ""
        self.have_response = False

    def notify_unit_topic_msg(self, msg):
        try:
            jobj = json.loads(msg.msg)
            msgstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
            ### self.node.get_logger().info(f'notify_unit_topic_msg message formatted {self.id}: {msg.topic} - {msgstr}')
            if self.use_tst_execution:
                if msg.topic == "tst/response":
                    response_to = jobj["response-to"]
                    self.response[response_to] = jobj
                if msg.topic == "tst/feedback":
                    node_uuid = jobj["node-uuid"]
                    self.feedback[node_uuid] = jobj
            else:
                ### self.node.get_logger().info(f'notify_unit_topic_msg message formatted {self.id}: {msg.topic} - {msgstr}')
                if msg.topic == "exec/response" or msg.topic == "exec/feedback":
                    ## self.node.get_logger().info(f'notify_unit_topic_msg message formatted: {msg.topic} - {msgstr}')
                    if msg.topic == "exec/response":
                        self.node.get_logger().info(f'notify_unit_topic_msg message formatted pong uuid - obj uuid: {self.pong_uuid} - {jobj["response-to"]} - {jobj["response"]}')
                    if msg.topic == "exec/response" and jobj["response"] == "pong":
                        self.pong_response[jobj["response-to"]] = True
                if msg.topic == "exec/response":
                    ### self.node.get_logger().info(f'RESPONSE notify_unit_topic_msg message formatted {self.id}: {msg.topic} - {msgstr}')
                    response_to = jobj["response-to"]
                    self.response[response_to] = jobj
                if msg.topic == "exec/feedback":
                    self.node.get_logger().info(f'FEEDBACK notify_unit_topic_msg message formatted {self.id}: {msg.topic} - {msgstr}')                    
                    task_uuid = jobj["task-uuid"]
                    self.feedback[task_uuid] = jobj            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION notify_unit_topic_msg: {type(exc)} - {exc}')

    def start_tst(self, treeobj, execution_unit):
        try:
            com_uuid = str(uuid.uuid4())
            tstobj = {
                "com-uuid": com_uuid,
                "command": "start-tst",
                "receiver": execution_unit,
                "sender": "ros2-json-command",
                "tst": treeobj
            }
            msgstr = json.dumps(tstobj, sort_keys=True, indent=4, separators=(',', ': '))
            self.node.get_logger().info(f'START TST: {self.id}')            
            ##print(f'START_TST: {msgstr} - {execution_unit}')

            if self.transport == "mqtt":
                req = SendTopicMsgToUnit.Request()
                req.unit = execution_unit
                req.topic = "tst/command"
                req.msg = msgstr
                future = self.stmtu_client.call_async(req)
                resp = self.wait_future(future)        
                ##print("remote.py send_topic_msg_to_unit resp:", resp)
            
            # self.client.publish(f'{self.prefix}/{self.unit}/tst/command', payload=msgstr)
            return com_uuid
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION start_tst: {type(exc)} - {exc}')

    def ping(self, execution_unit):
        try:
            com_uuid = str(uuid.uuid4())
            taskobj = {
                "com-uuid": com_uuid,
                "command": "ping",
            }
            msgstr = json.dumps(taskobj, sort_keys=True, indent=4, separators=(',', ': '))
            self.node.get_logger().info(f'+++++++++++++++++++++++PING {self.id}: {msgstr} - {execution_unit}')
            self.node.get_logger().info(f'TRANSPORT PING {self.transport}')            

            if self.transport == "zenoh":
                msg = String()
                msg.data = msgstr
                self.exec_command_pub.publish(msg)

            if self.transport == "mqtt":
                req = SendTopicMsgToUnit.Request()
                req.unit = execution_unit
                req.topic = "exec/command"
                req.msg = msgstr
                print("PING MQTT REQ:", req)
                future = self.stmtu_client.call_async(req)
                resp = self.wait_future(future)
                print("PING MQTT RESP:", resp)
                if not resp.success:
                    com_uuid = ""
                # do not wait for responce since that could cause exec/response arriving before rturning here ane pong_uuid == None
                #self.node.get_logger().info(f"remote.py send_topic_msg_to_unit resp {self.id}: {resp}")
                #if not resp.success:
                #    com_uuid = ""
                
                # self.client.publish(f'{self.prefix}/{self.unit}/tst/command', payload=msgstr)
            return com_uuid
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION ping {self.id}: {type(exc)} - {exc}')

    def exec_response_callback(self, msg):
        print("exec_response_callback:", msg)
        jobj = json.loads(msg.data)
        print("EXEC RESPONSE:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
        response_to = jobj["response-to"]
        if jobj["response"] == "pong":
            self.pong_response[jobj["response-to"]] = True
            
    def exec_feedback_callback(self, msg):
        print("exec_feedback_callback:", msg)
        jobj = json.loads(msg.data)
        print("EXEC FEEDBACK:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
        task_uuid = jobj["task-uuid"]
        status = jobj["status"]
        self.feedback[task_uuid] = jobj            
            
    def start_task(self, task, execution_unit):
        try:
            com_uuid = str(uuid.uuid4())
            task_uuid = str(uuid.uuid4())
            taskobj = {
                "com-uuid": com_uuid,
                "command": "start-task",
                "execution-unit": execution_unit,
                "sender": "remote",
                "task": task,
                "task-uuid": task_uuid
            }
            msgstr = json.dumps(taskobj, sort_keys=True, indent=4, separators=(',', ': '))
            self.node.get_logger().info(f'************************************START_TASK {self.id}: {execution_unit} - {task_uuid}')
            ## self.node.get_logger().info(f'START_TASK: {msgstr} - {execution_unit}')

            if self.transport == "zenoh":
                msg = String()
                msg.data = msgstr
                self.exec_command_pub.publish(msg)
            
            if self.transport == "mqtt":            
                req = SendTopicMsgToUnit.Request()
                req.unit = execution_unit
                req.topic = "exec/command"
                req.msg = msgstr
                future = self.stmtu_client.call_async(req)
                resp = self.wait_future(future)        
                self.node.get_logger().info(f"remote.py send_topic_msg_to_unit resp {self.id}: {resp}")
            
                # self.client.publish(f'{self.prefix}/{self.unit}/tst/command', payload=msgstr)
            return (com_uuid, task_uuid)
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION start_task {self.id}: {type(exc)} - {exc}')

    def signal_task(self, task_uuid, signal, execution_unit):
        try:
            com_uuid = str(uuid.uuid4())
            obj = {
                "com-uuid": com_uuid,
                "command": "signal-task",
                "signal": signal,
                "sender": "remote",
                "task-uuid": task_uuid
            }
            msgstr = json.dumps(obj, sort_keys=True, indent=4, separators=(',', ': '))
            self.node.get_logger().info(f'SIGNAL_TASK {self.id}: {msgstr} - {execution_unit}')

            if self.transport == "mqtt":            
                req = SendTopicMsgToUnit.Request()
                req.unit = execution_unit
                req.topic = "exec/command"
                req.msg = msgstr
                future = self.stmtu_client.call_async(req)
                resp = self.wait_future(future)        
                self.node.get_logger().info(f"remote.py send_topic_msg_to_unit resp {self.id}: {resp}")
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION signal_task {self.id}: {type(exc)} - {exc}')

    def do_work_begin(self):            
        try:
            self.node.get_logger().info(f'RemoteExecutor do_work_begin {self.id}: {self.tni.type}')            

            jobjstr = self.get_exec_json()
            ##self.node.get_logger().info(f'JOBJSTR: {jobjstr}')
            if jobjstr:
                self.jobj = json.loads(jobjstr)
                fjobjstr = json.dumps(self.jobj, sort_keys=True, indent=4, separators=(',', ': '))
                ##self.node.get_logger().info(f'FJOBJSTR: {fjobjstr}')
            else:
                self.fail(f"{self.id}: could not get json for node")
                self.node.get_logger().error(f"{self.id}: could not get json for node")                
                return

            execution_ns = self.jobj["execution_ns"]
            self.execution_ns = execution_ns

            if self.transport == "zenoh":
                ptopic = f"{execution_ns}/exec_command"
                stopic = f"{execution_ns}/exec_response"
                ftopic = f"{execution_ns}/exec_feedback"
                self.node.get_logger().info(f'RemoteExecutor exec_response_sub: {stopic}')
                self.node.get_logger().info(f'RemoteExecutor exec_command_pub: {ptopic}')
                self.node.get_logger().info(f'RemoteExecutor exec_feedback_sub: {ftopic}')
                self.exec_response_sub = self.node.create_subscription(String, stopic, self.exec_response_callback, 5, callback_group=self.node.group)
                self.exec_command_pub = self.node.create_publisher(String, ptopic, 5, callback_group=self.node.group)
                self.exec_feedback_sub = self.node.create_subscription(String, ftopic, self.exec_feedback_callback, 5, callback_group=self.node.group)                


            #
            # Wait for pong
            #
            
            self.pong_uuid = self.ping(execution_ns)

            if not self.pong_uuid:
                self.node.get_logger().error(f'FAILED sending ping: {self.id}')
                self.fail(f'FAILED sending ping: {self.id}')
                return
            
            self.counter = 600  # was 30
            
            #while not self.have_pong and counter:
            #    time.sleep(0.1)
            #    counter -= 1
            #if counter == 0:
            #    self.fail(f"{self.id}: Did not get a response to ping")
            #    return
            

            #if execution_ns in ["/dji0", "/dji1", "/dji2", "/dji3", "/dji4", "/dji5"]:
            #    self.mqtt_handler.prefix = f'waraps/unit/air/real'
            #    self.mqtt_handler.unit = execution_ns.lstrip("/")
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION do_work_begin: {type(exc)} - {exc}')            
            
    def do_work(self):
        try:
            ##self.node.get_logger().info(f'****************RemoteExecutor do_work {self.id}: {self.tni.type} - {self.have_pong} - {self.have_started} - {self.finished} - {self.counter}')

            # Check pong_response

            if not self.have_pong:
                if self.pong_uuid in self.pong_response and self.pong_response[self.pong_uuid]:
                    self.have_pong = True

            if not self.have_pong:
                self.node.get_logger().info(f'****************RemoteExecutor do_work {self.id}: {self.tni.type} - {self.have_pong} - {self.have_started} - {self.finished} - {self.counter}')
                if self.counter > 0:
                    self.counter -= 1
                    return
                else:
                    self.fail(f"{self.id}: Did not get a response to ping")
                    self.do_work_finished()
                    return

            ##self.node.get_logger().info(f'****************RemoteExecutor do_work {self.id}: HAVE PONG')
            
            if not self.have_started:
                if self.use_tst_execution:
                    self.jobj["node_uuid"] = str(uuid.uuid4()) # should this be done???
                    self.root_uuid = str(uuid.uuid4())
                    treeobj = json_basic_seq(self.execution_ns, node_uuid=self.root_uuid)
                    treeobj["children"].append(self.jobj)

                    treestr = json.dumps(treeobj, sort_keys=True, indent=4, separators=(',', ': '))
                    ## self.node.get_logger().info(f'TREESTR: {treestr}')
                    
                    # com_uuid = self.mqtt_handler.start_tst(treeobj)
                    self.com_uuid = self.start_tst(treeobj, self.execution_ns.lstrip("/"))
                else:
                    name = self.jobj["name"]
                    params = self.jobj["params"]
                    params.pop("use_lock", None)
                    task = {"name": name,
                            "params": params}
                    (self.com_uuid, self.task_uuid) = self.start_task(task, self.execution_ns.lstrip("/"))
                    
                if not self.com_uuid:
                    self.node.get_logger().error(f'com_uuid empty or undefined')
                    self.fail(f"{self.id}: com_uuid empty or undefined")
                    self.do_work_finished()                                        
                    return

                self.have_started = True
                return

            ##self.node.get_logger().info(f'****************RemoteExecutor do_work {self.id}: HAVE STARTED')               


            if not self.have_response and self.com_uuid and self.com_uuid in self.response:
                jobj = self.response[self.com_uuid]
                msgstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                # self.node.get_logger().error(f'have response: {msgstr}')
                self.have_response = True
                if self.use_tst_execution:
                    if jobj["response"] != "started":
                        fail_reason = jobj["fail-reason"]
                        self.fail(f"{self.id}: start tst failed: {fail_reason}")
                        self.do_work_finished()
                        return
                else:
                    if jobj["response"] != "running" and jobj["response"] != "finished":
                        fail_reason = jobj["fail-reason"]
                        self.fail(f"{self.id}: start task failed: {fail_reason}")
                        self.do_work_finished()                                            
                        return
                    if jobj["response"] == "finished":
                        self.do_work_finished()
                        

            if self.root_uuid in self.feedback:
                jobj = self.feedback[self.root_uuid]
                msgstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                ##self.node.get_logger().info(f'have feedback {self.id}: {msgstr}')
                    
                self.status = jobj["status"]
                if self.old_status != self.status:
                    self.node.get_logger().info(f"STATUS {self.id}: {self.old_status} -> {self.status}")
                    self.old_status = self.status
                    
                if self.status == "succeeded":
                    self.do_work_finished()
                    
                if self.status == "failed":
                    self.fail(jobj["fail-reason"])
                    self.do_work_finished()
                
                if self.status == "aborted":
                    self.abort_fail(jobj["fail-reason"])
                    self.do_work_finished()                    
                
                if self.status == "revoked":
                    self.do_work_finished()                                        
                
                if self.status == "running":
                    pass
                
                if self.status == "paused":
                    pass
                
                #if self.have_signal():
                #    signal = self.pop_signal()
                #    self.node.get_logger().info(f'RemoteExecutor got signal: {signal}')
                #if signal == "$finished":
                #    self.signal_finish()
                #    finished_flag = True                        
                #if signal == "$enough":
                #    finished_flag = True
                #if signal == "$abort":
                #    self.abort_fail("Got abort signal")
                #    finished_flag = True
                #if signal == "$pause":
                #    paused = True
                #if signal == "$continue":
                #    self.dji.mission_continue()                        
                #    paused = False


            if self.task_uuid in self.feedback and self.feedback[self.task_uuid]:
                jobj = self.feedback[self.task_uuid]
                self.feedback[self.task_uuid] = None
                msgstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                ## self.node.get_logger().info(f'have TASK feedback {self.id}: {msgstr}')

                self.status = jobj["status"]
                if self.old_status != self.status:
                    self.node.get_logger().info(f"STATUS {self.id}: {self.old_status} -> {self.status}")
                    self.old_status = self.status
                    
                if self.status == "finished":
                    self.do_work_finished()                                        
                    
                if self.status == "failed":
                    self.fail(jobj["fail-reason"])
                    self.do_work_finished()                                                            
                
                if self.status == "aborted":
                    self.abort_fail(jobj["fail-reason"])
                    self.do_work_finished()                                                            
                
                    if self.status == "running":
                        pass
                
                    if self.status == "paused":
                        pass
                
            if self.have_signal():
                signal = self.pop_signal()
                self.node.get_logger().info(f'RemoteExecutor got signal {self.id}: {signal}')
                if self.root_uuid in self.feedback:
                    pass
                if self.task_uuid in self.feedback:
                    if signal == "$abort":
                        self.node.get_logger().info(f'Send signal-task {signal} to remote: {self.execution_ns}')
                        self.signal_task(self.task_uuid, signal, self.execution_ns)
                    if signal == "$enough":
                        self.node.get_logger().info(f'Send signal-task {signal} to remote: {self.execution_ns}')
                        self.signal_task(self.task_uuid, signal, self.execution_ns)
                    if signal == "$pause":
                        self.node.get_logger().info(f'Send signal-task {signal} to remote: {self.execution_ns}')
                        self.signal_task(self.task_uuid, signal, self.execution_ns)
                    if signal == "$continue":
                        self.node.get_logger().info(f'Send signal-task {signal} to remote: {self.execution_ns}')
                        self.signal_task(self.task_uuid, signal, self.execution_ns)
                        #if signal == "$finished":
                        #    self.signal_finish()
                        #    finished_flag = True                        
                        #if signal == "$enough":
                        #    finished_flag = True
                        #if signal == "$abort":
                        #    self.abort_fail("Got abort signal")
                        #    finished_flag = True
                        #if signal == "$pause":
                        #    paused = True
                        #if signal == "$continue":
                        #    self.dji.mission_continue()                        
                        #    paused = False
                            

        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION do_work {self.id}: {type(exc)} - {exc}')            
            self.fail(f'Exception{self.id}: {exc}')
            self.do_work_finished()                    
        ## self.node.get_logger().info("RemoteExecutor end")
            
class RemoteExecutorFactory(TickedExecutorFactory):
    def __init__(self):
        super().__init__("remote")
        self.transport = "mqtt"
        if os.getenv("TST_REMOTE_TRANSPORT"):
            self.transport = os.getenv("TST_REMOTE_TRANSPORT")

        self.stmtu_client = None
        if self.transport == "mqtt":
            self.stmtu_client = self.create_client(SendTopicMsgToUnit, "send_topic_msg_to_unit", callback_group=self.group)
            self.unit_topic_msg_sub = self.create_subscription(UnitTopicMsg, "/unit_topic_msg", self.unit_topic_msg_callback, 10, callback_group=self.group)


    def unit_topic_msg_callback(self, msg):
        try:
            ###self.get_logger().info(f'=========================================unit_topic_msg_callback: {msg}')            
            for exec in self.executors:
                exec.notify_unit_topic_msg(msg)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION unit_topic_msg_callback: {type(exc)} - {exc}')            
        
    def get_executor(self, node, id):
        return self.add_executor(RemoteExecutor(node, id, stmtu_client=self.stmtu_client))

def main(args=None):
    rclpy.init(args=args)
    
    executor = MultiThreadedExecutor(num_threads=8)    

    node = RemoteExecutorFactory()
    ## node = RemoteExecutorFactory(mqtt_handler=mqtt_node)    
    
    executor.add_node(node)    
    
    print(f"Spinning {__name__}")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
    
