import time
import rclpy
from rclpy.executors import MultiThreadedExecutor

from lrs_exec.executor import Executor
from lrs_exec.executor_factory import ExecutorFactory
from lrs_msgs_tst.msg import TSTNodeStatus

class SuppExecutor(Executor):
    def __init__(self, node, id):
        super().__init__(node, id)
        self.set_signals(["$finished"])
        self.use_ticking = True
        self.uuids = []
        self.uuid_to_id = {}

        # tick part
        self.state = ""
        self.next_child = 0
        self.old_state = ""
        self.inside_tick = False
        self.started_flag = False
        self.finished = {}

    def tick(self):
        if self.inside_tick:
            return
        self.inside_tick = True
        ## self.node.get_logger().error(f'supp {self.id}: TICK: {self.started_flag}')
        try:
            if self.started_flag:
                if self.old_state != self.state:
                    self.node.get_logger().info(f'supp {self.id}: {self.old_state} -> {self.state}')
                    self.old_state = self.state
                if self.state == "start":
                    for uid in self.uuids:
                        if True:
                            resp = self.set_start_executor_flag(self.uuid_to_id[uid], True)
                            if not resp.success:
                                self.node.get_logger().error(f'Could not start executor for child')
                                self.fail("Could not start executor for child")
                                self.state = "finished"
                                self.inside_tick = False
                                return
                            else:
                                self.node.get_logger().info(f'Set start executor flag for: {uid}')
                                self.finished[uid] = False
                    self.state = "waiting"
                    self.inside_tick = False
                    return
                if self.state == "waiting":
                    fin = True
                    for uid in self.uuids:
                        fin &= self.finished[uid]
                    if fin:
                        self.state = "finished"
                    if self.have_signal():
                        signal = self.pop_signal()
                        self.node.get_logger().info(f'supp got signal: {signal}')            
                        if signal == "$finished":
                            for suid in self.uuids:
                                resp = self.send_signal(uid, "$finished")
                    self.inside_tick = False
                    return
                if self.state == "finished":
                    if self.normal_exit():
                        self.wait_for_postwork_conditions()
                    else:
                        unlock_res = self.unlock_resources()
                        exec_unlock_res = True
                        if self.tni.use_execution_lock:
                            exec_unlock_res = self.execution_lock_release(self, tni.root_uuid)
                    self.state = "exit"
                    self.inside_tick = False
                    return
        except Exception as exc:
            self.node.get_logger().error(f'Exception notify_node_status: {type(exc)} - {exc}')
        self.inside_tick = False
        

    def notify_node_status(self, msg):
        try:
            ## self.node.get_logger().error(f'supp node_status: {msg.node_uuid}')
            if msg.node_uuid in self.uuids:
                self.node.get_logger().info(f'***supp node_status {self.id}: {msg.id} - {msg.exec_status}')
                if msg.exec_status.finished:
                    if not msg.exec_status.succeeded:
                        self.fail("Child execution failed")
                    self.node.get_logger().info(f'***supp node_status FINISHED {self.id}: {msg.id}')
                    self.finished[msg.node_uuid] = True
        except Exception as exc:
            self.node.get_logger().error(f'Exception notify_node_status: {type(exc)} - {exc}')
        
        
    def start(self):
        try:
            self.node.get_logger().info(f'SuppExecutor start: {len(self.tni.children)}')
            fail_flag = False
            fail_reason = ""
            self.uuid_to_id = {}
            self.uuids = []
            for cid in self.tni.children:
                val = self.get_uuid_from_id(cid)
                if val:
                    self.uuid_to_id[val] = cid
                    self.uuids.append(val)
                else:
                    self.node.get_logger().info(f'could not get node uuid for child: {cid}')
                    self.fail("Could not get node uuid for a child")
                    self.started_flag = True
                    self.next_child = 0
                    self.state = "finished"
                    return
                
            self.started_flag = True
            self.state = "start"
            return
        
        except Exception as exc:
            self.node.get_logger().error(f'Exception supp: {type(exc)} - {exc}')            
            self.fail(f"supp Exception: {exc}")
        self.lprint("SuppExecutor end")

class SuppExecutorFactory(ExecutorFactory):
    def __init__(self):
        super().__init__("supp")
        self.node_status_sub = self.create_subscription(TSTNodeStatus, "tst_node_status", self.node_status, 10, callback_group=self.group)
        self.executors = []        
        self.tick_timer = self.create_timer(0.2, self.tick, callback_group=self.group)

    def tick(self):
        ## self.get_logger().error(f'supp: FACTICK: {len(self.executors)}')        
        try:
            executors = self.executors
            for obj in executors:
                ## self.get_logger().error(f'supp: OBJTICK: {len(self.executors)}')                        
                obj.tick()
            remove_list = []
            for obj in executors:                
                if obj.state == "exit":
                    remove_list.append(obj)
            self.executors_lock.acquire()                                        
            for obj in remove_list:
                if obj in self.executors:
                    self.get_logger().error(f'Remove from supp executors: {obj.id}')
                    self.executors.remove(obj)
            self.executors_lock.release()                    
        except Exception as exc:
            self.get_logger().error(f'supp fac tick Exception: {type(exc)} - {exc}')
                    

    def node_status(self, msg):
        try:
            # self.get_logger().error(f'node_status: {msg.node_uuid} {msg.id}')
            self.executors_lock.acquire()                    
            for obj in self.executors:
                obj.notify_node_status(msg)
            self.executors_lock.release()                                        
        except Exception as exc:
            self.get_logger().error(f'Exception node_status: {type(exc)} - {exc}')
        
        
    def get_executor(self, node, id):
        obj = SuppExecutor(node, id)
        self.executors_lock.acquire()                            
        self.executors.append(obj)
        self.executors_lock.release()        
        return obj

def main(args=None):
    rclpy.init(args=args)
    
    executor = MultiThreadedExecutor(num_threads=8)    
    
    node = SuppExecutorFactory()
    
    executor.add_node(node)    
    print(f"Spinning {__name__}")
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
    
