import time
import sys
import threading
import rclpy
from rclpy.node import Node

from threading import Lock

from lrs_msgs_common.msg import ResourceInfo, ResourcesInfo
from lrs_srvs_ra.srv import RALockResource, RAUnlockResource

class Resource(Node):

    def __init__(self):
        super().__init__('resource_agent')
        self.verbose = False

        self.declare_parameter("external", False)
        self.external = self.get_parameter('external').get_parameter_value().bool_value
        
        self.lock_resource_srv = self.create_service(RALockResource, 'ra/lock_resource', self.lock_resource)
        self.unlock_resource_srv = self.create_service(RAUnlockResource, 'ra/unlock_resource', self.unlock_resource)
        self.resource_info_publisher = self.create_publisher(ResourcesInfo, f'resources_info', 1)
        self.resource_locks = {}
        self.resource_uuids = {}
        self.lock = Lock()
        timer_period = 0.5  # seconds
        self.timer = self.create_timer(timer_period, self.timer_callback)

    def lprint(self, *val):
        if self.verbose:
            print(val)


    # published 2 hz
    def push_resource_info(self):
        self.lock.acquire(blocking=True)
        ri = ResourcesInfo()
        for resource in self.resource_locks:
            msg = ResourceInfo()
            msg.resource = resource
            msg.locked = self.resource_locks[resource]
            msg.node_uuid = self.resource_uuids[resource]
            ri.resinfo.append(msg)
        self.lock.release()
        ri.ns = self.get_namespace()
        ri.max_time = 3600
        self.resource_info_publisher.publish(ri)

    def timer_callback(self):
        self.push_resource_info()

    def lock_resource(self, request, response):
        try:
            self.lprint("lock_resource:", request)
            self.lock.acquire(blocking=True)
            if request.resource in self.resource_locks:
                if self.resource_locks[request.resource]:
                    response.got_lock = False
                else:
                    self.resource_locks[request.resource] = True
                    self.resource_uuids[request.resource] = request.node_uuid                    
                    response.got_lock = True
            else:
                self.resource_locks[request.resource] = True
                self.resource_uuids[request.resource] = request.node_uuid
                response.got_lock = True
            self.lock.release()
            response.success = True
            response.error = 0
            response.reason = ""
        except Exception as exc:
            print("EXCEPTION:", type(exc))
            print(exc)
            if self.lock.locked():            
                self.lock.release()            
            response.got_lock = False
            response.success = False
            response.error = 1
            response.reason = f'lock_resource Exception: {exc}'
        self.lprint("lock_resource resp:", response)
        return response
        
    def unlock_resource(self, request, response):
        response.success = True
        response.error = 0
        response.reason = ""
        try:
            self.lprint("unlock_resource:", request)
            self.lock.acquire(blocking=True)
            if request.resource == "":
                self.resource_locks.clear()
            else:
                if request.resource in self.resource_locks:
                    self.resource_locks[request.resource] = False
                else:
                    response.success = False
                    response.error = 1
                    response.reason = "unlock - resource does not exist"
            self.lock.release()
        except Exception as exc:
            print("EXCEPTION:", type(exc))
            print(exc)
            if self.lock.locked():
                self.lock.release()            
            response.success = False
            response.error = 2
            response.reason = f'unlock_resource Exception: {exc}'            
        self.lprint("unlock_resource resp:", response)
        return response
        


def main(args=None):
    print('Hi from resource server.')

    rclpy.init(args=args)

    resource_agent = Resource()

    try:
        rclpy.spin(resource_agent)
    except KeyboardInterrupt:
        pass

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    resource_agent.destroy_node()
    rclpy.shutdown()
    

if __name__ == '__main__':
    main()
    
