#
# ros2_bridge:
#     Service call to send_topic_msg_to_unit:   push on topic unit_topic_msg
#     unit_topic_msg_callback: if on right unit: translate to tst_command, tst_response and tst_feedback
# 
#
#
# remote executor: sub unit_topic_msg : save response and feedback from everybody
#                  send.topic_msg_unit client : used for start_tst
#
#
#

import os
import time
import sys
import uuid
import json
import threading
from threading import Lock

import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup

from lrs_msgs_common.srv import SendTopicMsgToUnit
from lrs_msgs_common.msg import UnitTopicMsg
from std_msgs.msg import String

class Ros2Bridge(Node):
    def __init__(self):
        super().__init__('ros2_bridge')
        self.group = ReentrantCallbackGroup()

        self.stmtu_srv = self.create_service(SendTopicMsgToUnit, "send_topic_msg_to_unit", self.send_topic_msg_to_unit, callback_group=self.group)
        self.unit_topic_msg_pub = self.create_publisher(UnitTopicMsg, f"/unit_topic_msg", 10, callback_group=self.group)

        self.tst_command_pub = self.create_publisher(String, "tst_command", 5, callback_group=self.group)
        self.tst_response_pub = self.create_publisher(String, "tst_response", 5, callback_group=self.group)
        self.tst_feedback_pub = self.create_publisher(String, "tst_feedback", 5, callback_group=self.group)


        self.unit_topic_msg_sub = self.create_subscription(UnitTopicMsg, "/unit_topic_msg", self.unit_topic_msg_callback, 10, callback_group=self.group)
        self.ns = self.get_namespace()
        self.unit = self.ns.lstrip("/")


    def unit_topic_msg_callback(self, msg):
        try:
            if msg.unit != self.unit:
                return
            msg2 = String()
            msg2.data = msg.msg
            self.get_logger().info(f'unit_topic_msg_callback: {msg.unit} - {msg.topic}')               
            if msg.topic == "tst/command":
                self.tst_command_pub.publish(msg2)
            if msg.topic == "tst/response":
                self.tst_response_pub.publish(msg2)
            if msg.topic == "tst/feedback":
                self.tst_feedback_pub.publish(msg2)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION unit_topic_msg_callback: {type(exc)} - {exc}')               
        

    def send_topic_msg_to_unit(self, request, response):
        try:
            msg = UnitTopicMsg()
            msg.unit = request.unit
            msg.topic = request.topic
            msg.msg = request.msg
            self.get_logger().info(f'send_topic_msg_to_unit: {msg.unit} - {msg.topic}')                           
            self.unit_topic_msg_pub.publish(msg)
            response.success = True            
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_topic_msg_to_unit: {type(exc)} - {exc}')               
            response.success = False
        return response

    
def main(args=None):
    print('Hi from ROS2 bridge server.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=8)
    node = Ros2Bridge()
    executor.add_node(node)

    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    # rclpy.spin(node)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    #node.destroy_node()
    #rclpy.shutdown()
    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
    
