import os
import functools
import time
import sys
import uuid
import json
import threading
import copy
from threading import Lock
#import geopy.distance
from pyproj import Proj

import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup

from rclpy.duration import Duration
from lrs_util.coordtrans import CoordTrans
from lrs_msgs_common.msg import PackageInfos, PackageInfo, PackageConsumed, DropStatus, WarapsSalientPointArray, SpotStatus, WarapsSalientPoint
from geographic_msgs.msg import GeoPoint

from lrs_team_msgs.msg import TeamManagerInfo
from lrs_team_msgs.srv import AddTask, SetTaskPriority, SignalTask
from lrs_util.json_exec_node import JsonExecNode

class ScenarioHandler(Node):

    def add_serv(self, type, srvname, callback):
        srv = self.create_service(type, f'{srvname}', callback, callback_group=self.group)
        self.srvs.append(srv)
    
    def __init__(self):
        super().__init__('scenario_handler')
        self.group = ReentrantCallbackGroup()

        self.get_logger().error(f'Scenario_Handler Init')

        try:
            self.declare_parameter("location", "granso")
            self.location = self.get_parameter('location').get_parameter_value().string_value
            self.get_logger().info(f'location: {self.location}')
            self.coordtrans = CoordTrans(location=self.location)

            self.declare_parameter("fake", False)
            self.fake = self.get_parameter('fake').get_parameter_value().bool_value

            self.declare_parameter("drop", False)
            self.drop_flag = self.get_parameter('drop').get_parameter_value().bool_value

            self.declare_parameter("search", False)
            self.search_flag = self.get_parameter('search').get_parameter_value().bool_value
            
            self.real_salient_points = []

            self.salient_points = {}
            self.handled_salient_points = {}
            self.packages = {}
            self.have_dropped = True
            self.have_added_search = True
            self.n_packages_in_bucket = {}
            self.n_object_detections = 0

            self.search_start_time = None

            self.persons_added = 0

            self.task_fail_time = {}
            self.task_requests = {}

            self.add_real_salient_point(0, 0, 0)

            # self.package_infos_pub = self.create_publisher(PackageInfos, f'sensor/package_infos', 10)

            self.add_task_client = self.create_client(AddTask, "team/add_task", callback_group=self.group)        

            self.spot_status_sub = self.create_subscription(SpotStatus, "global/spot_status", self.spot_status_callback, 10, callback_group=self.group)
            self.team_manager_info_sub = self.create_subscription(TeamManagerInfo, "/team_manager_info", self.team_manager_info_callback, 10, callback_group=self.group)
            self.package_infos_sub = self.create_subscription(PackageInfos, "sensor/package_infos", self.package_infos_callback, 10, callback_group=self.group)

            self.object_detections_sub = self.create_subscription(WarapsSalientPointArray, "global/object_detections", self.object_detections_callback, 10, callback_group=self.group)            

            
            self.period = 0.4
            self.slow_period = 3.0
            #self.timer = self.create_timer(self.period, self.timer_callback, callback_group=self.group) 
            self.slow_timer = self.create_timer(self.slow_period, self.slow_timer_callback, callback_group=self.group)

        except Exception as exc:
            self.get_logger().error(f'EXCEPTION scenario_handler: {type(exc)} - {exc}')


    def add_real_salient_point(self, x, y, z):
        gp = GeoPoint()
        (gp.longitude, gp.latitude, gp.altitude) = self.coordtrans.world_to_wgs84(x, y, z)
        self.real_salient_points.append(gp)

    def add_task(self, team_name, task_obj, priority=50, name="dummyname", task_type="json-tst"):
        try:
            print("ADD_TASK:", team_name, priority, name, task_obj)
            self.add_task_client.wait_for_service()
            print("GOT add_task_client service")
            req = AddTask.Request()
            req.team_name = team_name
            req.task.uuid = str(uuid.uuid4())
            req.task.priority = priority
            req.task.type = task_type
            req.task.task = json.dumps(task_obj, sort_keys=True, indent=4, separators=(',', ': '))
            req.task.name = name
            self.task_requests[req.task.uuid] = req
            self.add_task_client.call_async(req)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION: {type(exc)} - {exc}')
            

    def add_task_req(self, req):
        try:
            print("ADD_TASK_REQ:", req)
            self.add_task_client.wait_for_service()
            print("GOT add_task_client service")
            self.task_requests[req.task.uuid] = req
            self.add_task_client.call_async(req)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION: {type(exc)} - {exc}')
            

    def get_geopoint(self, x, y, z):
        gp = GeoPoint()
        (gp.longitude, gp.latitude, gp.altitude) = self.coordtrans.world_to_wgs84(x, y, z)
        return gp

    def add_to_spot_queue(self, sp):
        try:
            print("ADD TO SPOT QUEUE:", sp)
            lat = sp.geopoint.latitude
            lon = sp.geopoint.longitude
            alt = sp.geopoint.altitude
            dobj = JsonExecNode("", name="delivery-goal", lat=lat, lon=lon, alt=alt)
            tree = JsonExecNode("", name="seq", node_uuid=str(uuid.uuid4()))
            tree.add_child(dobj)
            jobj = tree.get_json()
        
            team_name = "deliver"
            jobj["common_params"]["execunit"] = ""
            jobj["children"][0]["common_params"]["execunit"] = ""
            print("JSON-TASK queue_task:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))

            self.add_task(team_name, jobj, name="deliver-medicine", task_type="json-tst")
            
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION: {type(exc)} - {exc}')

    def add_to_drop_queue(self, x, y, z):
        try:
            print("ADD TO DROP QUEUE:")
            (lon, lat, alt) = self.coordtrans.world_to_wgs84(x, y, z)
            (lon1, lat1, alt1) = self.coordtrans.world_to_wgs84(25.0, -4.0, 0.0)
            obj1 = JsonExecNode("", name="move-to", lat=lat, lon=lon, alt=alt, speed="slow")
            obj11 = JsonExecNode("", name="fly-to", lat=lat, lon=lon, alt=36.0, speed="slow")
            obj2 = JsonExecNode("", name="drop")
            obj3 = JsonExecNode("", name="add-pkg", lat=lat1, lon=lon1, alt=alt1, package_type="medicine")
            obj4 = JsonExecNode("", name="go-home", speed="slow")
            tree = JsonExecNode("", name="seq", node_uuid=str(uuid.uuid4()))
            tree.add_child(obj1)
            tree.add_child(obj11)
            tree.add_child(obj2)
            tree.add_child(obj3)
            tree.add_child(obj4)
            jobj = tree.get_json()
        
            team_name = "drop"
            jobj["common_params"]["execunit"] = ""
            jobj["children"][0]["common_params"]["execunit"] = ""
            print("JSON-TASK queue_task:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))

            self.add_task(team_name, jobj, name="drop-medicine", task_type="json-tst")
            self.have_dropped = True
            
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION: {type(exc)} - {exc}')

    def check_fake(self):
        if self.search_start_time:
            ## print("check_fake")
            t0 = self.get_clock().now()
            dur = t0 - self.search_start_time
            if dur > Duration(seconds=5):
                if self.persons_added == 0:
                    print("ADD FAKE OBSERVATION 1")
                    self.persons_added +=1
                    # self.add_salient_point(-1.3, 1.2, 0.0)
                    #self.add_salient_point(41.0, 17.0, 0.0)
            if dur > Duration(seconds=20):
                if self.persons_added == 1:
                    print("ADD FAKE OBSERVATION 1")
                    self.persons_added +=1
                    # self.add_salient_point(-1.5, -2.6, 0.0)
                    self.add_salient_point(29.0, 6.0, 0.0)
            if dur > Duration(seconds=25):
                if self.persons_added == 2:
                    print("ADD FAKE OBSERVATION 1")
                    self.persons_added +=1
                    # self.add_salient_point(-1.5, -12.6, 0.0)
                    self.add_salient_point(34.0, 23.0, 0.0)
        
    def slow_timer_callback(self):
        # print("slow_timer_callback")

        if self.fake:
            self.check_fake()

        ## self.check_restart_failed()
        
        n_medicine_pkgs = -1
        if "medicine" in self.packages:
            n_medicine_pkgs = len(self.packages["medicine"])
        print("n_medicine_pkgs:", n_medicine_pkgs)
        if n_medicine_pkgs == 0 and not self.have_dropped:
            self.add_to_drop_queue(26.0, 1.0, 20)
        for k, v in self.salient_points.items():
            if k not in self.handled_salient_points:
                if v.type == "person":
                    #if n_medicine_pkgs == 0:
                    #    return
                    self.add_to_spot_queue(v)
                    self.handled_salient_points[k] = v
                        # print("KEY:", k, v)
                else:
                    self.handled_salient_points[k] = v
                break
    
    def timer_callback(self):
        print("timer_callback")
        try:
            pass
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION timer_callback: {type(exc)} - {exc}')

    def spot_status_callback(self, msg):
        # print("spot_status_callback:", msg)
        for i in range(len(msg.package_types)):
            self.n_packages_in_bucket[msg.package_types[i]] = msg.number_of_packages_in_body_bucket[i]



#  "name": "search-area",                                                      
#    "params": {                                                                 
#        "area": [                                                               
#            {                                                                   
#                "altitude": 58,                                                 
#                "latitude": 57.760412547494035,                                 
#                "longitude": 16.68256568978499,                                 
#                "rostype": "GeoPoint"                                           
#            },                                                                  
#            {                                                                   
#                "altitude": 58,                                                 
#                "latitude": 57.760401097175794,                                 
#                "longitude": 16.682585502904335,                                
#                "rostype": "GeoPoint"                                           
#            },                                                                  
#            {                                                                   
#                "altitude": 58,                                                 
#                "latitude": 57.76068030775089,                                  
#                "longitude": 16.68314852572718,                                 
#                "rostype": "GeoPoint"                                           
#            },                                                                  
#            {                                                                   
#                "altitude": 58,                                                 
#                "latitude": 57.76068465130913,                                  
#                "longitude": 16.683144271373752,                                
#                "rostype": "GeoPoint"                                           
#            }                                                                   
#        ],                                                                      
#        "speed": "standard",                              
#



    def add_search_to_queue(self):
        #path = [[-5, -5, 0], [15, 15, 0], [20, -10, 0], [5, -10, 0]]
        #print("AREA:", path)                
        #area = self.world_to_geo_path(path)
        area = [                                                               
            {                                                                   
                "altitude": 58,                                                 
                "latitude": 57.760412547494035,                                 
                "longitude": 16.68256568978499,                                 
                "rostype": "GeoPoint"                                           
            },                                                                  
            {                                                                   
                "altitude": 58,                                                 
                "latitude": 57.760401097175794,                                 
                "longitude": 16.682585502904335,                                
                "rostype": "GeoPoint"                                           
            },                                                                  
            {                                                                   
                "altitude": 58,                                                 
                "latitude": 57.76068030775089,                                  
                "longitude": 16.68314852572718,                                 
                "rostype": "GeoPoint"                                           
            },                                                                  
            {                                                                   
                "altitude": 58,                                                 
                "latitude": 57.76068465130913,                                  
                "longitude": 16.683144271373752,                                
                "rostype": "GeoPoint"                                           
            }                                                                   
        ]
        print("AREA:", area)
        sa = JsonExecNode("", name="basic-search-area", area=area, speed="slow", target_type="person", target_size=2, area_type="field")
        tree = JsonExecNode("", name="seq", node_uuid=str(uuid.uuid4()))
        tree.add_child(sa)
        print("SEARCH AREA TREE:", json.dumps(tree.get_json(), sort_keys=True, indent=4, separators=(',', ': ')))                    
        jobj = tree.get_json()
        
        team_name = "search"
        jobj["common_params"]["execunit"] = ""
        jobj["children"][0]["common_params"]["execunit"] = ""
        print("JSON-TASK queue_task:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))

        self.add_task(team_name, jobj, name="search-area", task_type="json-tst")
        self.have_added_search = True

    def world_to_geo_path(self, path):
        try:
            geopath = []
            for el in path:
                # resp = self.world_to_geo(el[0], el[1], el[2])
                (lon, lat, alt) = self.coordtrans.world_to_wgs84(el[0], el[1], el[2])
                geopath.append({"altitude": alt,
                                "latitude": lat,
                                "longitude": lon,
                                "rostype": "GeoPoint"})
            return geopath
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION world_to_geo_path: {type(exc)} - {exc}')
        return None
        
    def team_manager_info_callback(self, msg):
        # print("team_manager_info_callback:", self.n_packages_in_bucket)
        if "medicine" in self.n_packages_in_bucket:
            n_packages = self.n_packages_in_bucket["medicine"]
            if msg.team_name == "deliver" and n_packages > 0 and len(msg.tasks) == 0 and not self.have_added_search:
                # We failed to find body and have package in bucket
                # add search mission
                print("ADD SEARCH MISSION")
                self.have_added_search = True
                self.add_search_to_queue()
                self.search_start_time = self.get_clock().now()
        # print("team_manager_info_callback:", msg)
        if msg.team_name == "deliver":
            for task in msg.tasks:
                if task.uuid not in self.task_fail_time:
                    self.task_fail_time[task.uuid] = None
                if task.state == 12:
                    # success
                    print("TASK SUCCESS:", task.uuid)
                    self.task_fail_time[task.uuid] = None
                    self.task_requests[task.uuid] = None
                if task.state == 11:
                    # aborted
                    if not self.task_fail_time[task.uuid]:
                        print("TASK ABORTED:", task.uuid)
                        self.task_fail_time[task.uuid] = self.get_clock().now()
                if task.state == 10:
                    # failed
                    if not self.task_fail_time[task.uuid]:
                        print("TASK FAILED:", task.uuid)
                        self.task_fail_time[task.uuid] = self.get_clock().now()


    def check_restart_failed(self):
        try:
            print("check_restart_failed")
            for k,v in self.task_fail_time.items():
                if v and k in self.task_requests and self.task_requests[k]:
                    t0 = self.get_clock().now()
                    dur = t0 - v
                    print("KEY:", k, v)
                    print("DURATION:", dur, Duration(seconds=5))
                    if dur > Duration(seconds=5) and self.n_medicine_pkgs > 0:
                    # if dur > Duration(seconds=5):
                        print("RE-ADD TASK TO QUEUE", self.task_requests[k])
                        self.task_fail_time[k] = None
                        req = copy.deepcopy(self.task_requests[k])
                        jobj = json.loads(req.task.task)
                        jobj["common_params"]["node-uuid"] = str(uuid.uuid4())
                        req.task.task = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                        req.task.uuid = str(uuid.uuid4())
                        self.add_task_req(req)
                        self.task_requests[k] = None
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION check_restart_failed: {type(exc)} - {exc}')


    def package_infos_callback(self, msg):
        #print("package_infos_callback:", msg)
        for el in msg.package_infos:
            #print("EL:", el)
            self.packages[el.package_type] = el.geopoints
        if "medicine" in self.packages:
            self.n_medicine_pkgs = len(self.packages["medicine"])

    def approve_salient_point(self, sp):
        return True

    def object_detections_callback(self, msg):
        self.n_object_detections += 1
        if self.n_object_detections < 12:
            return
        for sp in msg.points:
            # print("object_detections_callback:", sp)
            if self.approve_salient_point(sp):
                self.salient_points[sp.uuid] = sp

    def add_salient_point(self, x, y, z):
        print("add_salient_point:", x, y, z)
        sp = WarapsSalientPoint()
        sp.source = 3
        sp.uuid = str(uuid.uuid4())
        sp.type = "person"
        sp.confidence = 1.0
        (lon, lat, alt) = self.coordtrans.world_to_wgs84(x, y, z)
        sp.geopoint.latitude = lat
        sp.geopoint.longitude = lon
        sp.geopoint.altitude = alt
        self.salient_points[sp.uuid] = sp

def main(args=None):
    print('Hi from wdb server.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=8)
    node = ScenarioHandler()
    executor.add_node(node)

    #node.add_search_to_queue()
    #return

    if node.search_flag:
        node.add_search_to_queue ()

    if node.drop_flag:
        node.add_to_drop_queue(26.0, 1.0, 20)

    try:
        print('Hi from scenario handler. SPINNING!')
        executor.spin()
    except KeyboardInterrupt:
        pass

    # rclpy.spin(node)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    #node.destroy_node()
    #rclpy.shutdown()
    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()

