import time
import sys
import uuid
import json
import threading
from threading import Lock
from timeit import default_timer as timer
from collections import deque

import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup

from lrs_msgs_common.msg import DisplayRequest
from lrs_srvs_tst.srv import TSTCreateTree, TSTGetTreeStringFromUuid, TSTSetFlag, TSTGetExecInfo, TSTGetNodeFromUuid
from lrs_srvs_exec.srv import TSTExecute, TSTCreateExecutor, TSTStartExecutor, TSTExecutorJsonExpand
from lrs_exec.parameters import Parameters
from lrs_util.jsonutil import get_exec_unit

class TreeExecutor(Node):

    def __init__(self):
        try:
            super().__init__('tree_executor')
            self.group = ReentrantCallbackGroup()
            self.srv = self.create_service(TSTExecute, f'tst_execute', self.tst_execute, callback_group=self.group)

            self.display = True
            self.expandable_types = []
        
            self.create_tree_client = self.create_client(TSTCreateTree, "exec/tstfactory/create_tree", callback_group=self.group)
            self.get_tree_string_client = self.create_client(TSTGetTreeStringFromUuid, f'exec/tstfactory/get_tree_string_from_uuid', callback_group=self.group)
            self.set_executing_flag_client = self.create_client(TSTSetFlag, "exec/tstfactory/set_executing_flag", callback_group=self.group)
            self.set_waiting_flag_client = self.create_client(TSTSetFlag, "exec/tstfactory/set_waiting_flag", callback_group=self.group)            
            self.set_succeeded_flag_client = self.create_client(TSTSetFlag, "exec/tstfactory/set_succeeded_flag", callback_group=self.group)
            self.set_finished_flag_client = self.create_client(TSTSetFlag, "exec/tstfactory/set_finished_flag", callback_group=self.group)        
            self.get_exec_info_client = self.create_client(TSTGetExecInfo, "exec/tstfactory/get_exec_info", callback_group=self.group)
        
            self.disp_pub = self.create_publisher(DisplayRequest, f"tst_display_request", 10, callback_group=self.group)

            self.expand_clients = {}
            self.create_clients = {}
            self.start_clients = {}
            
            self.tree_objects = []
            self.tree_strings = []
            self.tree_channels = []
            self.tree_strings_index = 0
            self.have_tree = {}
        
            self.inside_timer = False
            self.inside_timer2 = False
            
            #        self.start_sub = self.create_subscription(TSTStartExecutor, "/tst_start_executor", self.start_executor_callback, 10, callback_group=self.group)
            #        self.signal_sub = self.create_subscription(TSTSignalExecutor, "tst_signal_executor", self.signal_executor_callback, 10, callback_group=self.group)
            self.get_node_from_uuid_client = self.create_client(TSTGetNodeFromUuid, "exec/tstfactory/get_node_from_uuid", callback_group=self.group)        
            self.timer = self.create_timer(0.1, self.timer_callback, callback_group=self.group)
            self.timer2 = self.create_timer(0.1, self.timer2_callback, callback_group=self.group)
        except Exception as exc:
            self.get_logger().error(f'Exception constructor: {type(exc)} - {exc}')            

    def wait_future(self, future, print_waiting=False):
        try:
            while True:
                if future.done():
                    resp = future.result()
                    break
            return resp
        except Exception as exc:
            self.get_logger().error(f'Exception wait_future: {type(exc)} - {exc}')

            
    def is_expandable(self, jobj):
        try:
            res = False
            jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
            # self.get_logger().error(f'is_expandable: {jobjstr}')
            execns = get_exec_unit(jobj)
            ns = self.get_namespace()
            # self.get_logger().error(f'is_expandable execns: {execns}')            
            ntype = jobj["name"]
            # self.get_logger().error(f'is_expandable ntype: {ntype}')            
            xtype = ntype.replace("-","_")
            # self.get_logger().error(f'is_expandable xtype: {xtype}')            
            if xtype in self.expandable_types and execns == ns:
                res = True
            return res
        except Exception as exc:
            self.get_logger().error(f'Exception is_expandable: {type(exc)} - {exc}')            

    def get_node_from_uuid(self, nuuid):
        try:
            # self.lprint("get_node_from_uuid:", nuuid)
            req = TSTGetNodeFromUuid.Request()
            req.node_uuid = nuuid
            future = self.get_node_from_uuid_client.call_async(req)
            resp = self.wait_future(future)        
            # self.lprint("get_node_from_uuid resp:", resp)
            return resp
        except Exception as exc:
            self.get_logger().error(f'Exception is_expandable: {type(exc)} - {exc}')            
    
    def get_exec_info(self, nuuid):
        # print("get_exec_info:", nuuid)
        req = TSTGetExecInfo.Request()
        req.node_uuid = nuuid
        future = self.get_exec_info_client.call_async(req)
        resp = self.wait_future(future)        
        # print("get_exec_info resp:", resp)
        return resp

    def create_executor(self, node_type, node_id, prepare):
        self.get_logger().info(f'tree_executor create_executor: {node_type} - {node_id} - {prepare}')
        ntype = node_type.replace("-", "_")        
        try:
            timeout = 0.2
            if ntype not in self.create_clients:
                client = self.create_client(TSTCreateExecutor, f'tst_executor/{ntype}/create_executor', callback_group=self.group)
                self.create_clients[ntype] = client
                timeout = 5.0
                self.get_logger().info(f'tree_executorr create_executor: client created: {node_type} - {node_id} - {prepare}')
            if not self.create_clients[ntype].wait_for_service(timeout_sec=timeout):
                self.get_logger().error(f'tree_executor create_executor wait for service FAILED: {node_type} - {node_id} - {prepare} - {ntype} - {timeout}')
                return False
            req = TSTCreateExecutor.Request()
            req.id = node_id
            req.run_prepare = prepare
            self.get_logger().info(f'tree_executor call create_executor: {req}')
            future = self.create_clients[ntype].call_async(req)
            i = 0
            while not future.done():
                time.sleep(0.1)
                i +=1
                #if i>10:
                #    i = 0
                #    self.get_logger().error(f'tst_exec_manager call create_executor AGAIN: {req}')
                #    future = self.create_clients[ntype].call_async(req)
            resp = future.result()
            #self.get_logger().info(f'tst_exec_manager call create_executor: have future')
            #resp = self.wait_future(future)
            self.get_logger().info(f'tree_executor create_executor resp: {resp}')            
            return resp.success
            #return True
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')            
        return False
    
    def start_executor(self, node_type, node_id):
        self.get_logger().info(f'tree_executor start_executor: {node_type} - {node_id}')                    
        ntype = node_type.replace("-", "_")
        try:
            timeout = 0.1
            if not ntype in self.start_clients:
                client = self.create_client(TSTStartExecutor, f'tst_executor/{ntype}/start_executor', callback_group=self.group)
                self.start_clients[ntype] = client
                timeout = 1.0
            if not self.start_clients[ntype].wait_for_service(timeout_sec=timeout):
                return False
            req = TSTStartExecutor.Request()
            req.id = node_id
            self.get_logger().info(f'start_executor: client call: {req}')                    
            future = self.start_clients[ntype].call_async(req)
            i = 0
            while not future.done():
                time.sleep(0.1)
                #i +=1
                #if i>10:
                #    i = 0
                #    self.get_logger().error(f'tst_exec_manager call start_executor AGAIN: {req}')
                #    future = self.start_clients[ntype].call_async(req)
            resp = future.result()                    
            #resp = self.wait_future(future)
            #self.get_logger().info(f'start_executor: client call have future: {resp} - {future}')                             #self.get_logger().info(f'start_executor resp: {node_type} - {node_id} - {resp}')                    
            return resp.success
            # return True
        except Exception as exc:
            self.get_logger().error(f'Exception start_executor: {type(exc)} - {exc}')            
        return False
    

    def set_executing_flag(self, id, flag):
        future = self.set_executing_flag_client.call_async(self.get_set_flag_request(id, flag))
        resp = self.wait_future(future)
        return resp

    def set_waiting_flag(self, id, flag):
        future = self.set_waiting_flag_client.call_async(self.get_set_flag_request(id, flag))
        resp = self.wait_future(future)
        return resp

    def set_succeeded_flag(self, id, flag):
        future = self.set_succeeded_flag_client.call_async(self.get_set_flag_request(id, flag))
        resp = self.wait_future(future)
        return resp

    def set_finished_flag(self, id, flag):
        future = self.set_finished_flag_client.call_async(self.get_set_flag_request(id, flag))
        resp = self.wait_future(future)
        self.finished = True
        return resp
    
    def get_set_flag_request(self, id, flag):
        req = TSTSetFlag.Request()
        req.id = id
        req.val = flag
        return req

    def get_tree_string(self, uuid, format):
        # self.get_logger().info(f'get_tree_string: {uuid} {format}')
        try:
            req = TSTGetTreeStringFromUuid.Request()
            req.uuid = uuid
            req.type = format
            future = self.get_tree_string_client.call_async(req)
            resp = self.wait_future(future, print_waiting=False)
            # self.get_logger().info(f'get_tree_string resp: {resp}')
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION get_tree_string: {type(exc)} - {exc}')
        return None
    
    def create_tree(self, jobjstr):
        try:
            self.create_tree_client.wait_for_service()
            req = TSTCreateTree.Request()
            req.json = jobjstr
            req.free_id = 1
            req.approve_flag = True
            future = self.create_tree_client.call_async(req)
            resp = self.wait_future(future, print_waiting=True)
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION create_tree: {type(exc)} - {exc}')
        return None

    def is_seq(self, jobj):
        try:
            if self.is_expandable(jobj):
                return True
            if jobj["name"] == "seq":
                return True
            return False
        except Exception as exc:
            self.get_logger().error(f'Exception is_seq: {type(exc)} - {exc}')            

    def traverse_update(self, jobj):
        print("U", end="", flush=True)
        try:
            # print("HANDLE_TREE:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
            ### print("TRAVERSE UPDATE:", jobj["id"])
            
            if "status" in jobj:
                if jobj["status"] in ["success", "fail"]:
                    return

            ## print("TRAVERSE UPDATE DOING CHECK:", jobj["id"])                
            if len(jobj["children"]) > 0:
                success = True
                fail = False
                if self.is_seq(jobj):
                    # self.get_logger().info(f'traverse_update: {jobj}')
                    ### self.get_logger().info(f'traverse_update SEQ: {jobj["name"]}')                        
                    for cobj in jobj["children"]:
                        self.traverse_update(cobj)
                        if not "status" in cobj:
                            success = False
                            break
                        ## print("CHILD STATUS:", jobj["id"], cobj["status"])              
                        if cobj["status"] == "active":
                            success = False
                            fail = False
                            break
                        if cobj["status"] == "fail":                            
                            success = False
                            fail = True
                            break
                    ## self.get_logger().info(f'traverse_update SEQ: {jobj["name"]} - {success}')                        
                if jobj["name"] == "conc":
                    # self.get_logger().info(f'traverse_update: {jobj}')                        
                    for cobj in jobj["children"]:
                        self.traverse_update(cobj)

                    all_children_finished = True
                    for cobj in jobj["children"]:                        
                        if not "status" in cobj:
                            success = False
                            all_children_finished = False
                        if "status" in cobj and cobj["status"] == "active":                            
                            success = False
                            fail = False
                            all_children_finished = False
                        if "status" in cobj and cobj["status"] == "fail":                            
                            success = False
                            fail = True
                    if not all_children_finished:
                        fail = False
                        
                if jobj["name"] == "test-if":
                    test_obj = jobj["children"][0]
                    yes_obj = jobj["children"][1]
                    no_obj = jobj["children"][2]
                    unknown_obj = jobj["children"][3]

                    if not "status" in test_obj:
                        success = False
                    else:
                        success = False
                        if test_obj["status"] == "fail":
                            fail = True
                        else:
                            if "status" in yes_obj:
                                #print("STATUS yes_obj:", yes_obj["status"])
                                if yes_obj["status"] not in ["success", "fail"]:
                                    self.traverse_update(yes_obj)
                                if yes_obj["status"] == "success":
                                    success = True
                                if yes_obj["status"] == "fail":
                                    fail = True
                            if "status" in no_obj:
                                if no_obj["status"] not in ["success", "fail"]:
                                    self.traverse_update(no_obj)
                                if no_obj["status"] == "success":
                                    success = True
                                if  no_obj["status"] == "fail":
                                    fail = True
                            if "status" in unknown_obj:
                                if unknown_obj["status"] not in ["success", "fail"]:
                                    self.traverse_update(unknown_obj)
                                if unknown_obj["status"] == "success":
                                    success = True
                                if unknown_obj["status"] == "fail":
                                    fail = True

                ## print("TRAVUPDATE:", jobj["id"], success, jobj["status"])
                
                if success and jobj["status"] != "success":
                    jobj["status"] = "success"
                    self.set_succeeded_flag(jobj["id"], True)
                    self.set_finished_flag(jobj["id"], True)
                    self.get_logger().info(f'FINISH non-terminal: {jobj["name"]} - {jobj["id"]}')
                        
                if fail and jobj["status"] != "fail":
                    jobj["status"] = "fail"
                    self.set_succeeded_flag(jobj["id"], False)
                    self.set_finished_flag(jobj["id"], True)
                    self.get_logger().info(f'FAIL non-terminal: {jobj["name"]} - {jobj["id"]}')                        
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION traverse_update: {type(exc)} - {exc}')

    def traverse_terminals(self, jobj):
        print("T", end="", flush=True)
        try:
            # print("traverse_terminals:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))

            if len(jobj["children"]) > 0:
                for cobj in jobj["children"]:
                    self.traverse_terminals(cobj)
            else:
                # Add check of status here
                if not "status" in jobj:
                    return
                if jobj["status"] in ["success", "fail"]:
                    return
                print("G", end="", flush=True)
                resp = self.get_exec_info(jobj["node_uuid"])
                print("G", end="", flush=True)                
                if resp.exec_status.finished:
                    if resp.exec_status.succeeded:
                        print("1", end="", flush=True)                
                        self.set_succeeded_flag(jobj["id"], True)
                        print("2", end="", flush=True)                                        
                        self.set_finished_flag(jobj["id"], True)
                        print("3", end="", flush=True)                
                        jobj["status"] = "success"
                        self.get_logger().info(f'FINISH terminal success: {jobj["name"]}')
                    else:
                        print("4", end="", flush=True)                                        
                        self.set_succeeded_flag(jobj["id"], False)
                        print("5", end="", flush=True)                
                        self.set_finished_flag(jobj["id"], True)
                        print("6", end="", flush=True)                
                        jobj["status"] = "fail"
                        self.get_logger().info(f'FINISH terminal fail: {jobj["name"]}')                        
                # self.get_logger().info(f'traverse_terminals resp: {resp}')
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION traverse_tree_start: {type(exc)} - {exc}')


    def traverse_expansion(self, jobj):
        print("E", end="", flush=True)
        try:
            jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
            # self.get_logger().info(f'traverse_expansion: {jobjstr}')
            self.get_logger().info(f'traverse_expansion: START')
        
            if len(jobj["children"]) == 0:
                if "common_params" in jobj:
                    ns = jobj["common_params"]["execunit"]
                else:
                    ns = jobj["execution_ns"]

                if ns != self.get_namespace():
                    # Do not try to expand if execution is on another unit
                    return True
                    
                self.get_logger().info(f'traverse_expansion NS: {ns}')
                if ns == self.get_namespace():
                    ntype = jobj["name"].replace("-","_")
                    self.get_logger().info(f'traverse_expansion TYPE: {ntype}')
                    timeout = 0.1
                    if not ntype in self.expand_clients:
                        self.expand_clients[type] = self.create_client(TSTExecutorJsonExpand, f'tst_executor/{ntype}/executor_json_expand', callback_group=self.group)
                        timeout = 1.0
                    service_exist = self.expand_clients[type].wait_for_service(timeout_sec=timeout)
                    self.get_logger().info(f'service_exist tst_executor/{ntype}/executor_json_expand: {service_exist}')
                    if not service_exist:
                        self.get_logger().info(f'traverse_expansion Expand service did not exist: tst_executor/{ntype}/executor_json_expand')
                        return False

                    req = TSTExecutorJsonExpand.Request()
                    req.node_json = jobjstr
                    future = self.expand_clients[type].call_async(req)
                    resp = self.wait_future(future)
                    # self.get_logger().info(f'traverse_expansion RESP: {resp}')                    
                    if resp.expansion_json:
                        cobj = json.loads(resp.expansion_json)
                        cobjstr = json.dumps(cobj, sort_keys=True, indent=4, separators=(',', ': '))
                        self.get_logger().info(f'traverse_expansion LINKING IN: {cobjstr}')
                        jobj["children"].append(cobj)
                        if not ntype in self.expandable_types:
                            self.expandable_types.append(ntype)
                            self.expandable_types.append(jobj["name"])
                        self.traverse_expansion(jobj)
                        return True
                    else:
                        return True
                    ## jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                    ## self.get_logger().error(f'traverse_expansion EXPANDED: {jobjstr}')
                    
                
            if len(jobj["children"]) > 0:
                res = True
                for cobj in jobj["children"]:
                    if not self.traverse_expansion(cobj):
                        res = False
                        break
            
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION traverse_expansion: {type(exc)} - {exc}')
            return False
        
        return res

    def traverse_start_non_terminal(self, jobj):
        print("S", end="", flush=True)
        try:
            if (not "status" in jobj and False):
                jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                self.get_logger().info(f'traverse_start_non_terminal: {jobjstr}')

            ## self.get_logger().info(f'traverse_start_non_terminal: {jobj["execution_ns"]} - {self.get_namespace()}')                
                
            if jobj["execution_ns"] != self.get_namespace():
                ## self.get_logger().info(f'traverse_start_non_terminal: RETURN FALSE')
                jobj["status"] = "fail"
                return False
            
            if not "status" in jobj:
                self.set_executing_flag(jobj["id"], True)
                ## self.get_logger().info(f'START non-terminal: {jobj["name"]} - {jobj["id"]}')
                jobj["status"] = "active"

            ## self.get_logger().info(f't**************************')                                
            if jobj["status"] in ["success", "fail"]:
                ### print("traverse_start_non_terminal: ", jobj["id"], "RETURN status success or fail")
                return True
            ## self.get_logger().info(f't=======================')                                            
            
            if self.is_seq(jobj):
                ## print("SEQ STATUS:", jobj["status"])
                for cobj in jobj["children"]:
                    if not "status" in cobj:
                        ##print("*****************start seq")
                        res = self.traverse_tree_start(cobj)
                        return res
                    if cobj["status"] == "active":
                        ##print("*****************start active")
                        res = self.traverse_tree_start(cobj)
                        return res
                    ##print("STATUS:", cobj["status"])
                    if cobj["status"] == "fail":
                        return False
                    if cobj["status"] == "success":
                        continue
                print("EVERYTHING FINISHED: Should status be set to success?")
                ## jobj["status"] = "success"
                
            if jobj["name"] == "conc":
                for cobj in jobj["children"]:
                    if not "status" in cobj:
                        res = self.traverse_tree_start(cobj)
                        #if not res:
                        #    return res
                    if cobj["status"] == "active":
                        res = self.traverse_tree_start(cobj)
                        #if not res:
                        #    return res
                    # If child has failed, what should happen?
                    if cobj["status"] == "fail":
                        res = self.traverse_tree_start(cobj)
                        #return False

                # Why is this needed???
                for cobj in jobj["children"]:                    
                    if cobj["status"] == "fail":
                        return False
                    if cobj["status"] == "active":
                        return True

            if jobj["name"] == "test-if":
                test_obj = jobj["children"][0]
                yes_obj = jobj["children"][1]
                no_obj = jobj["children"][2]
                unknown_obj = jobj["children"][3]
                
                if not "status" in test_obj:
                    print("*****************start test-if")
                    res = self.traverse_tree_start(test_obj)
                    if not res:
                        return res

                # print("TESTIBJ STATUS:", test_obj["status"])
                if test_obj["status"] == "fail":
                    return False
            
                if test_obj["status"] == "success":
                    resp = self.get_node_from_uuid(test_obj["node_uuid"])
                    params = Parameters(resp.info)
                    answer = params.get_value("answer")
                    ## self.get_logger().info(f'test_if answer: {answer}')
                    if answer == "true":
                        res = self.traverse_tree_start(yes_obj)
                    if answer == "false":
                        res = self.traverse_tree_start(no_obj)                        
                    if answer == "unknown":
                        res = self.traverse_tree_start(unknown_obj)

                    print ("===================RETURN:", res)
                    return res

            return True
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION traverse_start_non_terminal: {type(exc)} - {exc}')
            return False
    
    def traverse_start_terminal(self, jobj):
        print("s", end="", flush=True)
        try:
            res = True
            if not "status" in jobj and False:
                jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                #self.get_logger().error(f'traverse_start_terminal: {jobjstr}')
                self.get_logger().info(f'traverse_start_terminal: {jobj["id"]}')
                
            if jobj["execution_ns"] == self.get_namespace():
                if not "status" in jobj:
                    # self.set_executing_flag(jobj["id"], True)
                    self.set_waiting_flag(jobj["id"], True)
                    jobj["status"] = "active"
                    self.get_logger().info(f'traverse_start_terminal SETTING stats active: {jobj["id"]}')
                    res = self.create_executor(jobj["name"], jobj["id"], True)
                    if res:
                        res = self.start_executor(jobj["name"], jobj["id"])
                    else:
                        self.get_logger().error(f'traverse_start_terminal FAILED TO CREATE EXECUTOR: {jobj["id"]}')
                        self.set_succeeded_flag(jobj["id"], False)
                        self.set_finished_flag(jobj["id"], True)
            else:
                if not "status" in jobj:
                    self.set_executing_flag(jobj["id"], True)
                    jobj["status"] = "active"
                    self.get_logger().info(f'create remote executor for: {jobj["id"]}')
                    res = self.create_executor("remote", jobj["id"], True)
                    if res:
                        res = self.start_executor("remote", jobj["id"])
                        if not res:
                            self.set_succeeded_flag(jobj["id"], False)
                            self.set_finished_flag(jobj["id"], True)
                    else:
                        self.set_succeeded_flag(jobj["id"], False)
                        self.set_finished_flag(jobj["id"], True)
            return res
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION handle_terminal: {type(exc)} - {exc}')

    def traverse_tree_start(self, jobj):
        print("Q", end="", flush=True)
        try:
            #print("HANDLE_TREE:", json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
            ###print("HANDLE_TREE:", jobj["id"])

            if len(jobj["children"]) > 0:
                ###print("NONTERMINAL")
                res = self.traverse_start_non_terminal(jobj)
            else:
                res = self.traverse_start_terminal(jobj)
            # print("HANDLE_TREE RES:")                
            return res
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION traverse_tree_start: {type(exc)} - {exc}')
    
    def tst_execute(self, request, response):
        try:
            self.get_logger().info(f'tst_execute')
            self.get_logger().info(f'REQUEST: {request}')

            jobj = json.loads(request.json_tst)
            root_uuid = ""
            if "common_params" in jobj:
                cobj = jobj["common_params"]
                if "node-uuid" in cobj:
                    root_uuid = cobj["node-uuid"]
            else:
                if "node_uuid" in jobj:
                    root_uuid = jobj["node_uuid"]
                if "node-uuid" in jobj:
                    root_uuid = jobj["node-uuid"]

            if root_uuid in self.have_tree:
                self.get_logger().error(f'DUPLICATE IGNORING tst_execute ({root_uuid}): {len(self.tree_strings)} - {self.tree_strings_index}')
                response.success = False
                response.uuid = ""
                return response
            else:
                self.have_tree[root_uuid] = True

            self.tree_strings.append(request.json_tst)
            self.tree_channels.append(request.channel)
            
            self.get_logger().info(f'tst_execute: {len(self.tree_strings)} - {self.tree_strings_index}')
            response.success = True
            response.uuid = ""
            # self.get_logger().error(f'tst_execute RESP: {response}')            
            return response            
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION exec_tree: {type(exc)} - {exc}')

    def tree_finished(self, jobj):
        if "status" in jobj and jobj["status"] in ["success", "fail"]:
            return True
        else:
            return False

    def timer_callback(self):
        try:
            if self.inside_timer:
                return
            self.inside_timer = True
            remove_index = []
            index = 0
            for jobj in self.tree_objects:
                # self.get_logger().error(f'tree_executor timer_callback: {jobj}')
                if jobj:
                    res = self.traverse_tree_start(jobj)
                    res = self.traverse_terminals(jobj)
                    self.traverse_update(jobj)
                    if self.tree_finished(jobj):
                        remove_index.append(index)
                index += 1
            for i in remove_index:
                print("REMOVE TREE:", i)
                self.tree_objects[i] = None
                
            self.inside_timer = False                
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION timer_callback: {type(exc)} - {exc}')

    def timer2_callback(self):
        try:
            if self.inside_timer2:
                return
            self.inside_timer2 = True

            # str = self.tree_strings[self.tree_strings_index:]
            # self.get_logger().error(f'tree_executor timer2_callback str: {str} - {self.tree_strings_index}')
            #if len(self.tree_strings) > 0:
            #    self.get_logger().error(f'timer2_callback: {len(self.tree_strings)} - {self.tree_strings_index}')
            for json_tst in self.tree_strings[self.tree_strings_index:]:
                # self.get_logger().error(f'tree_executor timer2_callback: {json_tst}')
                self.get_logger().info(f'ADDING TREE OBJ: {self.tree_strings_index}')

                channel = self.tree_channels[self.tree_strings_index]

                self.tree_strings_index += 1

                jobj = json.loads(json_tst)
                expansion_flag = self.traverse_expansion(jobj)
                self.get_logger().info(f'EXPANSION_FLAG: {expansion_flag}')
                #expansion_flag = self.traverse_expansion(jobj)
                #self.get_logger().info(f'EXPANSION_FLAG: {expansion_flag}')

                jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                #self.get_logger().info(f'ADDING TREE OBJ EXPANDED TREE: {jobjstr}')
                
                displaytype="exec"            
                resp = self.create_tree(jobjstr)
                self.get_logger().info(f'CREATE_TREE RESP: {resp}')

                if resp.success:
                    rootid = resp.id
                    free_id = resp.free_id
                    root_uuid = resp.global_tree_uuid
                else:
                    break
            
                if self.display:
                    msg = DisplayRequest()
                    msg.title="tst_command"
                    msg.type=displaytype
                    msg.root_uuid=root_uuid
                    msg.channel = channel
                    print("DISPLAY PUB:", msg)
                    # time.sleep(2.5) # seems not to be needed... was needed in ros1
                    self.disp_pub.publish(msg)
                    # self.get_logger().error(f'DISPLAY: {msg}')

                resp = self.get_tree_string(root_uuid, 4)
                # self.get_logger().error(f'STORED_TREE RESP: {resp}')
                jobj = json.loads(resp.data)
                jobjstr = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                ## self.get_logger().error(f'STORED_TREE RESP FORMATTED: {jobjstr}')
                ## self.get_logger().info(f'STORED_TREE RESP FORMATTED: {jobjstr}')
                self.tree_objects.append(jobj)
                    
            self.inside_timer2 = False                
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION timer2_callback: {type(exc)} - {exc}')

def main(args=None):
    print('Hi from tree_executor server.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=8)
    node = TreeExecutor()
    executor.add_node(node)

    try:
        node.get_logger().info(f'SPINNING tree_executor')
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()

