import time
import sys
import uuid
import json
import threading
from threading import Lock

import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup

from lrs_msgs_tst.msg import TSTConstraintCommand, TSTConstraintResponse
from lrs_srvs_tst.srv import TSTAddConstraintsToTree

from lrs_srvs_exec.srv import TSTExecutorJsonConstraints
from lrs_exec.expand import get_exec_unit
from lrs_msgs_common.msg import UnitInfoSlow

class ConstraintNode(Node):

    def __init__(self):
        super().__init__('tst_constraint')
        self.group = ReentrantCallbackGroup()

        self.add_constraints_srv = self.create_service(TSTAddConstraintsToTree, "tst_add_constraints_to_tree", self.add_constraints, callback_group=self.group)

        self.constraint_command_sub = self.create_subscription(TSTConstraintCommand, "/tst_constraint_command", self.constraint_command_callback, 10, callback_group=self.group)
        self.constraint_response_sub = self.create_subscription(TSTConstraintResponse, "/tst_constraint_response", self.constraint_response_callback, 10, callback_group=self.group)

        self.constraint_command_pub = self.create_publisher(TSTConstraintCommand, "/tst_constraint_command", 10, callback_group=self.group)
        self.constraint_response_pub = self.create_publisher(TSTConstraintResponse, "/tst_constraint_response", 10, callback_group=self.group)
        
        self.com_uuid_to_jobj = {}
        self.trees = {}
        self.constraints_clients = {}

    def wait_future(self, future, print_waiting=False):
        try:
            while not future.done():
                if print_waiting:
                    print(f"wait_future executor")
            resp = future.result()
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION wait_future: {type(exc)} - {exc}')
        
        
    def do_add_constraints(self, jobj):
        try:
            if len(jobj["children"]) > 0:
                for c in jobj["children"]:
                    self.do_add_constraints(c)
            else:
                ns = get_exec_unit(jobj)
                name = jobj["name"].replace("-", "_")                
                if ns:
                    msg = TSTConstraintCommand()
                    com_uuid = str(uuid.uuid4())
                    self.com_uuid_to_jobj[com_uuid] = jobj
                    jobj["constraints_added"] = False
                    msg.com_uuid = com_uuid
                    msg.ns = ns
                    msg.type = name
                    msg.json = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                    self.constraint_command_pub.publish(msg)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION do_add_constraints: {type(exc)} - {exc}')
                    
    def is_finished(self, jobj):
        # print("is_expanded:", jobj)
        try:
            if "constraints_added" in jobj:
                flag = jobj["constraints_added"]
            else:
                flag = True
            if "children" in jobj:
                for c in jobj["children"]:
                    flag &= self.is_finished(c)
            return flag
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION is_finished: {type(exc)} - {exc}')
        return True

    def remove_extra_info(self, jobj):
        try:
            if "constraints_added" in jobj:
                del jobj["constraints_added"]
            if "children" in jobj:
                for c in jobj["children"]:
                    self.remove_extra_info(c)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION remove_extra_info: {type(exc)} - {exc}')

    def add_constraints(self, request, response):
        try:
            # print("add_constraints:", request)
            jobj = json.loads(request.input_json)
            #tree_uuid = str(uuid.uuid4())
            #self.trees[tree_uuid] = jobj

            self.do_add_constraints(jobj)
            while not self.is_finished(jobj):
                # print("wait for expansion")
                #print(json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
                time.sleep(0.1)
            self.remove_extra_info(jobj)
                
            response.output_json = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
            response.success = True
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION add_constraints: {type(exc)} - {exc}')
            response.output_json = ""
            response.success = False
            response.error = 1
            response.reason = str(exc)
        return response

    def constraint_command_callback(self, msg):        
        try:
            if msg.ns == self.get_namespace():
                print("constraint_command_callback:", msg)
                # jobj = json.loads(msg.json)
                # print(json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))

                # Later maybe add optimisation similar to expand

                resp = None

                if msg.type:
                    type = msg.type.replace("-","_")                    
                    self.constraints_clients[type] = self.create_client(TSTExecutorJsonConstraints, f'tst_executor/{type}/executor_json_constraints', callback_group=self.group)
                    # print("CLIENT:", self.expand_clients[type])
                    req = TSTExecutorJsonConstraints.Request()
                    req.node_json = msg.json
                    # print("SERVICE:", f'tst_executor/{type}/executor_json_expand')
                    # print("REQ:", req)
                    future = self.constraints_clients[type].call_async(req)
                    resp = self.wait_future(future)
                    # print("RESP:", resp)

                respmsg = TSTConstraintResponse()
                respmsg.response_to = msg.com_uuid
                respmsg.ns = msg.ns
                if resp:
                    respmsg.json = resp.constraints_json
                else:
                    respmsg.json = ""
                self.constraint_response_pub.publish(respmsg)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION constraint_command_callback: {type(exc)} - {exc}')

    def constraint_response_callback(self, msg):
        try:
            # print("constraint_response_callback:", msg)
            if msg.response_to in self.com_uuid_to_jobj:
                jobj = self.com_uuid_to_jobj[msg.response_to]
                jobj["constraints_added"] = True
                #
                # Merging should be done here...
                #
                if msg.json:
                    jobj["constraints"] = json.loads(msg.json)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION constraint_response_callback: {type(exc)} - {exc}')


def main(args=None):
    print('Hi from tst constraint server.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=8)
    node = ConstraintNode()
    executor.add_node(node)

    try:
        executor.spin()
    except KeyboardInterrupt:
        pass
    
    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()

        
