#
# class implementing distribution of exec tst trees. Enables by inheriting into ExecFactory
#

import uuid
import json

from lrs_srvs_tst.srv import TSTDistributeTreeFromUuid
from lrs_msgs_tst.msg import TSTDistributeTreeInfo, TSTUpdateNodeInfo
from lrs_msgs_common.srv import GetResponse
from lrs_msgs_common.msg import Response
from lrs_srvs_wdb.srv import WDBGetPossibleUnits

class TstDistribute ():
    
    def add_serv(self, type, srvname, callback):
        srv = self.create_service(type, f'{self.prefix}/{srvname}', callback, callback_group=self.group)
        self.srvs.append(srv)

    def __init__(self, external_prefix):
        self.verbose = False
        self.com_uuid_to_name = {}
        self.com_uuid_to_units = {}
        self.com_uuid_to_units_responded = {}

        #
        # Service calls
        #

        self.add_serv(TSTDistributeTreeFromUuid, "distribute_tree_from_uuid", self.distribute_tree)
        self.add_serv(GetResponse, "get_response", self.get_response)        

        #
        # clients
        #
        
        self.get_possible_units_client = self.create_client(WDBGetPossibleUnits, "wdb/get_possible_units", callback_group=self.group)
        
        
        #
        # Publishers
        #
        
        self.distribute_tree_pub = self.create_publisher(TSTDistributeTreeInfo, f'{external_prefix}/distribute_tree', 10)
        self.update_node_pub = self.create_publisher(TSTUpdateNodeInfo, f'{external_prefix}/update_node', 1000)
        self.response_pub = self.create_publisher(Response, f'{external_prefix}/response', 100)

        #
        # Subscribers
        #
        
        self.distribute_tree_sub = self.create_subscription(TSTDistributeTreeInfo, "/distribute_tree", self.distribute_tree_callback, 10)
        self.distribute_node_sub = self.create_subscription(TSTUpdateNodeInfo, "/update_node", self.distribute_node_callback, 10)
        self.response_sub = self.create_subscription(Response, "/response", self.response_callback, 10)

    def lprint(self, *val):
        if self.verbose:
            print(val)

    def distribute_tree_callback(self, msg):
        try:
            if msg.sender == self.get_namespace():
                return
            self.lprint("distribute_tree_callback:", msg.uuid)
            free_id = 1000  ## WHY NOT USE CURRENTLY HIGHEST???  Maybe this is intended to be overwitten by json value
            error = 0
            fail_reasone = ""
            resp = Response()
            resp.uuid = str(uuid.uuid4())
            resp.responder = self.get_namespace()
            resp.cuuid = msg.uuid
            
            jsonobj = json.loads(msg.tree_json)
            root_uuid = jsonobj["node_uuid"]
            
            if root_uuid in self.nodes:
                print("Duplicate distributed tree. Should not happen")
                resp.success = True
                self.response_pub.publish(resp)
                return

            (have_created_tree, root, free_id) = self.basic_create_tree(msg.tree_json, free_id, False)

            if have_created_tree:
                # update data structures
                # root_tree_map[id] = t;
                # node_tree_map[id] = t;
                # tree_list.push_back (t);
                # t->set_distribute_to(com_uuid_to_units_map[msg->uuid]); WHY THIS???
                resp.success = True
            else:
                resp.success = False
            
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')            
            resp.success = True

        self.response_pub.publish(resp)

    def distribute_node_callback(self, msg):
        try:
            if msg.sender == self.get_namespace():
                return

            self.lprint("distribute_node_callback:", msg.sender, msg.node_uuid)
            
            resp = Response()
            resp.uuid = str(uuid.uuid4())
            resp.responder = self.get_namespace()
            resp.cuuid = msg.uuid

            if msg.node_uuid in self.nodes:
                node = self.nodes[msg.node_uuid]
                notify_start_executor = msg.start_executor and not node.start_executor
                node.update(msg)
                self.notify_status(node)
                if notify_start_executor:
                    self.notify_start_executor(node)
                resp.success = True
            else:
                resp.success = False
            
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')            
            resp.success = True

        self.lprint("DISTRIBUTE_NODE_CALLBACK RESP PUB:", resp)
        self.response_pub.publish(resp)            
        
    def response_callback(self, msg):
        if msg.responder == self.get_namespace():
            return
        # print("response_callback:", msg)
        if msg.cuuid in self.com_uuid_to_units:
            have_resp = False
            for el in self.com_uuid_to_units_responded[msg.cuuid]:
                if el == msg.responder:
                    have_resp = True
            if not have_resp:
                if msg.responder in self.com_uuid_to_units[msg.cuuid]:
                    self.com_uuid_to_units_responded[msg.cuuid].append(msg.responder)

    def get_response(self, request, response):
        try:
            self.lprint("get_response:", request)
            cuuid = request.cuuid
            if cuuid in self.com_uuid_to_units:
                flag = (len(self.com_uuid_to_units[cuuid]) == len(self.com_uuid_to_units_responded[cuuid]))
                if len(self.com_uuid_to_units[cuuid]) == 0:
                    # no distribution
                    flag = True
                response.have_response = flag
                response.name = self.com_uuid_to_name[cuuid]
                response.need_response_from = self.com_uuid_to_units[cuuid]
                response.have_response_from = self.com_uuid_to_units_responded[cuuid]
                response.success = flag
            else:
                response.have_response = False
                response.name = ""
                response.success = False                
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')            
            response.have_response = False
            response.name = ""
            response.success = False
        self.lprint("get_response resp:", response)
        return response
    
            
    def distribute_tree(self, request, response):
        try:
            self.get_logger().error(f'DISTRIBUTE TREE - to: {request}')
            root_uuid = request.uuid
            units = request.units

            ## check that tree exists

            if not root_uuid in self.trees:
                response.com_uuid = ""
                response.success = False
                response.error = 1
                response.reason = "Tree does not exist"
                self.lprint("distribute_tree resp:", response)
                return response

            if len(units) == 0:
                resp = self.get_possible_units()
                for u in resp.units:
                    if u != self.get_namespace():
                        units.append(u)

            self.trees[root_uuid].set_distribute_to(units)
            dti = TSTDistributeTreeInfo()
            dti.uuid = str(uuid.uuid4())
            dti.sender = self.get_namespace()
            exec_obj = self.trees[root_uuid].get_json_obj()
            dti.tree_json = json.dumps(exec_obj, sort_keys=True, indent=4, separators=(',', ': '))
            self.lprint("DTI:", dti)
            self.com_uuid_to_name[dti.uuid] = "distribute_tree"
            self.com_uuid_to_units[dti.uuid] = units
            self.com_uuid_to_units_responded[dti.uuid] = []
            self.distribute_tree_pub.publish(dti)

            response.com_uuid = dti.uuid
            response.success = True
            response.error = 0
            response.reason = ""
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')
            response.com_uuid = ""
            response.success = False
            response.error = 2
            response.reason = f'distribute_tree Exception: {exc}'            
        self.lprint("distribute_tree resp:", response)
        return response


    def get_possible_units(self, team=""):
        self.lprint("get_possible_units")
        req = WDBGetPossibleUnits.Request()
        req.team = team
        future = self.get_possible_units_client.call_async(req)
        resp = self.wait_future(future)
        self.lprint("get_possible_units_client resp:", resp)
        return resp
    
    def distribute_node(self, node):
        try:
            distribute_to = node.get_distribute_to()
            for el in distribute_to:
                self.get_logger().error(f'DISTRIBUTE NODE - to: {el}')
            if not distribute_to:
                self.get_logger().error(f'DISTRIBUTE_NODE WARNING: distribute_to is empty')
            msg = node.get_update_node_info_msg()
            msg.sender = self.get_namespace()
            msg.uuid = str(uuid.uuid4())
            self.com_uuid_to_name[msg.uuid] = "distribute_node"
            self.com_uuid_to_units[msg.uuid] = distribute_to
            self.com_uuid_to_units_responded[msg.uuid] = []
            self.update_node_pub.publish(msg)
            return msg.uuid
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')
        return ""
    
