#
# All functionality for distributed exec TST is in tst_distribute.py
#
# 
#


import time
import sys
import uuid
import json
import threading
from threading import Lock
from collections import deque

import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup
from rclpy.qos import qos_profile_system_default, qos_profile_sensor_data, QoSProfile, QoSReliabilityPolicy, QoSHistoryPolicy

# exec_tst_type_def.append(["set_start_executor_from_uuid", "lrs_srvs_tst/srv/TSTSetStartExecutorFlagFromUuid"])

from lrs_srvs_tst.srv import TSTSendSignal, TSTSignalTree, TSTSignalUnit
from lrs_srvs_tst.srv import TSTCreateTree, TSTGetTreeStringFromUuid
from lrs_srvs_tst.srv import TSTSetStartExecutorFlagFromUuid, TSTSetFlag, TSTGetNodeFromUuid, TSTSetFailReason
from lrs_srvs_tst.srv import TSTSetStartExecutorFlag, TSTGetNode, TSTGetIdFromUuid, TSTGetUuidFromId
from lrs_srvs_tst.srv import TSTGetExecInfo, TSTGetTimeInfo, TSTSetExecInfo
from lrs_srvs_tst.srv import TSTSetNodeParameterString

from lrs_msgs_tst.msg import TSTStartExecutor, TSTNodeStatus, TSTSignalExecutor
from lrs_exec.exec_tst_node import ExecTstNode

from lrs_exec.tst_distribute import TstDistribute

class ExecFactory(Node, TstDistribute):
    

    def add_serv(self, type, srvname, callback):
        srv = self.create_service(type, f'{self.prefix}/{srvname}', callback, callback_group=self.group)
        self.srvs.append(srv)

    def __init__(self):
        #super().__init__('tst_exec_factory')
        self.group = ReentrantCallbackGroup()
        prefix = "exec/tstfactory"
        self.prefix = prefix
        self.srvs = []
        Node.__init__(self, 'tst_exec_factory')

        self.declare_parameter("external", False)
        self.external = self.get_parameter('external').get_parameter_value().bool_value
        external_prefix = ""
        if self.external:
            external_prefix = "/ext"
        # prefix = "/ext"

        # self.get_logger().error(f'External_Prefix: {external_prefix}')
        
        TstDistribute.__init__(self, external_prefix)
        self.trees = {}
        self.idtrees = {}
        self.nodes = {}
        self.idnodes = {}
        self.id_to_uuid = {}
        self.uuid_to_id = {}
        self.factory_local_id = 1    # counter to give a factory unique id for a node
        self.distribute_flag = False

        #
        # srv to setup things
        #

        self.add_serv(TSTCreateTree, "create_tree", self.create_tree)        
        self.add_serv(TSTGetTreeStringFromUuid, "get_tree_string_from_uuid", self.get_tree_string)
        self.add_serv(TSTSetStartExecutorFlagFromUuid, "set_start_executor_from_uuid", self.set_start_executor_from_uuid)                
        self.add_serv(TSTSetStartExecutorFlag, "set_start_executor", self.set_start_executor)                

        #
        # srvs during execution
        #

        self.add_serv(TSTSetFlag, "set_active_flag", self.set_active_flag)
        self.add_serv(TSTSetFlag, "set_waiting_flag", self.set_waiting_flag)
        self.add_serv(TSTSetFlag, "set_executing_flag", self.set_executing_flag)
        self.add_serv(TSTSetFlag, "set_finished_flag", self.set_finished_flag)                
        self.add_serv(TSTSetFlag, "set_was_finished_flag", self.set_was_finished_flag)                
        self.add_serv(TSTSetFlag, "set_aborted_flag", self.set_aborted_flag)
        self.add_serv(TSTSetFlag, "set_paused_flag", self.set_paused_flag)
        self.add_serv(TSTSetFlag, "set_succeeded_flag", self.set_succeeded_flag)
        #  self.add_serv(TSTSetFailReason, "set_fail_reason", self.set_fail_reason)
        self.add_serv(TSTGetNode, "get_node", self.get_node)
        self.add_serv(TSTGetNodeFromUuid, "get_node_from_uuid", self.get_node_from_uuid)
        self.add_serv(TSTGetIdFromUuid, "get_id_from_uuid", self.get_id_from_uuid)                
        self.add_serv(TSTGetUuidFromId, "get_uuid_from_id", self.get_uuid_from_id)
        self.add_serv(TSTGetExecInfo, "get_exec_info", self.get_exec_info)
        self.add_serv(TSTSetExecInfo, "set_exec_info", self.set_exec_info)                
        self.add_serv(TSTGetTimeInfo, "get_time_info", self.get_time_info)        
        self.add_serv(TSTSendSignal, "send_signal", self.send_signal)        
        self.add_serv(TSTSignalTree, "signal_tree", self.signal_tree)        
        self.add_serv(TSTSignalUnit, "signal_unit", self.signal_unit)        
        self.add_serv(TSTSetNodeParameterString, "set_node_parameter_string", self.set_node_parameter_string)        
        
        #
        # publishers
        #
        
        self.start_executor_pub = self.create_publisher(TSTStartExecutor, '/tst_start_executor',
                                                        QoSProfile(reliability=QoSReliabilityPolicy.RELIABLE,
                                                                   history=QoSHistoryPolicy.KEEP_LAST,
                                                                   depth=100),
                                                        callback_group=self.group)
        self.node_status_pub = self.create_publisher(TSTNodeStatus, 'tst_node_status',
                                                     QoSProfile(reliability=QoSReliabilityPolicy.RELIABLE,
                                                                history=QoSHistoryPolicy.KEEP_LAST,
                                                                depth=100),
                                                     callback_group=self.group)
        self.signal_executor_pub = self.create_publisher(TSTSignalExecutor, "tst_signal_executor",
                                                         QoSProfile(reliability=QoSReliabilityPolicy.RELIABLE,
                                                                    history=QoSHistoryPolicy.KEEP_LAST,
                                                                    depth=100),
                                                         callback_group=self.group)
        
        
        self.notify_queue = deque()
        self.notify_start_executor_queue = deque()
        self.inside_timer = False
        self.timer = self.create_timer(0.05, self.timer_callback, callback_group=self.group)

    def timer_callback(self):
        try:
            if self.inside_timer:
                return
            # self.get_logger().error(f'timer_callback')            
            self.inside_timer = True
            finished = False
            while not finished:
                try:
                    msg = self.notify_queue.popleft()
                    ## self.get_logger().info(f'timer_callback node_status_pub: {msg}')            
                    self.node_status_pub.publish(msg)                    
                except Exception as exc:
                    ## self.get_logger().info(f'timer_callback notify_queue empty')                                
                    finished = True
                    
            finished = False
            while not finished:
                try:
                    node = self.notify_start_executor_queue.popleft()
                    ## self.get_logger().info(f'timer_callback POP start_executor_queue')                                
                    self.do_notify_start_executor(node)
                except Exception as exc:
                    ## self.get_logger().info(f'timer_callback start_executor_queue empty')                                
                    finished = True
            self.inside_timer = False
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')            
            # self.get_logger().error(f'timer_callback exit')            
        self.inside_timer = False
        
    def wait_future(self, future, print_waiting=False):
        while not future.done():
            if print_waiting:
                print("wait_future tst_exec_factory")
        resp = future.result()
        return resp
    

    def assign_nodes(self, node):
        self.nodes[node.get_node_uuid()] = node
        self.idnodes[node.get_factory_local_id()] = node
        self.id_to_uuid[node.get_factory_local_id()] = node.get_node_uuid()
        self.uuid_to_id[node.get_node_uuid()] = node.get_factory_local_id()
        for child in node.children:
            self.assign_nodes(child)

    def basic_create_tree(self, jsonstr, free_id, approve=True):
        ## self.get_logger().info(f'basic_create_tree - {jsonstr}')            
        root = ExecTstNode()
        try:
            jsonobj = json.loads(jsonstr)
            (res, free_id, factory_local_id) = root.create_tree(jsonobj, free_id, self.factory_local_id,
                                                                self.factory_local_id,
                                                                approve=approve)
            self.factory_local_id = factory_local_id
            self.trees[root.node_uuid] = root
            self.idtrees[root.get_factory_local_id()] = root
            if res:
                self.assign_nodes(root)
                execobj = root.get_json_obj()
                newjsonstr = json.dumps(jsonobj, sort_keys=True, indent=4, separators=(',', ': '))
                execobjstr = json.dumps(execobj, sort_keys=True, indent=4, separators=(',', ': '))
                #self.get_logger().info(f"JSON: {newjsonstr}")
                #self.get_logger().info(f"EXEC JSON: {execobjstr}")
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION basic_create_tree: {type(exc)} - {exc}')
        return(res, root, free_id)

    def create_tree(self, request, response):
        try:
            self.get_logger().info(f"tst_exec_factory create_tree: {request}")
            jsonstr = request.json
            free_id = request.free_id
            (res, root, free_id) = self.basic_create_tree (jsonstr, free_id, approve=request.approve_flag)
            self.get_logger().info(f"NEW FREE_ID: {free_id}")
            if res:
                response.success = True
                response.error = 0
                response.reason = ""
                response.global_tree_uuid = root.node_uuid
                response.id = root.get_factory_local_id() # deprecated value, should never be used...
                response.free_id = free_id
            else:
                response.success = False
                response.error = 1
                response.reason = "Failed to parse json"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION create_tree: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'create_tree Exception: {exc}'            
        #self.get_logger().info(f'create_tree resp: {response}')
        return response

    #
    # Local notification about node status (use for update of display of TST tree, and used to sequence updates)
    #
    def notify_status(self, node):
        try:
            msg = node.get_status_msg()
            ###self.get_logger().info(f'tst_exec_factory notify_status: {msg}')
            msg.ns = self.get_namespace()
            self.notify_queue.append(msg)
            # self.node_status_pub.publish(msg)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION notify_status: {type(exc)} - {exc}')

    #
    # Local notification of the tst exec manager to start execution
    #
    def do_notify_start_executor(self, node):
        try:        
            se = TSTStartExecutor()
            se.exec_ns = node.get_exec_ns()
            se.type = node.get_type()
            se.ns = self.get_namespace()
            se.id = node.get_factory_local_id()
            # The following fix is a bit strange, is the design that this should only be local then why listen on a
            # global topic...
            if se.ns == se.exec_ns:
                ## self.get_logger().info(f'start_executor_pub: {se}')                        
                self.start_executor_pub.publish(se)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION notify_start_executor: {type(exc)} - {exc}')

    def notify_start_executor(self, node):
        self.notify_start_executor_queue.append(node)

    def get_node_string(self, nuuid, format=4):
        res = ""
        if nuuid in self.nodes:
            jsonobj = self.nodes[nuuid].get_json_obj()
            res = json.dumps(jsonobj, sort_keys=True, indent=4, separators=(',', ': '))
        return res
        
    def get_tree_string(self, request, response):
        try:
            ## self.get_logger().error(f'get_tree_string: {request}')                        
            tuuid = request.uuid
            if tuuid in self.trees:
                response.data = "DUMMY STRING"
                if request.type == request.FORMAT_JSON_EXEC:
#                if request.type == 4:
                    ### This should be a string...
                    jsonobj = self.trees[tuuid].get_json_obj()
                    response.data = json.dumps(jsonobj, sort_keys=True, indent=4, separators=(',', ': '))
#                if request.type == request.FORMAT_DOT:
                if request.type == 2:
                    response.data = self.trees[tuuid].get_dot()                    
                response.success = True
                response.error = 0
                response.reason = ""
            else:
                str = self.get_node_string(request.uuid)
                if str:
                    response.data = str
                    response.success = True
                    response.error = 0
                    response.reason = ""
                else:
                    response.data = ""
                    response.success = False
                    response.error = 1
                    response.reason = "Did not find tree or node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION get_tree_string: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        ## self.get_logger().info(f'get_tree_string_from_uuid: {response}')                        
        return response

    def set_start_executor_from_uuid(self, request, response):
        #
        # Always a root node but maybe should not be handled differently
        #
        try:
            # self.get_logger().info(f'tst_exec_factory set_start_executor_from_uuid: {request}')
            tuuid = request.uuid
            if tuuid in self.trees:
                self.trees[tuuid].set_start_executor(request.val)
                if request.val:
                    # local notification
                    self.notify_start_executor(self.trees[tuuid])

                if self.distribute_flag:
                    # distribute all possible changes
                    com_uuid = self.distribute_node(self.trees[tuuid])
                    if com_uuid:
                        response.com_uuid = com_uuid
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.com_uuid = ""
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.com_uuid = ""
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                response.com_uuid = ""
                response.success = False
                response.error = 1
                response.reason = "Did not find tree"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_start_executor_from_uuid: {type(exc)} - {exc}')
            response.com_uuid = ""            
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        return response
    
    def set_start_executor(self, request, response):
        try:
            # self.get_logger().info(f'tst_exec_factory set_start_executor: {request}')            
            id = request.id
            if id in self.idnodes:
                self.idnodes[id].set_start_executor(request.val)
                tuuid = self.idnodes[id].get_node_uuid()
                if request.val:
                    self.notify_start_executor(self.idnodes[id])
                    
                if self.distribute_flag:
                    # distribute all possible changes
                    self.get_logger().info(f'tst_exec_factory set_start_executor DISTRIBUTE: {request}')
                    com_uuid = self.distribute_node(self.idnodes[id])
                    if com_uuid:
                        response.com_uuid = com_uuid
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        self.get_logger().error(f'tst_exec_factory set_start_executor: Could not find node in distribution: {request}')            
                        response.com_uuid = ""
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.com_uuid = ""
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                self.get_logger().error(f'tst_exec_factory set_start_executor: Did not find tree: {request}')
                response.com_uuid = ""
                response.success = False
                response.error = 1
                response.reason = "Did not find tree"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_start_executor: {type(exc)} - {exc}')
            response.com_uuid = ""            
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        #self.get_logger().info(f'tst_exec_factory set_start_executor RESP: {response}')
        return response
    
    def get_node(self, request, response):
        try:
            ## self.get_logger().error(f'get_node: {request}')
            if request.id in self.idnodes:
                response.info = self.idnodes[request.id].get_node_info()
                if response.info.factory_local_id >= 0:
                    response.success = True
                    response.error = 0
                    response.reason = ""
                else:
                    response.success = False
                    response.error = 2
                    response.reason = "id less than zero"
            else:
                response.success = False
                response.error = 1
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION get_node: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        # self.get_logger().info(f'get_node response: {response} - {response.info.strings_params}')        
        return response

    def get_exec_info(self, request, response):
        ###self.get_logger().info(f'get_exec_info: {request}')        
        try:
            ## print("S", end="", flush=True)
            # print("get_exec_info:", request)
            if request.node_uuid in self.nodes:
                node = self.nodes[request.node_uuid]
                response.execution_ns = node.get_exec_ns()
                response.type = node.get_type()
                response.exec_status = node.get_exec_status_msg()
                response.success = True
                response.error = 0
                response.reason = ""
            else:
                response.success = False
                response.error = 1
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION get_exec_info: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        # print("get_exec_info resp:", response)
        ## print("E", end="", flush=True)
        ###self.get_logger().info(f'exit get_exec_info: {response}')        
        return response

    def send_signal_basic(self, com_uuid, node_uuid, signal):
        if node_uuid in self.nodes:
            node = self.nodes[node_uuid]
            msg = TSTSignalExecutor()
            msg.uuid = com_uuid
            msg.sender = self.get_namespace()
            msg.exec_ns = node.get_exec_ns()
            msg.type = node.get_type()
            msg.ns = ""
            msg.node_uuid = node_uuid
            msg.signal = signal
            self.signal_executor_pub.publish(msg)
            self.get_logger().info(f'send_signal have been published: {msg}')
            return (True, 0, "")
        else:
            self.get_logger().error(f'send_signal_basic: node does not exist')
            return (False, 2, "Node does not exist")

    def send_signal(self, request, response):
        try:
            self.get_logger().info(f'send_signal: {request}')
            response.com_uuid = str(uuid.uuid4())
            (res, error, reason) = self.send_signal_basic(response.com_uuid, request.node_uuid, request.signal)
            response.success = res
            response.error = error
            response.reason = reason
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION send_signal: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        self.get_logger().info(f'send_signal resp: {response}')            
        return response
    
    def signal_tree(self, request, response):
        try:
            self.get_logger().info(f'signal_tree: {request}')

            node_uuid = request.node_uuid

            if node_uuid in self.trees:
                tree = self.trees[node_uuid]
            else:
                response.success = False
                response.error = 1
                response.reason = "Could not find the tree"
                self.get_logger().error(f'signal_tree resp: {response}')
                return response

            if request.signal == "$abort":
                node_uuids = tree.get_unexecuted_terminals()
                self.get_logger().error(f'unexecuted terinals: {node_uuids}')
                for node_uuid in node_uuids:
                    node = self.nodes[node_uuid]
                    node.abort_when_starting = True
                    
                # do set_abort_when_starting on t->get_unexecuted_terminals()

            fail_flag = False
            node_uuids = []
            if request.signal == "$abort":
                node_uuids = tree.get_executing_abortable_nodes()
                self.get_logger().error(f'executing_abortable: {node_uuids}')
            if request.signal == "$enough":
                node_uuids = tree.get_executing_enoughable_nodes()
            if request.signal == "$pause":
                node_uuids = tree.get_executing_pauseable_nodes()
            if request.signal == "$continue":
                node_uuids = tree.get_executing_pauseable_nodes()
            for node_uuid in node_uuids:
                self.get_logger().info(f'{request.signal} EXECUTING TREE NODE: {node_uuid}')
                com_uuid = str(uuid.uuid4())
                (res, error, reason) = self.send_signal_basic(com_uuid, node_uuid, request.signal)
                if res:
                    response.com_uuids.append(com_uuid)
                else:
                    self.get_logger().error(f'FAILED TO SEND SIGNAL: {error} {reason}')
                    fail_flag = True
            response.success = not fail_flag
            response.error = 0
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION signal_tree: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        self.get_logger().info(f'signal_tree resp: {response}')
        return response
    
    def signal_unit(self, request, response):
        try:
            self.get_logger().info(f'signal_unit: {request}')

            if request.signal == "$abort":
                # do set_abort_when_starting on t->get_unexecuted_terminals()
                pass

            fail_flag = False
            for key,tree in self.trees.items():
                self.get_logger().info(f'signal_unit checking tree with key: {key}')
                node_uuids = []
                if request.signal == "$abort":
                    node_uuids = tree.get_executing_abortable_nodes_on_unit(request.unit)
                if request.signal == "$enough":
                    node_uuids = tree.get_executing_enoughable_nodes_on_unit(request.unit)
                if request.signal == "$pause":
                    node_uuids = tree.get_executing_pauseable_nodes_on_unit(request.unit)
                if request.signal == "$continue":
                    node_uuids = tree.get_executing_pauseable_nodes_on_unit(request.unit)
                for node_uuid in node_uuids:
                    self.get_logger().info(f'{request.signal} EXECUTING UNIT NODE: {request.unit} {node_uuid}')
                    com_uuid = str(uuid.uuid4())
                    (res, error, reason) = self.send_signal_basic(com_uuid, node_uuid, request.signal)
                    if res:
                        response.com_uuids.append(com_uuid)
                    else:
                        self.get_logger().error(f'FAILED TO SEND SIGNAL: {error} {reason}')
                        fail_flag = True
            response.success = not fail_flag
            response.error = 0
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION signal_unit: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        self.get_logger().info(f'signal_unit resp: {response}')
        return response

    def get_time_info(self, request, response):
        try:
            if request.node_uuid in self.nodes:
                response.info = self.nodes[request.node_uuid].get_time_info()
                response.success = True
                response.error = 0
                response.reason = ""
            else:
                response.success = False
                response.error = 1
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION get_time_info: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        # print("get_time_info resp:", response)
        return response

    def get_node_from_uuid(self, request, response):
        try:
            # self.get_logger().info(f'get_node_from_uuid: {request}')
            if request.node_uuid in self.nodes:
                response.info = self.nodes[request.node_uuid].get_node_info()
                if response.info.factory_local_id >= 0:
                    response.success = True
                    response.error = 0
                    response.reason = ""
                else:
                    response.success = False
                    response.error = 2
                    response.reason = "id less than zero"
            else:
                response.success = False
                response.error = 1
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION get_node_from_uuid: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        # print("get_node_from_uuid resp:", response)
        return response
    
    def get_id_from_uuid(self, request, response):
        try:
            if request.uuid in self.uuid_to_id:
                response.id = self.uuid_to_id[request.uuid]
                response.success = True
                response.error = 0
                response.reason = ""
            else:
                response.success = False
                response.error = 1
                response.reason = "Did not find id from uuid"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION get_id_from_uuid: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        return response
    
    def get_uuid_from_id(self, request, response):
        try:
            if request.id in self.id_to_uuid:
                response.uuid = self.id_to_uuid[request.id]
                response.success = True
                response.error = 0
                response.reason = ""
            else:
                response.success = False
                response.error = 1
                response.reason = "Did not find uuid from id"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION get_uuid_from_id: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        return response
    
    def set_exec_info(self, request, response):
        try:
            ### self.get_logger().info(f'set_exec_info ######################: {request}')            
            if request.id in self.idnodes:
                self.idnodes[request.id].set_exec_info(request.val)
                response.success = True
                response.error = 0
                response.reason = ""
            else:
                response.success = False
                response.error = 1
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_exec_info: {type(exc)} - {exc}')
            response.success = False
            response.error = 2
            response.reason = f'Exception: {exc}'
        self.get_logger().info(f'set_exec_info resp %%%%%%%%%%%%%%%%%%%%: {response}')
        return response

    def set_active_flag(self, request, response):
        try:
            self.get_logger().info(f'set_active_flag: {request}')
            if request.id in self.idnodes:
                node = self.idnodes[request.id]
                node.set_active_flag(request.val)
                # print("set_active_flag TYPE:", node.get_type())
                self.notify_status(node)
                if self.distribute_flag:
                    com_uuid = self.distribute_node(node)
                    if com_uuid:
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                response.success = False
                response.error = 2
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_active_flag: {type(exc)} - {exc}')
            response.success = False
            response.error = 3
            response.reason = f'Exception: {exc}'
        # print("set_active_flag resp:", response)
        return response

    def set_succeeded_flag(self, request, response):
        try:
            # print("set_succeeded_flag:", request)            
            if request.id in self.idnodes:
                node = self.idnodes[request.id]
                node.set_succeeded_flag(request.val)
                self.get_logger().info(f'#########tst_exec_factory set_succeeded_flag: {node.get_type()}')                
                self.notify_status(node)
                if self.distribute_flag:                
                    com_uuid = self.distribute_node(node)
                    if com_uuid:
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                response.success = False
                response.error = 2
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_succeeded_flag: {type(exc)} - {exc}')
            response.success = False
            response.error = 3
            response.reason = f'Exception: {exc}'
        #print("set_succeeded_flag resp:", response)
        return response

    def set_finished_flag(self, request, response):
        try:
            # print("set_finished_flag:", request)            
            if request.id in self.idnodes:
                node = self.idnodes[request.id]
                node.set_finished_flag(request.val)
                self.get_logger().info(f'#########tst_exec_factory set_finished_flag: {node.get_type()}')
                self.notify_status(node)
                if self.distribute_flag:                                
                    com_uuid = self.distribute_node(node)
                    if com_uuid:
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                response.success = False
                response.error = 2
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_finished_flag: {type(exc)} - {exc}')
            response.success = False
            response.error = 3
            response.reason = f'Exception: {exc}'
        #print("set_finished_flag resp:", response)
        return response

    def set_was_finished_flag(self, request, response):
        try:
            #print("set_was_finished_flag:", request)            
            if request.id in self.idnodes:
                node = self.idnodes[request.id]
                node.set_was_finished_flag(request.val)
                # print("set_was_finished_flag TYPE:", node.get_type())
                self.notify_status(node)                
                if self.distribute_flag:
                    com_uuid = self.distribute_node(node)
                    if com_uuid:
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                response.success = False
                response.error = 2
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_was_finished_flag: {type(exc)} - {exc}')
            response.success = False
            response.error = 3
            response.reason = f'Exception: {exc}'
        #print("set_was_finished_flag resp:", response)
        return response

    def set_aborted_flag(self, request, response):
        try:
            #print("set_aborted_flag:", request)            
            if request.id in self.idnodes:
                node = self.idnodes[request.id]
                node.set_aborted_flag(request.val)
                # print("set_aborted_flag TYPE:", node.get_type())
                self.notify_status(node)                
                if self.distribute_flag:
                    com_uuid = self.distribute_node(node)
                    if com_uuid:
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                response.success = False
                response.error = 2
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_aborted_flag: {type(exc)} - {exc}')
            response.success = False
            response.error = 3
            response.reason = f'Exception: {exc}'
        #print("set_aborted_flag resp:", response)
        return response

    def set_paused_flag(self, request, response):
        try:
            #print("set_paused_flag:", request)            
            if request.id in self.idnodes:
                node = self.idnodes[request.id]
                node.set_paused_flag(request.val)
                # print("set_paused_flag TYPE:", node.get_type())
                self.notify_status(node)                
                if self.distribute_flag:
                    com_uuid = self.distribute_node(node)
                    if com_uuid:
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                response.success = False
                response.error = 2
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_paused_flag: {type(exc)} - {exc}')
            response.success = False
            response.error = 3
            response.reason = f'Exception: {exc}'
        #print("set_paused_flag resp:", response)
        return response

    def set_waiting_flag(self, request, response):
        try:
            #print("set_waiting_flag:", request)            
            if request.id in self.idnodes:
                node = self.idnodes[request.id]
                node.set_waiting_flag(request.val)
                # print("set_waiting_flag TYPE:", node.get_type())
                self.notify_status(node)                
                if self.distribute_flag:
                    com_uuid = self.distribute_node(node)
                    if com_uuid:
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                response.success = False
                response.error = 2
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_waiting_flag: {type(exc)} - {exc}')
            response.success = False
            response.error = 3
            response.reason = f'Exception: {exc}'
        #print("set_waiting_flag resp:", response)
        return response

    def set_executing_flag(self, request, response):
        try:
            #print("set_executing_flag:", request)            
            if request.id in self.idnodes:
                node = self.idnodes[request.id]                
                node.set_executing_flag(request.val)
                # print("set_executing_flag TYPE:", node.get_type())
                self.notify_status(node)                
                if self.distribute_flag:
                    com_uuid = self.distribute_node(node)
                    if com_uuid:
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.success = False
                    response.error = 2
                    response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_executing_flag: {type(exc)} - {exc}')
            response.success = False
            response.error = 3
            response.reason = f'Exception: {exc}'
        #print("set_executing_flag resp:", response)
        return response

    def set_node_parameter_string(self, request, response):
        try:
            if request.id in self.idnodes:
                node = self.idnodes[request.id]
                node.set_node_parameter(request.name, request.value)
                if self.distribute_flag:
                    com_uuid = self.distribute_node(node)
                    if com_uuid:
                        response.success = True
                        response.error = 0
                        response.reason = ""
                    else:
                        response.success = False
                        response.error = 1
                        response.reason = "Could not find node"
                else:
                    response.success = True
                    response.error = 0
                    response.reason = ""
            else:
                response.success = False
                response.error = 2
                response.reason = "Did not find node"
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION set_waiting_flag: {type(exc)} - {exc}')
            response.success = False
            response.error = 3
            response.reason = f'Exception: {exc}'
        #print("set_node_parameter resp:", response)
        return response

    
     
def main(args=None):
    print('Hi from tst exec factory server.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=8)
    node = ExecFactory()
    executor.add_node(node)
    
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    # rclpy.spin(node)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    #node.destroy_node()
    #rclpy.shutdown()
    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()

