import time
import sys
import uuid
import json
import threading
from threading import Lock
from timeit import default_timer as timer
from collections import deque

import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup

from lrs_msgs_tst.msg import TSTStartExecutor, TSTSignalExecutor
import lrs_srvs_exec
from lrs_srvs_exec.srv import TSTCreateExecutor
from lrs_srvs_tst.srv import TSTGetNodeFromUuid

class ExecManager(Node):

    def __init__(self):
        super().__init__('tst_exec_manager')
        self.group = ReentrantCallbackGroup()
        self.start_sub = self.create_subscription(TSTStartExecutor, "/tst_start_executor", self.start_executor_callback, 10, callback_group=self.group)
        self.signal_sub = self.create_subscription(TSTSignalExecutor, "tst_signal_executor", self.signal_executor_callback, 10, callback_group=self.group)
        self.get_node_from_uuid_client = self.create_client(TSTGetNodeFromUuid, "exec/tstfactory/get_node_from_uuid", callback_group=self.group)        
        self.create_clients = {}
        self.start_clients = {}
        self.signal_clients = {}
        self.verbose = False
        self.already_created = {}
        self.already_started = {}
        self.start_executor_queue = deque()
        self.signal_executor_queue = deque()
        self.inside_timer = False
        self.timer = self.create_timer(0.1, self.timer_callback, callback_group=self.group)

    def timer_callback(self):
        try:
            if self.inside_timer:
                return
            # self.get_logger().error(f'timer_callback')            
            self.inside_timer = True
            finished = False
            while not finished:
                try:
                    msg = self.start_executor_queue.popleft()
                    self.start_executor_handler(msg)
                except Exception as exc:
                    finished = True
            finished = False
            while not finished:
                try:
                    msg = self.signal_executor_queue.popleft()
                    self.signal_executor_handler(msg)
                except Exception as exc:
                    finished = True
            self.inside_timer = False
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')            
            # self.get_logger().error(f'timer_callback exit')            
        self.inside_timer = False
        
    def lprint(self, *val):
        if self.verbose:
            print(val)

    def wait_future(self, future, print_waiting=False):
        # self.get_logger().error(f'wait_future: {future}')            
        while not future.done():
            if print_waiting:
                print(f"wait_future executor")
        resp = future.result()
        # self.get_logger().error(f'wait_future finished: {future}')                    
        return resp

    def get_node_from_uuid(self, nuuid):
        # self.lprint("get_node_from_uuid:", nuuid)
        req = TSTGetNodeFromUuid.Request()
        req.node_uuid = nuuid
        future = self.get_node_from_uuid_client.call_async(req)
        resp = self.wait_future(future)        
        # self.lprint("get_node_from_uuid resp:", resp)
        return resp
    
    def create_executor(self, node_type, node_id, prepare):
        self.get_logger().error(f'tst_exec_manager create_executor: {node_type} - {node_id} - {prepare}')
        ntype = node_type.replace("-", "_")        
        try:
            if ntype not in self.create_clients:
                client = self.create_client(TSTCreateExecutor, f'tst_executor/{ntype}/create_executor', callback_group=self.group)
                self.create_clients[ntype] = client
                self.get_logger().info(f'tst_exec_manager create_executor: client created: {node_type} - {node_id} - {prepare}')
            req = TSTCreateExecutor.Request()
            req.id = node_id
            req.run_prepare = prepare
            self.get_logger().info(f'tst_exec_manager call create_executor: {req}')
            future = self.create_clients[ntype].call_async(req)
            i = 0
            while not future.done():
                time.sleep(0.1)
                i +=1
                if i>10:
                    i = 0
                    self.get_logger().error(f'tst_exec_manager call create_executor AGAIN: {req}')
                    future = self.create_clients[ntype].call_async(req)
            resp = future.result()
            #self.get_logger().info(f'tst_exec_manager call create_executor: have future')
            #resp = self.wait_future(future)
            #self.get_logger().info(f'tst_exec_manager create_executor resp: {resp}')            
            return resp.success
            #return True
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')            
        return False

    
            
    def signal_executor(self, node_type, node_uuid, signal):
        self.get_logger().error(f'signal_executor: {node_type} {node_uuid}')
        ntype = node_type.replace("-", "_")
        self.get_logger().error(f'signal_executor ntype: {ntype}')                        
        try:
            if not ntype in self.signal_clients:
                client = self.create_client(lrs_srvs_exec.srv.TSTSignalExecutor, f'tst_executor/{ntype}/signal_executor', callback_group=self.group)
                self.signal_clients[ntype] = client
            req = lrs_srvs_exec.srv.TSTSignalExecutor.Request()
            req.node_uuid = node_uuid
            req.signal = signal
            future = self.signal_clients[ntype].call_async(req)
            resp = self.wait_future(future)
            self.get_logger().error(f'signal_executor RESP: {resp}')                        
            return resp.success
        except Exception as exc:
            self.get_logger().error(f'Exception: {type(exc)} - {exc}')                        
        return False
    
    def start_executor(self, node_type, node_id):
        self.get_logger().error(f'start_executor: {node_type} - {node_id}')                    
        ntype = node_type.replace("-", "_")
        try:
            if not ntype in self.start_clients:
                client = self.create_client(lrs_srvs_exec.srv.TSTStartExecutor, f'tst_executor/{ntype}/start_executor', callback_group=self.group)
                self.start_clients[ntype] = client
            req = lrs_srvs_exec.srv.TSTStartExecutor.Request()
            req.id = node_id
            self.get_logger().info(f'start_executor: client call: {req}')                    
            future = self.start_clients[ntype].call_async(req)
            i = 0
            while not future.done():
                time.sleep(0.1)
                i +=1
                if i>10:
                    i = 0
                    self.get_logger().error(f'tst_exec_manager call start_executor AGAIN: {req}')
                    future = self.start_clients[ntype].call_async(req)
            resp = future.result()                    
            #resp = self.wait_future(future)
            #self.get_logger().info(f'start_executor: client call have future: {resp} - {future}')                             #self.get_logger().info(f'start_executor resp: {node_type} - {node_id} - {resp}')                    
            return resp.success
            # return True
        except Exception as exc:
            self.get_logger().error(f'Exception start_executor: {type(exc)} - {exc}')            
        return False
    
    def start_executor_handler(self, msg):
        self.get_logger().info(f'start_executor_handler: {msg}')
        start = timer()
        try:
            if msg.exec_ns != self.get_namespace():
                self.get_logger().error(f'start_executor_handler - wrong namespace: {self.get_namespace} - {msg.exec_ns}')
                return
            
            if not msg.id in self.already_created or not self.already_created[msg.id]:
                self.get_logger().info(f'start_executor_handler: CREATE EXECUTOR {msg}')
                cstart = timer()
                if self.create_executor(msg.type, msg.id, True):
                    self.already_created[msg.id] = True 
                else:
                    self.get_logger().error(f'start_executor_handler: failed to create executor.')
                    cend = timer()
                    self.get_logger().info(f'start_executor_handler: EXECUTOR CREATED {msg}')                
            else:
                self.get_logger().error(f'start_executor_callback create duplicate id: {msg}')            

            # time.sleep(0.2)
            
            if msg.id in self.already_created and self.already_created[msg.id]:
                self.get_logger().info(f'start_executor_handler: started: {msg.type} {msg.id}')
                if not msg.id in self.already_started or not self.already_started[msg.id]:
                    if self.start_executor(msg.type, msg.id):
                        self.get_logger().info(f'start_executor_callback: started: {msg.type} {msg.id}')
                        self.already_started[msg.id] = True
                    else:
                        self.get_logger().error(f'start_executor_callback: failed to start executor.')            
            else:
                self.get_logger().error(f'start_executor_callback: executor already_started.')
                
        except Exception as exc:
            self.get_logger().error(f'Exception start_executor_callback: {type(exc)} - {exc}')
            end = timer()
            print("start_executor_callback timing:", (end-start)-(cend-cstart), cend-cstart)

    def start_executor_callback(self, msg):
        try:
            self.get_logger().info(f'start_executor_callback: {msg}')        
            self.start_executor_queue.append(msg)
        except Exception as exc:
            self.get_logger().error(f'Exception start_executor_callback: {type(exc)} - {exc}')
            
    def signal_executor_handler(self, msg):
        self.get_logger().error(f'signal_executor_handler: {msg}')                    
        try:        
            if (msg.type == ""):  # intended for external ROS nodes
                return

            ntype = msg.type
            if ntype == "unknown" or msg.exec_ns == "":
                # lookup ntype... if node does not exist return
                resp = self.get_node_from_uuid(msg.node_uuid)
                if resp.success:
                    ntype = resp.info.type
                    self.get_logger().info(f'signal_executor_handler: found node: setting type to: {ntype}')                                        
                else:
                    self.get_logger().error(f'signal_executor_handler: failed to find node with uuid: {msg.node_uuid}')                    
                    return
                
            # if exec_ns is empty string, we did not know who was executing it

            #if msg.exec_ns == self.get_namespace() or msg.exec_ns == "": # old way...


            if msg.exec_ns != self.get_namespace():
                ntype = "remote"
            
            if True:
                if self.signal_executor(ntype, msg.node_uuid, msg.signal):
                    self.get_logger().info(f'signal_executor_handler: signal_executor succeeded: {ntype} - {msg.node_uuid} - {msg.signal}')
                else:
                    self.get_logger().error(f'signal_executor_handler: Failed to signal executor: {ntype} {msg.signal}')
        except Exception as exc:
            self.get_logger().error(f'Exception start_executor_handler: {type(exc)} - {exc}')

    def signal_executor_callback(self, msg):
        try:
            self.get_logger().error(f'signal_executor_callback: {msg}')        
            self.signal_executor_queue.append(msg)
        except Exception as exc:
            self.get_logger().error(f'Exception signal_executor_callback: {type(exc)} - {exc}')
        
def main(args=None):
    print('Hi from tst exec manager server.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=8)
    node = ExecManager()
    executor.add_node(node)

    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()

