# How expansion works:
# 
# - There is a tst_expansion node that have a service TSTExpandTree ("tst_expand_tree"). The tree is in JSON format.
# - All running executor factories have a service call that can do expansion if needed.
# - When we get a service call to expand a tree the following happens:
#   1. For each terminal node that is expandable publish a TSTExpandCommand message on topic /tst_expand_command.
#   1. /tst_expand_command messages are distributed to all agents (using ROS2 topic or JSON-forwarding).
#   1. The tst_expansion node receives the TSTExpandCommand messages and if the expansion should be.
#      done on the agent then we do a service call to the correct ExecutorFactory that do the actual
#      expansion.
#   1. The result of the service call is published on the topic /tst_expand_response as a TSTExpandResponse message.
#   1. /tst_expand_command messages are distributed to all agents (using ROS2 topic or JSON-forwarding)       
#   1. We wait for /tst_expand_response messages for all outstanding requests and link in possible expansion
#           in the tree.
#   1. If the tree was expanded then repeat this procedure until the tree is not expanding anymore.
#   1. Return the result to the initial service call.
# 



import time
import sys
import uuid
import json
import threading
from threading import Lock

import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup

from lrs_msgs_tst.msg import TSTExpandCommand, TSTExpandResponse
from lrs_srvs_tst.srv import TSTExpandTree

from lrs_srvs_exec.srv import TSTExecutorJsonExpand
from lrs_exec.expand import get_exec_unit
from lrs_msgs_common.msg import UnitInfoSlow

class Expansion(Node):

    def __init__(self):
        super().__init__('tst_expansion')
        self.group = ReentrantCallbackGroup()

        self.declare_parameter("external", False)
        self.external = self.get_parameter('external').get_parameter_value().bool_value
        external_prefix = ""
        if self.external:
            external_prefix = "/ext"

        self.get_logger().info(f'tst_expansion external_prefix: {external_prefix}')            
        
        self.expand_tree_srv = self.create_service(TSTExpandTree, "tst_expand_tree", self.expand_tree, callback_group=self.group)
        self.expand_command_sub = self.create_subscription(TSTExpandCommand, "/tst_expand_command", self.expand_command_callback, 10, callback_group=self.group)
        self.expand_response_sub = self.create_subscription(TSTExpandResponse, "/tst_expand_response", self.expand_response_callback, 10, callback_group=self.group)
        self.unit_info_slow = self.create_subscription(UnitInfoSlow, "/unit_info_slow", self.unit_info_slow_callback, 10, callback_group=self.group)        

        self.expand_command_pub = self.create_publisher(TSTExpandCommand, f"{external_prefix}/tst_expand_command", 10, callback_group=self.group)
        self.expand_response_pub = self.create_publisher(TSTExpandResponse, f"{external_prefix}/tst_expand_response", 10, callback_group=self.group)

        self.com_uuid_to_jobj = {}
        self.trees = {}
        self.expand_clients = {}
        self.expandable_types = None
        self.have_expandable_types = False

    def wait_future(self, future, print_waiting=False):
        try:
            while not future.done():
                if print_waiting:
                    print(f"wait_future executor")
            resp = future.result()
            return resp
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION wait_future: {type(exc)} - {exc}')
        
        
    def do_expand_tree(self, jobj):
        try:
            self.get_logger().error(f'do_expand_tree: {jobj}')
            if len(jobj["children"]) > 0:
                # self.get_logger().error(f'do_expand_tree MSG: recursive expand')
                for c in jobj["children"]:
                    self.do_expand_tree(c)
            else:
                # self.get_logger().error(f'do_expand_tree MSG: TERMINAL expand')
                ns = get_exec_unit(jobj)
                # self.get_logger().error(f'do_expand_tree MSG: TERMINAL expand: NS: {ns}')
                if ns:
                    name = jobj["name"].replace("-", "_")
                    self.get_logger().info(f'name: {name}')
                    self.get_logger().info(f'have_expandable_types: {self.have_expandable_types}')
                    self.get_logger().info(f'expandable_types: {self.expandable_types}')
                    if not self.have_expandable_types or name in self.expandable_types:
                        msg = TSTExpandCommand()
                        com_uuid = str(uuid.uuid4())
                        self.com_uuid_to_jobj[com_uuid] = jobj
                        jobj["expanded"] = False
                        msg.com_uuid = com_uuid
                        msg.ns = ns
                        msg.type = jobj["name"]
                        msg.json = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
                        self.get_logger().error(f'do_expand_tree MSG: {msg.type}')
                        self.expand_command_pub.publish(msg)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION do_expand_tree: {type(exc)} - {exc}')
                    
    def is_expanded(self, jobj):
        # print("is_expanded:", jobj)
        try:
            if "expanded" in jobj:
                flag = jobj["expanded"]
            else:
                flag = True
            if "children" in jobj:
                for c in jobj["children"]:
                    flag &= self.is_expanded(c)
            return flag
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION is_expanded: {type(exc)} - {exc}')
        return True

    def remove_expanded(self, jobj):
        try:
            if "expanded" in jobj:
                del jobj["expanded"]
            if "children" in jobj:
                for c in jobj["children"]:
                    self.remove_expanded(c)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION remove_expanded: {type(exc)} - {exc}')

    def get_n_nodes(self, jobj):
        try:
            res = 1
            if "children" in jobj:
                for c in jobj["children"]:
                    res += self.get_n_nodes(c)
            return res
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION get_n_nodes: {type(exc)} - {exc}')
        

    def expand_tree(self, request, response):
        try:
            # self.get_logger().error(f'expand_tree: {request}')            
            jobj = json.loads(request.input_json)
            #tree_uuid = str(uuid.uuid4())
            #self.trees[tree_uuid] = jobj
            
            old_n_nodes = 0
            n_nodes = self.get_n_nodes(jobj)
            while old_n_nodes != n_nodes:
                # print("Expansion Number of NODES:", old_n_nodes, n_nodes)
                old_n_nodes = n_nodes
                self.do_expand_tree(jobj)
                while not self.is_expanded(jobj):
                    # print("wait for expansion")
                    #print(json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))
                    time.sleep(0.1)
                self.remove_expanded(jobj)
                n_nodes = self.get_n_nodes(jobj)
                
            response.expanded_json = json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': '))
            response.success = True
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION expand_tree: {type(exc)} - {exc}')
            response.expanded_json = ""
            response.success = False
            response.error = 1
            response.reason = str(exc)
        return response

    def expand_command_callback(self, msg):        
        try:
            if msg.ns == self.get_namespace():
                # self.get_logger().error(f'expand_command_callback: {msg}')
                # print("expand_command_callback:", msg)
                # jobj = json.loads(msg.json)
                # print(json.dumps(jobj, sort_keys=True, indent=4, separators=(',', ': ')))

                resp = None
                
                if msg.type:
                    type = msg.type.replace("-","_")
                    # print("TYPE:", type)
                    if not type in self.expand_clients:
                        self.expand_clients[type] = self.create_client(TSTExecutorJsonExpand, f'tst_executor/{type}/executor_json_expand', callback_group=self.group)
                    self.get_logger().error(f'CLIENT: {self.expand_clients[type]}')
                    req = TSTExecutorJsonExpand.Request()
                    req.node_json = msg.json
                    # print("SERVICE:", f'tst_executor/{type}/executor_json_expand')
                    # print("REQ:", req)
                    future = self.expand_clients[type].call_async(req)
                    resp = self.wait_future(future)
                    # self.get_logger().error(f'RESP: {resp}')
                    
                respmsg = TSTExpandResponse()
                respmsg.response_to = msg.com_uuid
                respmsg.ns = msg.ns
                if resp:
                    respmsg.json = resp.expansion_json
                else:
                    respmsg.json = ""
                self.expand_response_pub.publish(respmsg)
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION expand_command_callback: {type(exc)} - {exc}')

    def expand_response_callback(self, msg):
        try:
            # print("expand_response_callback:", msg)
            if msg.response_to in self.com_uuid_to_jobj:
                # print("******EXPANDABLE NODE FOUND: Attaching children*****")
                jobj = self.com_uuid_to_jobj[msg.response_to]
                jobj["expanded"] = True
                if msg.json:
                    jobj["children"].append(json.loads(msg.json))
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION expand_response_callback: {type(exc)} - {exc}')

    def unit_info_slow_callback(self, msg):
        if msg.ns == self.get_namespace():        
            if not self.have_expandable_types:
                self.get_logger().error(f'got expandable types - ready to expand')
            self.have_expandable_types = True
            self.expandable_types = msg.expandable_capabilities
            self.get_logger().error(f'expandable types: {self.expandable_types}')


def main(args=None):
    print('Hi from tst expansion server.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=28)
    node = Expansion()
    executor.add_node(node)

    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    # rclpy.spin(node)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    #node.destroy_node()
    #rclpy.shutdown()
    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()

        
