import time
import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup

from lrs_msgs_tst.msg import TSTNodeStatus, TSTNodeStatusList
from lrs_srvs_tst.srv import TSTGetTreeStringFromUuid
from lrs_msgs_common.msg import DisplayRequest, ResourcesInfo
from cv_bridge import CvBridge, CvBridgeError
from sensor_msgs.msg import CompressedImage

import pydot
import cv2
import numpy as np

class TstStreamer(Node):

    def __init__(self):
        super().__init__('tst_exec_manager')

        self.declare_parameter("display_type", "all")
        self.display_type = self.get_parameter("display_type").value
        self.declare_parameter("channel", 0)
        self.channel = self.get_parameter("channel").value

        self.been_finished = {}
        self.succeeded = {}
        self.fps = 20
        self.graph = None
        self.width = 1280
        self.height = 720
        self.bridge = CvBridge()
        self.nsc_list = []
        self.tree_started = False

        self.can_be_aborted = {}
        self.can_be_paused = {}
        self.can_be_enoughed = {}
        self.can_be_signaled = {}
        self.resources_info = None
        
        self.group = ReentrantCallbackGroup()
        self.resources_info_sub = self.create_subscription(ResourcesInfo, "resources_info", self.resources_info_callback, 1, callback_group=self.group)        
        self.display_request_sub = self.create_subscription(DisplayRequest, "tst_display_request", self.display_request, 10, callback_group=self.group)
        self.node_status_sub = self.create_subscription(TSTNodeStatus, "tst_node_status", self.node_status, 1000, callback_group=self.group)
        self.node_status_list_sub = self.create_subscription(TSTNodeStatusList, "tst_node_status_list", self.node_status_list, 10, callback_group=self.group)
        self.get_tree_string_client = self.create_client(TSTGetTreeStringFromUuid, f'exec/tstfactory/get_tree_string_from_uuid', callback_group=self.group)
        if self.channel == 0:
            self.image_pub = self.create_publisher(CompressedImage, f"tststream_{self.display_type}/image_raw/compressed", 10, callback_group=self.group)
        else:
            self.image_pub = self.create_publisher(CompressedImage, f"tststream{self.channel}_{self.display_type}/image_raw/compressed", 10, callback_group=self.group)

        self.timer = self.create_timer(1.0/self.fps, self.timer_callback, callback_group=self.group)         

    def wait_future(self, future, print_waiting=False):
        while not future.done():
            if print_waiting:
                print(f"wait_future executor")
        resp = future.result()
        return resp

    def get_tree_string(self, uuid, format):
        self.get_logger().info(f'get_tree_string: {uuid} {format}')
        try:
            req = TSTGetTreeStringFromUuid.Request()
            req.uuid = uuid
            req.type = format
            future = self.get_tree_string_client.call_async(req)
            resp = self.wait_future(future, print_waiting=False)
            # self.get_logger().info(f'get_tree_string: {resp}')
            return resp
        except Exception as exc:
            self.get_logger().error(f'Exception get_tree_string: {type(exc)} - {exc}')
        return None

    def update_node(self, data):
        ## key = "node_" + str(data.root_id) + "_" + str(data.unique_node_id)
        key = "node_" + str(data.root_id) + "_" + str(data.id)
        if not self.graph:
            return
        nd = self.graph.get_node(key)

        # data arrives in the wrong order
        if not data.id in self.been_finished:
            self.been_finished[data.id] = False
            self.succeeded[data.id] = False

        if data.exec_status.finished:
            self.been_finished[data.id] = True
            self.succeeded[data.id] = data.exec_status.succeeded
            
        # print("NUMBER OF NODES:", len(nd))
        if nd:
            # print("CHANGE STATUS NODE:", key)
            ## nd[0].set_label(key)
            # self.get_logger().info(f'UPDATE_NODE: {data.id} - {data.exec_status}')
            if data.exec_status.executing:
                self.tree_started = True
                nd[0].set_fillcolor("green")
            if data.exec_status.paused:
                nd[0].set_fillcolor("mistyrose")
            if self.been_finished[data.id]:
                nd[0].set_fillcolor("yellow")
            if self.been_finished[data.id] and not self.succeeded[data.id] and not data.exec_status.aborted:
                nd[0].set_fillcolor("red")
                ## nd[0].set_label(data.exec_status.fail_reason)
            if self.been_finished[data.id] and data.exec_status.aborted:
                nd[0].set_fillcolor("blue")
            if data.exec_status.revoked:
                nd[0].set_fillcolor("gray")
            if data.exec_status.was_finished:
                nd[0].set_fillcolor("gray")
            if not data.exec_status.active and not data.exec_status.finished and not data.exec_status.executing and not data.exec_status.revoked:
                #nd[0].set_fillcolor("seagreen")            
                nd[0].set_fillcolor("cornsilk2")
            # APES
            lstr = ""
            if data.exec_info.can_be_aborted and not data.id in self.can_be_aborted:
                lstr += "A"
                self.can_be_aborted[data.id] = True
            if data.exec_info.can_be_paused and not data.id in self.can_be_paused:
                lstr += "P"
                self.can_be_paused[data.id] = True
            if data.exec_info.can_be_enoughed and not data.id in self.can_be_enoughed:
                lstr += "E"
                self.can_be_enoughed[data.id] = True
            if data.exec_info.can_be_signaled and not data.id in self.can_be_signaled:
                lstr += "S"
                self.can_be_signaled[data.id] = True
            if lstr:
                oldlabel = nd[0].get_label()
                oldlabel = oldlabel.lstrip('"')
                oldlabel = oldlabel.rstrip('"')
                newlabel = f'{oldlabel}\n{lstr}'
                #print("OLDLABEL:", oldlabel)
                #print("NEWLABEL:", newlabel)
                nd[0].set_label(newlabel)
            return True
        return False

    def node_status_list(self, msg):
        pass
    
    def node_status(self, msg):
        try:
            # self.get_logger().info(f'NODE-STATUS: {msg.id} - {msg.exec_status}')            
            if not self.graph:
                return
            new_nsc_list = []
            for el in self.nsc_list:
                if self.update_node(el):
                    pass
                else:
                    new_nsc_list.append(el)
            if self.update_node(msg):
                pass
            else:
                new_nsc_list.append(msg)
            self.nsc_list = new_nsc_list
        except Exception as exc:
            self.get_logger().error(f'Exception node_status: {type(exc)} - {exc}')

    def resources_info_callback(self, msg):
        #print("CALLBACK:", msg)
        if msg.ns == self.get_namespace():
            #print("STORE CALLBACK:", msg)
            self.resources_info = msg
            
    def display_request(self, msg):
        try:
            self.get_logger().info(f'display_request: {msg}')
            if self.channel != msg.channel:
                return
            if self.display_type == "all" or self.display_type == msg.type:
                if msg.json == "":
                    resp = self.get_tree_string(msg.root_uuid, 2)
                    print("display_request dot string:", resp)
                    if resp.success:
                        self.handle_dot_string(resp.data)
                        self.tree_started = False
                else:
                    pass
                    
        except Exception as exc:
            self.get_logger().error(f'Exception display_request: {type(exc)} - {exc}')

    def handle_dot_string(self, dot_string):
        try:
            # print("DOTSTRING:", dot_string)
            graphs = pydot.graph_from_dot_data(dot_string)
            self.graph = graphs[0]
            # print(self.graph)
            self.graph.write_jpg('output.jpg')
            #node_list = self.graph.get_node_list()
            node_list = self.graph.get_nodes()
            for nd in node_list:
                label = nd.get_label()
                name = nd.get_name()
                # print("NODE:", name, label)
        except Exception as exc:
            self.get_logger().error(f'Exception handle_dot_string: {type(exc)} - {exc}')            

    def get_blank_image(self):
        time_text = time.strftime("%H:%M:%S")
        image2 = np.zeros((self.height,self.width,3), np.uint8)
        text_color = (255,255,255)
        cv2.putText(image2, time_text, (150,130), cv2.FONT_HERSHEY_PLAIN, 2.0, text_color, thickness=2)
        cv2.putText(image2, self.get_namespace(), (300,130), cv2.FONT_HERSHEY_PLAIN, 2.0, text_color, thickness=2)        
        return image2

    def add_resources_lock_info(self, image):
        sx = 20
        dx = 150
        dy = 60
        n = 0
        sy = self.height - 90
        text_color = (0, 0, 0)
        if self.resources_info:
            for ri in self.resources_info.resinfo:
                ### print("ADD RI:", ri, ri.locked)
                if ri.locked:
                    col=(0, 0, 255)                    
                else:
                    col=(0, 255, 0)
                cv2.rectangle(image, (sx, sy), (sx+dx, sy+dy), color=col,  thickness=-1)
                cv2.putText(image, ri.resource, (int(sx+5),int(sy+dy-20)), cv2.FONT_HERSHEY_PLAIN, 2.5, text_color, thickness=2)
                sx += dx+10
        
    def timer_callback(self):
        try:
            #print("timer_callback")
            if self.graph:
                jpg = self.graph.create_jpg()
                msg = CompressedImage()
                msg.header.stamp = self.get_clock().now().to_msg()                
                msg.header.frame_id = self.display_type
                msg.format = "jpeg"
                msg.data = jpg

                image = self.bridge.compressed_imgmsg_to_cv2(msg, "bgr8")
                [h, w, c] = image.shape
                wscale = float(w)/self.width
                hscale = float(h)/self.height
                if wscale > hscale:
                    ws = self.width
                    hs = int(h/wscale)
                else:
                    ws = int(w/hscale)
                    hs = self.height

                # print(width, height, w, h, ws, hs)
                img = cv2.resize(image, (ws, hs))
                image2 = cv2.copyMakeBorder(img, 0, self.height-hs, 0, self.width-ws, cv2.BORDER_CONSTANT,value=[255, 255, 255])

                # overlay time of day
                time_text = time.strftime("%H:%M:%S")
                text_color = (0, 0, 0)
                text_image = f'/{self.get_namespace().split("/")[1]}'
                offset = 400*3
                cv2.putText(image2, time_text, (offset+50,30), cv2.FONT_HERSHEY_PLAIN, 2.0, text_color, thickness=2)
                cv2.putText(image2, text_image, (offset+200,30), cv2.FONT_HERSHEY_PLAIN, 2.0, text_color, thickness=2)

                self.add_resources_lock_info(image2)
                
#                cv2.putText(image2, display_type, (300,30), cv2.FONT_HERSHEY_PLAIN, 2.0, text_color, thickness=2)
#                cv2.putText(image2, display_title, (400,30), cv2.FONT_HERSHEY_PLAIN, 2.0, text_color, thickness=2)


                
                msg2 = self.bridge.cv2_to_compressed_imgmsg(image2)
                msg2.header = msg.header
                self.image_pub.publish(msg2)
                # self.image_pub.publish(msg2)
            else:
                # print("timer_callback blank")
                image = self.get_blank_image()
                self.add_resources_lock_info(image)
                msg = self.bridge.cv2_to_compressed_imgmsg(image)
                msg.header.stamp = self.get_clock().now().to_msg()
                msg.header.frame_id = self.display_type
                self.image_pub.publish(msg)
        except Exception as exc:
            self.get_logger().error(f'Exception timer_callback: {type(exc)} - {exc}')                        

def main(args=None):
    print('Hi from tst streamer server.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=8)
    node = TstStreamer()
    executor.add_node(node)
    try:    
        executor.spin()
    except KeyboardInterrupt:
        pass
    executor.shutdown()
    rclpy.shutdown()

if __name__ == '__main__':
    main()

        
